/***************************************************************************
 * Copyright (c) 2002-2005, In-VIGO Team
 * Advanced Computing and Information Systems (ACIS) Laboratory
 * University of Florida
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * cache_attrib.c
 *
 * Author:  Renato J. Figueiredo, Ming Zhao
 *
 *    Functions that implement the gvfs proxy attribute cache.
 *    The attribute cache is stored in the local disk.
 *    Refer to cache_attrib.h for a description of the attribute cache
 *    organization
 *
 *
 * Updated: Jan., 2004
 *    Modified attributes cache functions to support write-back
 *    for data-cache;
 *    Implemented write-through with write-allocate for attributes
 *    cache;
 *    Added cache support for CREATE, REMOVE, RENAME, and LINK
 *    procedures;
 *    Corrected the problem that cache bank is not seeked properly
 *    in tag_match_and_inv and find_replace_frame functions;
 *    Corrected the problem that a cache_insert must be preceded by
 *    a cache_lookup because the block is not properly prepared.
 *
 * Updated:   Jan. 27, 2004
 *    Corrected the problem that the resources allocated RPC results
 *    are not freed in several places.
 *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <unistd.h>
#include <fcntl.h>
#include <math.h>

#include "nfs_prot.h"
#include "logging.h"
#include "cache_attrib.h"

/*
 * Global variables
 */

unsigned long int cache_size = MAX_CACHE_SIZE;
unsigned int cache_assoc = MAX_CACHE_ASSOC;
unsigned int cache_banks = MAX_CACHE_BANKS;
unsigned long int cache_sets; /* total number of cache sets */
unsigned long int sets_per_bank;
char *BaseCachePath;
char *SessionID;
char *scratch_path;
struct stat *statbuf;
cacheblock *block_scratch_RPC;
cacheblock *block_scratch_disk;
extern CLIENT *nfsdclient;


/*************************************************************************
 * Write back a dirty attr cache block
 * When write-back is enabled, write-through policy is actually adopted
 * for most attr cache updates. Only when a file is being appended, it
 * grows just in cache, but not in server, thus the cached cile size is
 * different. This is the only case that an attr cache block can become
 * dirty. When a dirty attr cache block is to be evicted, the proxy should
 * check whether the cached size agrees with the size in server. If not,
 * proxy update the size in server by writing a zero byte at the offset
 * of cached_size - 1. When large numbers of files are created, the attr
 * cache can quickly become out of space, so write-back is indeed
 * necessary. (TODO: make the cache fully associative)
 *************************************************************************/
int write_back_block(cacheblock *block) {
	attrstat res_attr;
	write3res res_write;
	writeargs arg;
	write3args arg3;

	Dprintf(D_CACHE, "write back acache block\n");

	switch(block->nfsprogid) {
	case 2:
		arg.file = block->contents.getattr_v2.handle;
		hexdump(arg.file.data, NFS_FHSIZE, D_CACHE);
		Dprintf(D_CACHE, "file size unknown; fetch attr from server\n");
		if((nfsproc_getattr_2(&arg.file, &res_attr, nfsdclient)) != RPC_SUCCESS ||
			res_attr.status != NFS_OK) {
			Dprintf(D_CACHE, "error fetch file attributes\n");
			return -1;
		}

		Dprintf(D_CACHE, "old size: %d, new_size: %d\n",
		res_attr.attrstat_u.attributes.size,
		block->contents.getattr_v2.attributes.size);
		if(res_attr.attrstat_u.attributes.size <
				block->contents.getattr_v2.attributes.size) {
			arg.offset = block->contents.getattr_v2.attributes.size - 1;
			arg.beginoffset = arg.offset;
			arg.totalcount = 1;
			arg.data.data_len = 1;
			if((arg.data.data_val = (char *)malloc(1)) == NULL)
				return -1;
			memset(arg.data.data_val, 0, 1);

			if((nfsproc_write_2(&arg, &res_attr, nfsdclient)) != RPC_SUCCESS ||
				res_attr.status != NFS_OK) {
	Dprintf(D_CACHE, "error write back\n");
	free(arg.data.data_val);
	return -1;
			}
			Dprintf(D_CACHE, "update at offset %d, return size as %d\n",
					 arg.offset, res_attr.attrstat_u.attributes.size);
			free(arg.data.data_val);
		}
		break;
	case 3:
		arg3.file.data.data_len = block->contents.getattr_v3.handle.length;
		if((arg3.file.data.data_val = (char *)malloc(arg3.file.data.data_len)) == NULL) return -1;
		if((memcpy(arg3.file.data.data_val, block->contents.getattr_v3.handle.handle,
							arg3.file.data.data_len))==NULL) {
			Dprintf(D_CACHE,"write_back_block: memcpy failed\n");
			return -1;
		}
		hexdump(arg3.file.data.data_val, arg3.file.data.data_len, D_CACHE);
		Dprintf(D_CACHE, "cached file size: %d\n", block->contents.getattr_v3.attributes.size);
//     Dprintf(D_CACHE, "file size unknown; fetch attr from server\n");
//     if((nfsproc_getattr_2(&arg.file, &res_attr, nfsdclient)) != RPC_SUCCESS ||
//     	res_attr.status != NFS_OK) {
//       Dprintf(D_CACHE, "error fetch file attributes\n");
//       return -1;
//     }

//     Dprintf(D_CACHE, "old size: %d, new_size: %d\n",
// 	  res_attr.attrstat_u.attributes.size,
// 	  block->contents.getattr_v2.attributes.size);
//     if(res_attr.attrstat_u.attributes.size <
//     		block->contents.getattr_v2.attributes.size) {
		if(block->contents.getattr_v3.attributes.size > 0) {
			arg3.offset = block->contents.getattr_v3.attributes.size - 1;
			arg3.count = 1;
			arg3.data.data_len = 1;
			if((arg3.data.data_val = (char *)malloc(1)) == NULL)
	return -1;
			memset(arg3.data.data_val, 0, 1);

			if((nfsproc3_write_3(&arg3, &res_write, nfsdclient)) != RPC_SUCCESS ||
				res_write.status != NFS3_OK) {
	Dprintf(D_CACHE, "error write back\n");
	free(arg3.data.data_val);
	return -1;
			}
			Dprintf(D_CACHE, "update at offset %d, return size as %d\n",
					 (int)arg3.offset, (int)res_write.write3res_u.resok.file_wcc.after.post_op_attr_u.attributes.size);
			free(arg3.data.data_val);
		}
		break;
	default:
		return -1;
	}

	return 0;
}

int flush_attr_cache(void) {
	int staterr;
	unsigned long int i;

	Dprintf(D_CACHE, "Flushing attribute cache...\n");
	for(i=0; i<cache_banks; i++) {
		/* Not yet checking for modified blocks!!! */
		if((strlen(BaseCachePath)+1+strlen(SessionID)+
	1+strlen("attr_cache_bank_")+16)>=MAX_STR) {
			Dprintf(L_ERROR,"cache block path too long\n");
			return -1;
		}
		sprintf(scratch_path,"%s/%s/attr_cache_bank_%lu",BaseCachePath,
			SessionID,i);
		staterr = stat(scratch_path, statbuf);
		if(staterr==0) { /* file exists; remove it */
			if((unlink(scratch_path))==-1) {
	Dprintf(L_ERROR, "Could not flush attribute cache bank!\n");
	return -1;
			}
		}
	}
	Dprintf(D_CACHE, "Done flushing attribute cache.\n");
	return 0;
}

/*************************************************************************
 * NFS-V2-LOOKUP cache entry manipulation functions
 *************************************************************************/

/*************************************************************************
 * Copy NFS-V2 lookup arguments into cacheblock data structure
 * that is passed as argument. Returns 0 if successful, -1 if
 * error
 *************************************************************************/
int argument_lookup_v2(diropargs *arg, cacheblock *block) {

	if(arg==NULL) {
		Dprintf(D_CACHE,"argument_lookup_v2: arg==NULL\n");
		return -1;
	}

	/* zero out cache block */
	bzero(block,sizeof(*block));

	/* mark block as NFS V2 lookup */
	block->nfsprogid = 2;
	block->nfsprocid = NFSPROC_LOOKUP;

	/* copy contents of argument to cache block */
	if((memcpy(&block->contents.lookup_v2.dir,&arg->dir,NFS_FHSIZE))==NULL) {
		Dprintf(D_CACHE,"argument_lookup_v2: memcpy failed\n");
		return -1;
	}
	if(strlen(arg->name) > MAX_STR) /* make sure string fits */ {
		Dprintf(D_CACHE,"argument_lookup_v2: string larger than MAX_STR\n");
		return -1;
	}
	strcpy((char *)&block->contents.lookup_v2.name,arg->name);
	return 0;
}

/*************************************************************************
 * Copy NFS-V3 lookup arguments into cacheblock data structure
 * that is passed as argument. Returns 0 if successful, -1 if
 * error
 *************************************************************************/
int argument_lookup_v3(diropargs3 *arg, cacheblock *block) {
	if(arg==NULL) {
		Dprintf(D_CACHE,"argument_lookup_v3: arg==NULL\n");
		return -1;
	}

	/* zero out cache block */
	bzero(block,sizeof(*block));

	/* mark block as NFS V2 lookup */
	block->nfsprogid = 3;
	block->nfsprocid = NFSPROC3_LOOKUP;

	/* copy contents of argument to cache block */
	block->contents.lookup_v3.dir.length = arg->dir.data.data_len;
	memset(block->contents.lookup_v3.dir.handle, 0, NFS3_FHSIZE);
	if((memcpy(block->contents.lookup_v3.dir.handle,
			 arg->dir.data.data_val, arg->dir.data.data_len))==NULL) {
		Dprintf(D_CACHE,"argument_lookup_v3: memcpy failed\n");
		return -1;
	}

	if(strlen(arg->name) > MAX_STR) /* make sure string fits */ {
		Dprintf(D_CACHE,"argument_lookup_v3: string larger than MAX_STR\n");
		return -1;
	}
	strcpy((char *)&block->contents.lookup_v3.name, arg->name);
	return 0;
}

/*************************************************************************
 * Copy NFS-V2 lookup results into cacheblock data structure
 * that is passed as argument. Returns 0 if successful, -1 if
 * error
 *************************************************************************/
int response_remote_RPC_lookup_v2(diropres *res, cacheblock *block) {

	if(res==NULL) {
		Dprintf(D_CACHE,"response_remote_RPC_lookup_v2: res==NULL\n");
		return -1;
	}

	block->contents.lookup_v2.status = res->status;
	if((memcpy(&block->contents.lookup_v2.file,
			 &res->diropres_u.diropres.file,NFS_FHSIZE))==NULL) {
		Dprintf(D_CACHE,"response_remote_RPC_lookup_v2: memcpy failed\n");
		return -1;
	}
	block->contents.lookup_v2.attributes = res->diropres_u.diropres.attributes;
	return 0;
}

/*************************************************************************
 * Copy NFS-V2 lookup cached data from cacheblock data structure to RPC
 * response that is passed as argument. Returns 0 if successful, -1 if
 * error
 *************************************************************************/
int response_from_disk_cache_lookup_v2(cacheblock *block, diropres *res) {

	if(res==NULL) {
		Dprintf(D_CACHE,"response_from_disk_cache_lookup_v2: res==NULL\n");
		return -1;
	}
	res->status = block->contents.lookup_v2.status;
	if((memcpy(&res->diropres_u.diropres.file,
			 &block->contents.lookup_v2.file,
			 NFS_FHSIZE))==NULL) {
		Dprintf(D_CACHE,"response_from_disk_cache_lookup_v2: memcpy failed\n");
		return -1;
	}
	res->diropres_u.diropres.attributes = block->contents.lookup_v2.attributes;
	return 0;
}

/*************************************************************************
 * Copy NFS-V3 lookup results into cacheblock data structure
 * that is passed as argument. Returns 0 if successful, -1 if
 * error
 *************************************************************************/
int response_remote_RPC_lookup_v3(lookup3res *res, cacheblock *block) {

	if(res==NULL) {
		Dprintf(D_CACHE,"response_remote_RPC_lookup_v3: res==NULL\n");
		return -1;
	}

	block->contents.lookup_v3.status = res->status;
	if(res->status == NFS3_OK) {
		block->contents.lookup_v3.file.length =
									res->lookup3res_u.resok.object.data.data_len;
		if((memcpy(block->contents.lookup_v3.file.handle,
								res->lookup3res_u.resok.object.data.data_val,
								res->lookup3res_u.resok.object.data.data_len))==NULL) {
			Dprintf(D_CACHE,"response_remote_RPC_lookup_v3: memcpy failed\n");
			return -1;
		}
	}
	return 0;
}

/*************************************************************************
 * Copy NFS-V3 lookup cached data from cacheblock data structure to RPC
 * response that is passed as argument. Returns 0 if successful, -1 if
 * error
 *************************************************************************/
int response_from_disk_cache_lookup_v3(cacheblock *block, lookup3res *res) {

	if(res==NULL) {
		Dprintf(D_CACHE,"response_from_disk_cache_lookup_v3: res==NULL\n");
		return -1;
	}
	res->status = block->contents.lookup_v3.status;
	res->lookup3res_u.resok.object.data.data_len = block->contents.lookup_v3.file.length;
	if((memcpy(res->lookup3res_u.resok.object.data.data_val,
			 block->contents.lookup_v3.file.handle,
			 block->contents.lookup_v3.file.length))==NULL) {
		Dprintf(D_CACHE,"response_from_disk_cache_lookup_v3: memcpy failed\n");
		return -1;
	}

	return 0;
}

/*************************************************************************
 * Hash V2 diropargs (fh of directory + file name)                       *
 * Return set number                                                     *
 * HERE: check efficiency of this hash                                   *
 *************************************************************************/
unsigned long int hash_diropargs_v2(nfs_fh *fhandle, char *filename) {
	int i,seof=0;
	double A=0.6180339887;
	unsigned long int B=0;

	/* accumulate file handle into B */
	for(i=0;i<NFS_FHSIZE;i++) {
		B+=(unsigned long int)(fhandle->data[i]);
	}

	/* also accumulate file name */
	for(i=0;seof==0;i++) {
		if(filename[i] == 0)
			seof=1;
		else
			B+=(unsigned long int)(filename[i]);
	}

	A = A * (double)B;
	A = A - floor(A);
	A = A * cache_sets;
	return (unsigned long int) (A) % cache_sets;
}

/*************************************************************************
 * Hash V3 diropargs (fh of directory + file name)
 * Return set number
 * HERE: check efficiency of this hash
 *************************************************************************/
unsigned long int hash_diropargs_v3(char *fhandle, char *filename) {
	int i,seof=0;
	double A=0.6180339887;
	unsigned long int B=0;

	/* accumulate file handle into B */
	for(i=0;i<NFS3_FHSIZE;i++) {
		B+=(unsigned long int)(fhandle[i]);
	}

	/* also accumulate file name */
	for(i=0;seof==0;i++) {
		if(filename[i] == 0)
			seof=1;
		else
			B+=(unsigned long int)(filename[i]);
	}

	A = A * (double)B;
	A = A - floor(A);
	A = A * cache_sets;
	return (unsigned long int) (A) % cache_sets;
}

/*************************************************************************
 * NFS-V2-GETATTR cache entry manipulation functions
 *************************************************************************/

/*************************************************************************
 * Copy NFS-V2 getattr arguments into cacheblock data structure
 * that is passed as argument. Returns 0 if successful, -1 if
 * error
 *************************************************************************/
int argument_getattr_v2(nfs_fh *arg, cacheblock *block) {
	if(arg==NULL) {
		Dprintf(D_CACHE,"argument_getattr_v2: arg==NULL\n");
		return -1;
	}

	/* zero out cache block */
	bzero(block,sizeof(*block));

	/* mark block as NFS V2 getattr */
	block->nfsprogid = 2;
	block->nfsprocid = NFSPROC_GETATTR;

	/* copy contents of argument to cache block */
	if((memcpy(&block->contents.getattr_v2.handle,
			 arg,NFS_FHSIZE))==NULL) {
		Dprintf(D_CACHE,"argument_getattr_v2: memcpy failed\n");
		return -1;
	}
	return 0;
}


/*************************************************************************
 * Copy NFS-V3 getattr arguments into cacheblock data structure
 * that is passed as argument. Returns 0 if successful, -1 if
 * error
 *************************************************************************/
int argument_getattr_v3(nfs_fh3 *arg, cacheblock *block) {
	if(arg==NULL) {
		Dprintf(D_CACHE,"argument_getattr_v3: arg==NULL\n");
		return -1;
	}

	/* zero out cache block */
	bzero(block,sizeof(*block));

	/* mark block as NFS V2 getattr */
	block->nfsprogid = 3;
	block->nfsprocid = NFSPROC3_GETATTR;

	/* copy contents of argument to cache block */
	block->contents.getattr_v3.handle.length = arg->data.data_len;
	memset(block->contents.getattr_v3.handle.handle, 0, NFS3_FHSIZE);
	if((memcpy(block->contents.getattr_v3.handle.handle,
			 arg->data.data_val, arg->data.data_len))==NULL) {
		Dprintf(D_CACHE,"argument_getattr_v3: memcpy failed\n");
		return -1;
	}
	return 0;
}

int argument_access_v3(nfs_fh3 *arg, cacheblock *block) {
	if(arg==NULL) {
		Dprintf(D_CACHE,"argument_getattr_v3: arg==NULL\n");
		return -1;
	}

	/* zero out cache block */
	bzero(block,sizeof(*block));

	/* mark block as NFS V2 getattr */
	block->nfsprogid = 3;
	block->nfsprocid = NFSPROC3_ACCESS;

	/* copy contents of argument to cache block */
	block->contents.getattr_v3.handle.length = arg->data.data_len;
	memset(block->contents.getattr_v3.handle.handle, 0, NFS3_FHSIZE);
	if((memcpy(block->contents.getattr_v3.handle.handle,
			 arg->data.data_val, arg->data.data_len))==NULL) {
		Dprintf(D_CACHE,"argument_getattr_v3: memcpy failed\n");
		return -1;
	}
	return 0;
}
/*************************************************************************
 * Copy NFS-V2 getattr results into cacheblock data structure
 * that is passed as argument. Returns 0 if successful, -1 if
 * error
 *************************************************************************/
int response_remote_RPC_getattr_v2(attrstat *res, cacheblock *block) {

	if(res==NULL) {
		Dprintf(D_CACHE,"response_remote_RPC_getattr_v2: res==NULL\n");
		return -1;
	}

	block->contents.getattr_v2.status = res->status;
	block->contents.getattr_v2.attributes =
		res->attrstat_u.attributes;

	return 0;
}

/*************************************************************************
 * Copy NFS-V2 getattr cached data from cacheblock data structure to RPC
 * response that is passed as argument. Returns 0 if successful, -1 if
 * error
 *************************************************************************/
int response_from_disk_cache_getattr_v2(cacheblock *block, attrstat *res) {

	if(res==NULL) {
		Dprintf(D_CACHE,"response_from_disk_cache_getattr_v2: res==NULL\n");
		return -1;
	}
	res->status = block->contents.getattr_v2.status;
	res->attrstat_u.attributes = block->contents.getattr_v2.attributes;
	return 0;
}

/*************************************************************************
 * Copy NFS-V3 getattr results into cacheblock data structure
 * that is passed as argument. Returns 0 if successful, -1 if
 * error
 *************************************************************************/
int response_remote_RPC_getattr_v3(getattr3res *res, cacheblock *block) {

	if(res==NULL) {
		Dprintf(D_CACHE,"response_remote_RPC_getattr_v3: res==NULL\n");
		return -1;
	}

	block->contents.getattr_v3.status = res->status;
	block->contents.getattr_v3.attributes = res->getattr3res_u.attributes;
	block->contents.getattr_v3.access = -1;

	return 0;
}

int response_remote_RPC_access_v3(access3res *res, cacheblock *block) {

	if(res==NULL) {
		Dprintf(D_CACHE,"response_remote_RPC_access_v3: res==NULL\n");
		return -1;
	}

	block->contents.getattr_v3.status = res->status;
	block->contents.getattr_v3.attributes =
		res->access3res_u.resok.obj_attributes.post_op_attr_u.attributes;
	block->contents.getattr_v3.access = res->access3res_u.resok.access;
	if(res->access3res_u.resok.access == -1) Dprintf(D_CACHE, "alarm\n");
	Dprintf(D_CACHE, "%d\n", res->access3res_u.resok.access);

	return 0;
}

/*************************************************************************
 * Copy NFS-V3 getattr cached data from cacheblock data structure to RPC
 * response that is passed as argument. Returns 0 if successful, -1 if
 * error
 *************************************************************************/
int response_from_disk_cache_getattr_v3(cacheblock *block, getattr3res *res) {

	if(res==NULL) {
		Dprintf(D_CACHE,"response_from_disk_cache_getattr_v3: res==NULL\n");
		return -1;
	}
	res->status = block->contents.getattr_v3.status;
	res->getattr3res_u.attributes = block->contents.getattr_v3.attributes;

	return 0;
}

int response_from_disk_cache_access_v3(cacheblock *block, access3res *res) {
	if(res==NULL) {
		Dprintf(D_CACHE,"response_from_disk_cache_access_v3: res==NULL\n");
		return -1;
	}
	res->status = block->contents.getattr_v3.status;
	res->access3res_u.resok.obj_attributes.post_op_attr_u.attributes = block->contents.getattr_v3.attributes;
	res->access3res_u.resok.obj_attributes.present = TRUE;
	res->access3res_u.resok.access = block->contents.getattr_v3.access;
	if(res->access3res_u.resok.access == -1) Dprintf(D_CACHE, "alarm\n");
	return 0;
}

unsigned long int hash_nfs_fh_v2(nfs_fh *fhandle) {
	return hash_diropargs_v2(fhandle,"xx");
}

unsigned long int hash_nfs_fh_v3(char *fhandle) {
	return hash_diropargs_v3(fhandle,"xx");
}

/*************************************************************************
 * Determine the file name that a cache block maps to                    *
 * Returns set number,                                                   *
 *  and the path to the cache bank file if successfull, -1 otherwise     *
 *************************************************************************/
long int block_file_path_and_set(cacheblock *block, char *path) {
	unsigned long int hash, bank_number;

	if(block==NULL) {
		Dprintf(D_CACHE,"block_file_path_and_set: block==NULL\n");
		return -1;
	}

	switch(block->nfsprogid) {
		case 2:
			switch(block->nfsprocid) {
				case NFSPROC_LOOKUP:
					hash = hash_diropargs_v2(&(block->contents.lookup_v2.dir),
								 (char *)&(block->contents.lookup_v2.name));
					bank_number = (unsigned long int) (hash/sets_per_bank);
					if((strlen(BaseCachePath)+1+strlen(SessionID)+
							1+strlen("attr_cache_bank_")+16)>MAX_STR) {
						Dprintf(L_ERROR,"cache block path too long\n");
						return -1;
					}
					sprintf(path,"%s/%s/attr_cache_bank_%lu",BaseCachePath,SessionID,
						bank_number);
					Dprintf(D_CACHE,"HASH lookup = %lu (bank %lu)\n",hash,bank_number);
					return (unsigned long int) hash;
				case NFSPROC_GETATTR:
				case NFSPROC_CREATE:
				case NFSPROC_LINK:
					hash = hash_nfs_fh_v2(&(block->contents.getattr_v2.handle));
					bank_number = (unsigned long int) (hash/sets_per_bank);
					if((strlen(BaseCachePath)+1+strlen(SessionID)+
							1+strlen("attr_cache_bank_")+16)>MAX_STR) {
						Dprintf(L_ERROR,"cache block path too long\n");
						return -1;
					}
					sprintf(path,"%s/%s/attr_cache_bank_%lu",BaseCachePath,SessionID,
						bank_number);
					Dprintf(D_CACHE,"HASH(%d) = %lu (bank %lu)\n",
							block->nfsprocid,hash,bank_number);
					return (unsigned long int) hash;
				default:
					return -1;
			} /* switch block->nfsprocid */
			break; /* case 2 */
		case 3:
			switch(block->nfsprocid) {
				case NFSPROC3_GETATTR:
				case NFSPROC3_ACCESS:
				case NFSPROC3_LINK:
					hash = hash_nfs_fh_v3(block->contents.getattr_v3.handle.handle);
					bank_number = (unsigned long int) (hash/sets_per_bank);
					if((strlen(BaseCachePath)+1+strlen(SessionID)+
							1+strlen("attr_cache_bank_")+16)>MAX_STR) {
						Dprintf(L_ERROR,"cache block path too long\n");
						return -1;
					}
					sprintf(path,"%s/%s/attr_cache_bank_%lu",BaseCachePath,SessionID,
						bank_number);
					Dprintf(D_CACHE,"HASH(%d) = %lu (bank %lu)\n",
							block->nfsprocid,hash,bank_number);
					return (unsigned long int) hash;
				case NFSPROC3_LOOKUP:
					hash = hash_diropargs_v3(block->contents.lookup_v3.dir.handle,
								 (char *)&(block->contents.lookup_v3.name));
					bank_number = (unsigned long int) (hash/sets_per_bank);
					if((strlen(BaseCachePath)+1+strlen(SessionID)+
							1+strlen("attr_cache_bank_")+16)>MAX_STR) {
						Dprintf(L_ERROR,"cache block path too long\n");
						return -1;
					}
					sprintf(path,"%s/%s/attr_cache_bank_%lu",BaseCachePath,SessionID,
						bank_number);
					Dprintf(D_CACHE,"HASH lookup = %lu (bank %lu)\n",hash,bank_number);
					return (unsigned long int) hash;
				default:
					return -1;
			} /* switch block->nfsprocid */
			break;
		default:
			return -1;
	}
	return -1;
}

/*************************************************************************
 * Check whether tag of NFS arguments matches cached data on disk,       *
 *  possibly invalidating cached data (if invalidate==1)                 *
 * block_RPC : has NFS arguments                                         *
 * block_disk: returns contents of a match on a hit                      *
 * return values:                                                        *
 * 0 : no match                                                          *
 * 1 : tag match                                                         *
 * -1: error                                                             *
 *************************************************************************/
int tag_match_and_inv(cacheblock *block_RPC, cacheblock *block_disk,
					int invalidate) {
	int fd,reads,framescan;
	int file_handle_match = 0;
	int file_name_match = 0;
	int overall_match = 0;
	int seekset, seekr, seekcur;
	unsigned long int set_number;
	int staterr;

	if(block_RPC==NULL) {
		Dprintf(D_CACHE,"tag_match: block_RPC==NULL\n");
		return -1;
	}

	/* obtain path and set number */
	set_number = block_file_path_and_set(block_RPC, scratch_path);
	if(set_number==-1) {
		Dprintf(D_CACHE,"tag_match: could not obtain set\n");
		return -1;
	}

	/* open & seek into 1st block of appropriate set in cache file */
	staterr = stat(scratch_path, statbuf);
	if(staterr != 0) {
		Dprintf(D_CACHE, "tag_match: bank does not exist\n");
		return 0;
	}
	if((fd = open(scratch_path, O_RDONLY)) == -1) {
		Dprintf(D_CACHE,"tag_match: could not open cache bank %s for reading\n",
			scratch_path);
		return -1;
	}

	/* obtain a shared lock for cache lookup */
	if ((flock(fd,LOCK_SH))==-1) {
		Dprintf(D_CACHE,"tag_match: could not lock cache bank %s for reading %d\n",
			scratch_path, fd);
		return -1;
	}

	seekset = (set_number%sets_per_bank)*cache_assoc*sizeof(*block_RPC);
	Dprintf(D_CACHE, "tag_match: seeking %lu -> %lu\n",set_number,seekset);
	seekr = lseek(fd,seekset,SEEK_SET);
	if(seekr != seekset) {
		Dprintf(D_CACHE,"tag_match: could not seek initial block %i in %s\n",
			seekset, scratch_path);
		flock(fd,LOCK_UN);
		close(fd);
		return -1;
	}

	/* associative lookup; traverse all sets */
	for(framescan=0; framescan<cache_assoc && !overall_match; framescan++) {

		reads = read(fd,block_disk,sizeof(*block_disk));
		if(reads != sizeof(*block_disk)) { /* read was not right */
			Dprintf(D_CACHE,"tag_match: could not read scan block %i in %s\n",
				framescan, scratch_path);
			flock(fd,LOCK_UN);
			close(fd);
			return -1;
		}

		/* different NFS calls require diff. tag matching */
		switch(block_RPC->nfsprogid) {
		case 2:
			switch(block_RPC->nfsprocid) {
			case NFSPROC_LOOKUP:
				if (block_disk->state == ST_INV) /* invalid block; no match */
					break;
				/* check if blocks' prog/proc ids match */
				if(!((block_RPC->nfsprogid==block_disk->nfsprogid) &&
						(block_RPC->nfsprocid==block_disk->nfsprocid)))
					break;
				/* check if file names match */
				file_name_match = strcmp((char *)&block_RPC->contents.lookup_v2.name,
							 (char *)&block_disk->contents.lookup_v2.name);
				if (file_name_match != 0)  /* no file name match */
					break;
				/* check if file handles match */
				file_handle_match = memcmp(&block_RPC->contents.lookup_v2.dir,
								 &block_disk->contents.lookup_v2.dir,
								 NFS_FHSIZE);
				overall_match = (file_handle_match==0? 1 : 0);
				if(!overall_match) break;

				/* invalidate block in disk if required */
				if(invalidate > 0) {
					/* keep current seek point in seekcur; will need it after
						 fd is reopened */
					seekcur = lseek(fd, 0, SEEK_CUR);
					/* release shared lock  */
					if ((flock(fd,LOCK_UN))==-1) {
						Dprintf(D_CACHE,
							"tag_match: could not unlock cache bank %s for reading\n",
							scratch_path);
						return -1;
					}
					/* reopen bank file in read/write mode  for invalidation */
					close(fd);
					if((fd = open(scratch_path, O_RDWR)) == -1) {
						Dprintf(D_CACHE,"tag_match: could not open cache bank %s R/W\n",
							scratch_path);
						return -1;
					}
					/* obtain exclusive lock for invalidation */
					if ((flock(fd,LOCK_EX))==-1) {
						Dprintf(D_CACHE,
							"tag_match: could not lock cache bank %s for writing\n",
							scratch_path);
						return -1;
					}
					/* seek back one block from seekcur */
					seekset = seekcur-sizeof(*block_RPC); /* point back one block */
					seekr = lseek(fd, seekset, SEEK_SET);
					if(seekr!=seekset) {
						Dprintf(D_CACHE, "tag_match: could not seek %i in file bank %s\n",
							seekset, scratch_path);
						flock(fd,LOCK_UN);
						close(fd);
						return -1;
					}
					if(invalidate == 1) {
						/* write an invalid block (all zeroes) */
						bzero(block_disk,sizeof(*block_disk));
					} else {
						memcpy(block_disk, block_RPC, sizeof(*block_disk));
						block_disk->state = ST_RO_CLEAN; /* Mark this block as valid */
					}
					if((write(fd,block_disk,sizeof(*block_disk)))!=sizeof(*block_disk)) {
						Dprintf(D_CACHE,"tag_match: could not invalidate block in %s\n",
							scratch_path);
						flock(fd,LOCK_UN);
						close(fd);
						return -1;
					}
				}
				break;

			case NFSPROC_GETATTR:
				if (block_disk->state == ST_INV) /* invalid block; no match */
					break;
				/* check if blocks' prog/proc ids match */
				if(!((block_RPC->nfsprogid==block_disk->nfsprogid) &&
						(block_RPC->nfsprocid==block_disk->nfsprocid)))
					break;
				/* check if file handles match */
				file_handle_match = memcmp(&block_RPC->contents.getattr_v2.handle,
								 &block_disk->contents.getattr_v2.handle,
								 NFS_FHSIZE);
				overall_match = (file_handle_match==0 ? 1 : 0);
				if(!overall_match) break;

				/* invalidate block in disk if required */
				if(invalidate) {
					/* keep current seek point in seekcur; will need it after
						 fd is reopened */
					seekcur = lseek(fd, 0, SEEK_CUR);
					/* release shared lock  */
					if ((flock(fd,LOCK_UN))==-1) {
						Dprintf(D_CACHE,
							"tag_match: could not unlock cache bank %s for reading\n",
							scratch_path);
						return -1;
					}
					/* reopen bank file in read/write mode  for invalidation */
					close(fd);
					if((fd = open(scratch_path, O_RDWR)) == -1) {
						Dprintf(D_CACHE,"tag_match: could not open cache bank %s R/W\n",
							scratch_path);
						return -1;
					}
					/* obtain exclusive lock for invalidation */
					if ((flock(fd,LOCK_EX))==-1) {
						Dprintf(D_CACHE,
							"tag_match: could not lock cache bank %s for writing\n",
							scratch_path);
						return -1;
					}
					/* seek back one block from seekcur */
					seekset = seekcur-sizeof(*block_RPC); /* point back one block */
					seekr = lseek(fd, seekset, SEEK_SET);
					if(seekr!=seekset) {
						Dprintf(D_CACHE, "tag_match: could not seek %i in file bank %s\n",
							seekset, scratch_path);
						flock(fd,LOCK_UN);
						close(fd);
						return -1;
					}

					if(invalidate == 1) {
						/* write an invalid block (all zeroes) */
						bzero(block_disk,sizeof(*block_disk));
					} else {
/*            if(block_disk->contents.getattr_v2.attributes.size > block_RPC->contents.getattr_v2.attributes.size) {
							Dprintf(D_CACHE, "keep the original size: %d > %d\n",
							block_disk->contents.getattr_v2.attributes.size, block_RPC->contents.getattr_v2.attributes.size);
							block_RPC->contents.getattr_v2.attributes.size = block_disk->contents.getattr_v2.attributes.size;
						}*/
						block_disk->contents.getattr_v2.attributes = block_RPC->contents.getattr_v2.attributes;
					}

					if((write(fd,block_disk,sizeof(*block_disk)))!=sizeof(*block_disk)) {
						Dprintf(D_CACHE,"tag_match: could not invalidate block in %s\n",
							scratch_path);
						flock(fd,LOCK_UN);
						close(fd);
						return -1;
					}
				} /* end if invalidate */
				break;

			default:
				return -1;
			} /* end switch(block_RPC->nfsprocid) */
			break; /* case 2 */

		case 3:
			switch(block_RPC->nfsprocid) {
			case NFSPROC3_GETATTR:
				if (block_disk->state == ST_INV) /* invalid block; no match */ {
					break;
				}
				/* check if blocks' prog/proc ids match */
				if(!((block_RPC->nfsprogid==block_disk->nfsprogid) &&
						(block_RPC->nfsprocid==block_disk->nfsprocid))) {
					break;
				}
				/* check if file handles match */
				file_handle_match =
						memcmp(block_RPC->contents.getattr_v3.handle.handle,
						block_disk->contents.getattr_v3.handle.handle,
						block_RPC->contents.getattr_v3.handle.length);
				overall_match = (file_handle_match==0 ? 1 : 0);
				if(!overall_match) break;

				/* invalidate block in disk if required */
				if(invalidate > 0) {
					/* keep current seek point in seekcur; will need it after
						 fd is reopened */
					seekcur = lseek(fd, 0, SEEK_CUR);
					/* release shared lock  */
					if ((flock(fd,LOCK_UN))==-1) {
						Dprintf(D_CACHE,
							"tag_match: could not unlock cache bank %s for reading\n",
							scratch_path);
						return -1;
					}
					/* reopen bank file in read/write mode  for invalidation */
					close(fd);
					if((fd = open(scratch_path, O_RDWR)) == -1) {
						Dprintf(D_CACHE,"tag_match: could not open cache bank %s R/W\n",
							scratch_path);
						return -1;
					}
					/* obtain exclusive lock for invalidation */
					if ((flock(fd,LOCK_EX))==-1) {
						Dprintf(D_CACHE,
							"tag_match: could not lock cache bank %s for writing\n",
							scratch_path);
						return -1;
					}
					/* seek back one block from seekcur */
					seekset = seekcur-sizeof(*block_RPC); /* point back one block */
					seekr = lseek(fd, seekset, SEEK_SET);
					if(seekr!=seekset) {
						Dprintf(D_CACHE, "tag_match: could not seek %i in file bank %s\n",
							seekset, scratch_path);
						flock(fd,LOCK_UN);
						close(fd);
						return -1;
					}

					if(invalidate == 1) {
						/* write an invalid block (all zeroes) */
						bzero(block_disk,sizeof(*block_disk));
					} else {
						block_disk->contents.getattr_v3.attributes = block_RPC->contents.getattr_v3.attributes ;
					}

					if((write(fd,block_disk,sizeof(*block_disk)))!=sizeof(*block_disk)) {
						Dprintf(D_CACHE,"tag_match: could not invalidate block in %s\n",
										scratch_path);
						flock(fd,LOCK_UN);
						close(fd);
						return -1;
					}
				} /* end if invalidate > 0 */
				break;

			case NFSPROC3_ACCESS:
				if (block_disk->state == ST_INV) /* invalid block; no match */ {
					break;
				}
				/* check if blocks' prog/proc ids match */
				if(!((block_RPC->nfsprogid==block_disk->nfsprogid) &&
						(NFSPROC3_GETATTR==block_disk->nfsprocid))) {
					break;
				}
				/* check if file handles match */
				file_handle_match =
						memcmp(block_RPC->contents.getattr_v3.handle.handle,
						block_disk->contents.getattr_v3.handle.handle,
						block_RPC->contents.getattr_v3.handle.length);
				overall_match = (file_handle_match==0 ? 1 : 0);
				if(!overall_match) break;

				/* invalidate block in disk if required */
				if(invalidate > 0) {
					/* keep current seek point in seekcur; will need it after
						 fd is reopened */
					seekcur = lseek(fd, 0, SEEK_CUR);
					/* release shared lock  */
					if ((flock(fd,LOCK_UN))==-1) {
						Dprintf(D_CACHE,
							"tag_match: could not unlock cache bank %s for reading\n",
							scratch_path);
						return -1;
					}
					/* reopen bank file in read/write mode for invalidation */
					close(fd);
					if((fd = open(scratch_path, O_RDWR)) == -1) {
						Dprintf(D_CACHE,"tag_match: could not open cache bank %s R/W\n",
							scratch_path);
						return -1;
					}
					/* obtain exclusive lock for invalidation */
					if ((flock(fd,LOCK_EX))==-1) {
						Dprintf(D_CACHE,
							"tag_match: could not lock cache bank %s for writing\n",
							scratch_path);
						return -1;
					}
					/* seek back one block from seekcur */
					seekset = seekcur-sizeof(*block_RPC); /* point back one block */
					seekr = lseek(fd, seekset, SEEK_SET);
					if(seekr!=seekset) {
						Dprintf(D_CACHE, "tag_match: could not seek %i in file bank %s\n",
							seekset, scratch_path);
						flock(fd,LOCK_UN);
						close(fd);
						return -1;
					}

					if(invalidate == 1) {
						/* write an invalid block (all zeroes) */
						bzero(block_disk,sizeof(*block_disk));
					} else {
						block_disk->contents.getattr_v3.access = block_RPC->contents.getattr_v3.access;
						Dprintf(D_CACHE, "%d\n", block_disk->contents.getattr_v3.access);
						block_disk->contents.getattr_v3.attributes = block_RPC->contents.getattr_v3.attributes ;
					}

					if((write(fd,block_disk,sizeof(*block_disk)))!=sizeof(*block_disk)) {
						Dprintf(D_CACHE,"tag_match: could not invalidate block in %s\n",
										scratch_path);
						flock(fd,LOCK_UN);
						close(fd);
						return -1;
					}
				} /* end if invalidate > 0 */
				break;


			case NFSPROC3_LOOKUP:
				if (block_disk->state == ST_INV) /* invalid block; no match */
					break;
				/* check if blocks' prog/proc ids match */
				if(!((block_RPC->nfsprogid==block_disk->nfsprogid) &&
						(block_RPC->nfsprocid==block_disk->nfsprocid)))
					break;
				/* check if file names match */
				file_name_match = strcmp((char *)&block_RPC->contents.lookup_v3.name,
							 (char *)&block_disk->contents.lookup_v3.name);
				if (file_name_match != 0)  /* no file name match */
					break;
				/* check if file handles match */
				file_handle_match =
						memcmp(block_RPC->contents.lookup_v3.dir.handle,
						block_disk->contents.lookup_v3.dir.handle,
						block_RPC->contents.lookup_v3.dir.length);
				overall_match = (file_handle_match==0? 1 : 0);
				if(!overall_match) break;

				/* invalidate block in disk if required */
				if(invalidate > 0) {
					/* keep current seek point in seekcur; will need it after
						 fd is reopened */
					seekcur = lseek(fd, 0, SEEK_CUR);
					/* release shared lock  */
					if ((flock(fd,LOCK_UN))==-1) {
						Dprintf(D_CACHE,
							"tag_match: could not unlock cache bank %s for reading\n",
							scratch_path);
						return -1;
					}
					/* reopen bank file in read/write mode  for invalidation */
					close(fd);
					if((fd = open(scratch_path, O_RDWR)) == -1) {
						Dprintf(D_CACHE,"tag_match: could not open cache bank %s R/W\n",
							scratch_path);
						return -1;
					}
					/* obtain exclusive lock for invalidation */
					if ((flock(fd,LOCK_EX))==-1) {
						Dprintf(D_CACHE,
							"tag_match: could not lock cache bank %s for writing\n",
							scratch_path);
						return -1;
					}
					/* seek back one block from seekcur */
					seekset = seekcur-sizeof(*block_RPC); /* point back one block */
					seekr = lseek(fd, seekset, SEEK_SET);
					if(seekr!=seekset) {
						Dprintf(D_CACHE, "tag_match: could not seek %i in file bank %s\n",
							seekset, scratch_path);
						flock(fd,LOCK_UN);
						close(fd);
						return -1;
					}
					/* write an invalid block (all zeroes) */
					if(invalidate == 1) {
						/* write an invalid block (all zeroes) */
						bzero(block_disk,sizeof(*block_disk));
					} else {
						memcpy(block_disk, block_RPC, sizeof(*block_disk));
						block_disk->state = ST_RO_CLEAN;  /* Mark this block as valid */
					}
					if((write(fd,block_disk,sizeof(*block_disk)))!=sizeof(*block_disk)) {
						Dprintf(D_CACHE,"tag_match: could not invalidate block in %s\n",
							scratch_path);
						flock(fd,LOCK_UN);
						close(fd);
						return -1;
					}
				}
				break;

			default:
				return -1;
			} /* end switch(block_RPC->nfsprocid) */
			break;
		}  /* end switch(block_RPC->nfsprogid) */
	}

	/* release shared lock  */
	if ((flock(fd,LOCK_UN))==-1) {
		Dprintf(D_CACHE,
			"tag_match: could not unlock cache bank %s for reading\n",
			scratch_path);
		return -1;
	}
	close(fd);

	Dprintf(D_CACHE,"tag_match(%i,%i): returns %i (frame %i, state %d)\n",
		    block_RPC->nfsprogid, block_RPC->nfsprocid,
		    overall_match, framescan-1, block_disk->state);
	return overall_match;
}

/*************************************************************************
 * Finds if there is a frame with specified state for a cache block      *
 * creates new file for cache bank if necessary                          *
 * return values:                                                        *
 * 0 =< x < cache_assoc        : free frame found in x                   *
 * cache_assoc:                : no free frames found                    *
 * -1: error                                                             *
 *************************************************************************/
int find_replace_frame(unsigned long int set_number, cacheblock *block_disk,
					 int state) {
	int fd,reads,framescan;
	int overall_match = 0;
	int seekset, seekr;

	/* open & seek into 1st block of appropriate set in cache file */
	if((fd = open(scratch_path, O_RDONLY)) == -1) {
		Dprintf(D_CACHE,
			"find_replace_frame: could not open cache bank %s for reading\n",
			scratch_path);
		return -1;
	}

	/* obtain a shared lock for cache lookup */
	if ((flock(fd,LOCK_SH))==-1) {
		Dprintf(D_CACHE,
			"find_replace_frame: could not lock cache bank %s for reading\n",
			scratch_path);
		return -1;
	}

	seekset = (set_number%sets_per_bank)*cache_assoc*sizeof(cacheblock);
	Dprintf(D_CACHE,
		"find_replace_frame: seeking %lu -> %lu\n",set_number,seekset);
	seekr = lseek(fd,seekset,SEEK_SET);
	if(seekr != seekset) {
		Dprintf(D_CACHE,
			"find_replace_frame: could not seek initial block %i in %s\n",
			seekset, scratch_path);
		flock(fd,LOCK_UN);
		close(fd);
		return -1;
	}

	/* associative lookup; traverse all sets */
	for(framescan=0; framescan<cache_assoc && !overall_match; framescan++) {

		reads = read(fd,block_disk,sizeof(*block_disk));
		if(reads != sizeof(*block_disk)) { /* read was not right */
			Dprintf(D_CACHE,
				"find_replace_frame: could not read scan block %i in %s %d\n",
				framescan, scratch_path, fd);
			flock(fd,LOCK_UN);
			close(fd);
			return -1;
		}

		if(block_disk->state == state) {
			/* release shared lock  */
			if ((flock(fd,LOCK_UN))==-1) {
				Dprintf(D_CACHE,
					"find_replace_frame: could not unlock cache bank %s\n",
					scratch_path);
				return -1;
			}
			close(fd);
			Dprintf(D_CACHE,
				"find_replace_frame: found %i (state %d)\n",
				framescan, block_disk->state);
			return framescan;
		}
	}

	/* release shared lock  */
	if ((flock(fd,LOCK_UN))==-1) {
		Dprintf(D_CACHE,
			"find_replace_frame: could not unlock cache bank %s\n",
			scratch_path);
		return -1;
	}
	close(fd);
	return cache_assoc;
}


/*************************************************************************
 * Insert result from RPC call in disk cache                             *
 * 0 : inserted                                                          *
 *-1 : error                                                             *
 *************************************************************************/
int insert_in_cache(cacheblock *block_RPC, cacheblock *block_disk) {
	int freeset, replaceset, foundset, err;
	unsigned long int set_number;
	int staterr, fd;
	int written, seekset, seekr, seekcur;
	char c = '\0';

	if(block_RPC==NULL)
		return -1;

	/* determine if cache bank exists; if not, create it */
	set_number = block_file_path_and_set(block_RPC, scratch_path);
	staterr = stat(scratch_path, statbuf);

	if(staterr!=0) { /* create new bank */
		if((fd = creat(scratch_path, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP ))==-1) {
			Dprintf(D_CACHE,"insert_in_cache: could not create %s for R/W\n",
				scratch_path);
			return -1;
		}
		lseek(fd, sets_per_bank*cache_assoc*sizeof(*block_disk)-1, SEEK_SET);
		if(write(fd, &c, 1) != 1) {
			Dprintf(D_CACHE,"insert_in_cache: could not write to %s\n",
				scratch_path);
			close(fd);
			return -1;
		}
		close(fd);
		Dprintf(D_CACHE,"insert_in_cache: bank %s created\n", scratch_path);
	} else { /* bank file exists; check permissions */
		if(!((statbuf->st_mode & S_IRUSR) && (statbuf->st_mode & S_IWUSR))) {
			Dprintf(D_CACHE,"insert_in_cache: wrong permissions in %s\n",
				scratch_path);
			return -1;
		}
	}

	/* look for empty block to replace with */
	if((freeset = find_replace_frame(set_number, block_disk, ST_INV))==-1) {
		Dprintf(D_CACHE,"insert_in_cache: error finding replace frame ST_INV\n");
		return -1;
	}
	if(freeset == cache_assoc) {
		Dprintf(D_CACHE, "No empty blocks\n");
		/* look for clean block to replace with */
		if((replaceset = find_replace_frame(set_number, block_disk,
					ST_RO_CLEAN))==-1) {
			Dprintf(D_CACHE,
				"insert_in_cache: error finding replace frame ST_RO_CLEAN\n");
			return -1;
		}
		if(replaceset == cache_assoc) {
			Dprintf(D_CACHE, "No clean blocks\n");
			/* look for dirty block to replace with */
			if((replaceset = find_replace_frame(set_number, block_disk,
										ST_RW_DIRTY)) == -1) {
				Dprintf(D_CACHE,
					"insert_in_cache: error finding replace frame ST_RW_DIRTY\n");
				return -1;
			}
			if(replaceset == cache_assoc) {
				Dprintf(D_CACHE,
				"insert_in_cache: could not find block to evict\n");
				return -1;
			}
		}
		/* need write back if the block to be replaces with is dirty */
		if(block_disk->state == ST_RW_DIRTY) {
			if((err=write_back_block(block_disk)) != 0)
		Dprintf(D_CACHE,"tag_match: write back error(%d)\n", err);
		}
		foundset = replaceset;
		Dprintf(D_CACHE,"insert_in_cache: replaces block %i\n", replaceset);
	} else {
		foundset = freeset;
		Dprintf(D_CACHE,"insert_in_cache: found empty frame %i\n", freeset);
	}

	/* open cache bank */
	if((fd = open(scratch_path, O_RDWR))==-1) {
		Dprintf(D_CACHE, "insert_in_cache: could not open file bank %s\n",
			scratch_path);
		return -1;
	}

	/* obtain exclusive lock for insertion */
	if ((flock(fd,LOCK_EX))==-1) {
		Dprintf(D_CACHE,
			"insert_in_cache: could not lock cache bank %s for writing\n",
			scratch_path);
		return -1;
	}

	/* seek into appropriate insert position */
	/* first, see into first block of set */
	seekset = (set_number%sets_per_bank)*cache_assoc*sizeof(*block_RPC);
	seekr = lseek(fd,seekset,SEEK_SET);
	if(seekr != seekset) {
		Dprintf(D_CACHE,
			"insert in cache: could not seek initial block %i in %s\n",
			seekset, scratch_path);
		flock(fd,LOCK_UN);
		close(fd);
		return -1;
	}

	/* then, seek within set for replaced frame */
	seekcur = seekr; /* remember current seek point */
	seekset = foundset*sizeof(*block_RPC); /* determine offset within set */
	seekr = lseek(fd, seekset, SEEK_CUR); /* offset seek */
	if((seekcur+seekset) != seekr) {
		Dprintf(D_CACHE, "insert_in_cache: could not seek %i in file bank %s\n",
			seekset, scratch_path);
		flock(fd,LOCK_UN);
		close(fd);
		return -1;
	}

	/* determine cache block state (RO, RW) */
	switch(block_RPC->nfsprogid) {
	case 2:
		switch(block_RPC->nfsprocid) {
		case NFSPROC_LOOKUP:
		case NFSPROC_GETATTR:
			block_RPC->state = ST_RO_CLEAN;
			break;
		default:
			block_RPC->state = ST_INV;
		}
		break; /* case 2 */
	case 3:
		switch(block_RPC->nfsprocid) {
		case NFSPROC3_GETATTR:
		case NFSPROC3_LOOKUP:
		case NFSPROC3_ACCESS:
			block_RPC->state = ST_RO_CLEAN;
			break;
		default:
			block_RPC->state = ST_INV;
		}
		break;
	default:
		block_RPC->state = ST_INV;
	}

	/* write block into cache */
	written = write(fd, block_RPC, sizeof(*block_RPC));
	if(written != sizeof(*block_RPC)) {
		Dprintf(L_ERROR,"Could not write into cache!\n");
		flock(fd,LOCK_UN);
		close(fd);
		return -1;
	}
	/* release exclusive lock  */
	if ((flock(fd,LOCK_UN))==-1) {
		Dprintf(D_CACHE,
			"insert_in_cache: could not unlock cache bank %s for writing\n",
			scratch_path);
		return -1;
	}
	close(fd);
	return 0;
}

/*************************************************************************
 * Initialize global variables                                           *
 * path    : base path to file cache directory                           *
 * return values:                                                        *
 * 0 : successfull                                                       *
 * -1 : unsuccessful                                                     *
 *************************************************************************/
int init_attr_cache_global(char *basepath, char *sessionid) {
	int staterr, mkdirerr;

	if (!(scratch_path = (char *)malloc(sizeof(*scratch_path)*MAX_STR)))
		return -1;

	if (!(SessionID = (char *)malloc(sizeof(*SessionID)*MAX_STR)))
		return -1;
	if(strlen(sessionid)>(sizeof(*SessionID)*MAX_STR))
		return -1;
	strcpy(SessionID,sessionid);

	if (!(BaseCachePath = (char *)malloc(sizeof(*BaseCachePath)*MAX_STR)))
		return -1;
	if(strlen(basepath)>(sizeof(*BaseCachePath)*MAX_STR))
		return -1;
	strcpy(BaseCachePath,basepath);

	if (!(statbuf = (struct stat *) malloc (sizeof(*statbuf))))
		return -1;

	if (!(block_scratch_RPC = (cacheblock *)malloc(sizeof(*block_scratch_RPC))))
		return -1;

	if (!(block_scratch_disk =(cacheblock *)malloc(sizeof(*block_scratch_disk))))
		return -1;

	/* check the existence of base path and session ID directories */
	staterr = stat(BaseCachePath, statbuf);
	if(!((staterr==0) && (statbuf->st_mode & S_IRUSR) &&
			 (statbuf->st_mode & S_IWUSR))) {
		Dprintf(L_ERROR, "BaseCachePath directory %s does not exist\n",
			BaseCachePath);
		return -1;
	}
	if((strlen(BaseCachePath)+1+strlen(SessionID))>MAX_STR) {
		Dprintf(L_ERROR,"cache block path too long\n");
		return -1;
	}
	sprintf(scratch_path,"%s/%s",BaseCachePath,SessionID);
	staterr = stat(scratch_path, statbuf);
	if(!((staterr==0) && (statbuf->st_mode & S_IRUSR) &&
			 (statbuf->st_mode & S_IWUSR))) {
		mkdirerr = mkdir(scratch_path,S_IRWXU | S_IRWXG);
		if(mkdirerr==-1) {
			Dprintf(L_ERROR, "Could not create SessionID directory %s\n",
				scratch_path);
			return -1;
		}
	}
	return 0;
}

void dump_attr_cache_parameters(FILE *fp) {
	fprintf(fp,"Attribute cache parameters:\n");
	fprintf(fp,"\tCache size: %lu blocks\n",cache_size);
	fprintf(fp,"\tCache assoc: %du\n",cache_assoc);
	fprintf(fp,"\t# cache banks: %du\n",cache_banks);
	fprintf(fp,"\t# cache sets: %lu\n",cache_sets);
	fprintf(fp,"\t# sets per bank: %lu\n",sets_per_bank);
	fprintf(fp,"\tSize of cache block: %i\n",sizeof(*block_scratch_RPC));
	fprintf(fp,"\tSize of cache bank file: %lu\n",
		sets_per_bank*cache_assoc*sizeof(*block_scratch_RPC));
}

/*************************************************************************
 * Initialize cache parameters                                           *
 * size    : cache size (# entries, pwr of 2)                            *
 * assoc   : cache associativity (pwr of 2)                              *
 * return values:                                                        *
 * 0 : successfull                                                       *
 *-1 : unsuccessful                                                      *
 *************************************************************************/
int init_attr_cache(unsigned long int size, int assoc, int banks,
				char *basepath, char *sessionid) {

	unsigned long int localsize, localassoc, localbanks;

	if(basepath==NULL || sessionid==NULL) {
		Dprintf(L_ERROR, "basepath/sessionid error\n");
		return -1;
	}

	if(size==0 || size > MAX_CACHE_SIZE)
		return -1;
	if (assoc==0 || assoc > MAX_CACHE_ASSOC)
		return -1;

	if(init_attr_cache_global(basepath,sessionid) == -1)
		return -1;

	localsize=size;
	while (localsize>1) {
		if(localsize%2)
			return -1;
		localsize>>=1;
	}

	localassoc=assoc;
	while (localassoc>1) {
		if(localassoc%2)
			return -1;
		localassoc>>=1;
	}

	localbanks=banks;
	while (localbanks>1) {
		if(localbanks%2)
			return -1;
		localbanks>>=1;
	}

	cache_size = size;
	cache_assoc = assoc;
	cache_banks = banks;
	cache_sets = cache_size/cache_assoc;
	sets_per_bank = cache_size/(cache_assoc*cache_banks);
	Dprintf(L_NOTICE, "attr cache init successful\n"
										"size: %d, assoc: %d, banks: %d "
				"sets: %d, sets per bank: %d\n",
				cache_size, cache_assoc, cache_banks,
				cache_sets, sets_per_bank);

	return 0;
}

/*************************************************************************
 * lookup block in cache                                                 *
 * prog_index, proc_index: type of RPC call                              *
 * argument = (void *)rpcargs from incoming RPC call                     *
 * return values:                                                        *
 * 0 : miss                                                              *
 * 1 : hit                                                               *
 * -1: error                                                             *
 * side-effects:                                                         *
 *  block_scratch_RPC contains copy of diropargs                         *
 *  block_scratch_disk contains copy of results in case of a hit         *
 *  res contains RPC result looked up from cache                         *
 *************************************************************************/
int cache_lookup(int prog_index, int proc_index, void *rpcargs, void *res) {
	int match=0;

	switch(prog_index) {
	case 2:
		switch(proc_index) {
		case NFSPROC_LOOKUP:
			/* create cache_scratch_RPC from rpcargs */
			if((argument_lookup_v2((diropargs *)rpcargs, block_scratch_RPC))==-1) {
				Dprintf(D_CACHE,"cache_lookup: argument copy failed\n");
				return -1;
			}
			if((match=tag_match_and_inv(block_scratch_RPC,
					block_scratch_disk, 0))==-1) {
				Dprintf(D_CACHE,"cache_lookup: tag match failed\n");
				return -1;
			}
			if (match == 1) { /* hit; copy data */
				if((response_from_disk_cache_lookup_v2(block_scratch_disk,
											 (diropres *)res))==-1) {
					Dprintf(D_CACHE,"cache_lookup: response copy failed\n");
					return -1;
				}
			}
			break;
		case NFSPROC_GETATTR:
			/* create cache_scratch_RPC from rpcargs */
			if((argument_getattr_v2((nfs_fh *)rpcargs, block_scratch_RPC))==-1) {
				Dprintf(D_CACHE,"cache_lookup: argument copy failed\n");
				return -1;
			}
			if((match=tag_match_and_inv(block_scratch_RPC,
					block_scratch_disk, 0))==-1) {
				Dprintf(D_CACHE,"cache_lookup: tag match failed\n");
				return -1;
			}
			if (match == 1) { /* hit; copy data */
				if((response_from_disk_cache_getattr_v2(block_scratch_disk,
											 (attrstat *)res))==-1) {
					Dprintf(D_CACHE,"cache_lookup: response copy failed\n");
					return -1;
				}
			}
			break;
		default:
			break;
		}
	break; /* case 2 */
	case 3:
		switch(proc_index) {
		case NFSPROC3_GETATTR:
			/* create cache_scratch_RPC from rpcargs */
			if((argument_getattr_v3((nfs_fh3 *)rpcargs, block_scratch_RPC))==-1) {
				Dprintf(D_CACHE,"cache_lookup: argument copy failed\n");
				return -1;
			}
			if((match=tag_match_and_inv(block_scratch_RPC,
					block_scratch_disk, 0))==-1) {
				Dprintf(D_CACHE,"cache_lookup: tag match failed\n");
				return -1;
			}
			if (match == 1) { /* hit; copy data */
				if((response_from_disk_cache_getattr_v3(block_scratch_disk,
											 (getattr3res *)res))==-1) {
					Dprintf(D_CACHE,"cache_lookup: response copy failed\n");
					return -1;
				}
			}
			break;
		case NFSPROC3_LOOKUP:
			/* create cache_scratch_RPC from rpcargs */
			if((argument_lookup_v3((diropargs3 *)rpcargs, block_scratch_RPC))==-1) {
				Dprintf(D_CACHE,"cache_lookup: argument copy failed\n");
				return -1;
			}
			if((match=tag_match_and_inv(block_scratch_RPC,
					block_scratch_disk, 0))==-1) {
				Dprintf(D_CACHE,"cache_lookup: tag match failed\n");
				return -1;
			}
			if (match == 1) { /* hit; copy data */
				if((response_from_disk_cache_lookup_v3(block_scratch_disk,
											 (lookup3res *)res))==-1) {
					Dprintf(D_CACHE,"cache_lookup: response copy failed\n");
					return -1;
				}
			}
			break;
		case NFSPROC3_ACCESS:
			/* create cache_scratch_RPC from rpcargs */
			if((argument_access_v3((nfs_fh3 *)rpcargs, block_scratch_RPC))==-1) {
				Dprintf(D_CACHE,"cache_lookup: argument copy failed\n");
				return -1;
			}
			if((match=tag_match_and_inv(block_scratch_RPC,
					block_scratch_disk, 0))==-1) {
				Dprintf(D_CACHE,"cache_lookup: tag match failed\n");
				return -1;
			}
			if (match == 1) { /* hit; copy data */
				if((response_from_disk_cache_access_v3(block_scratch_disk,
											 (access3res *)res))==-1) {
					Dprintf(D_CACHE,"cache_lookup: response copy failed\n");
					return -1;
				}
			}
			break;
		} /* switch(proc_index) */
		break;
	} /* switch(prog_index) */

	Dprintf(D_CACHE, "acache lookup: (%i,%i) = %s\n", prog_index, proc_index,
		    match ? "HIT" : "MISS");
	return match;
}

/*************************************************************************
 * insert block in cache                                                 *
 * prog_index, proc_index: type of RPC call                              *
 * argument = (void *)rpcargs from incoming RPC call                     *
 * return values:                                                        *
 * 0 : inserted                                                          *
 *-1 : did not insert                                                    *
 * side-effects:                                                         *
 *  block_scratch_RPC contains copy of results                           *
 *  block_scratch_disk contains copy of results in case of a hit         *
 *************************************************************************/
int cache_insert(int prog_index, int proc_index, void *rpcargs, void *res) {
	int ret = -1;

	switch(prog_index) {
	case 2:
		switch(proc_index) {
		case NFSPROC_LOOKUP:
			if(rpcargs != NULL)
				if((argument_lookup_v2((diropargs *)rpcargs, block_scratch_RPC))==-1) {
					Dprintf(D_CACHE,"cache_lookup: argument copy failed\n");
					break;
			}
			/* create cache_scratch_RPC from rpcargs */
			if((response_remote_RPC_lookup_v2((diropres *)res,
					block_scratch_RPC))==-1) {
				Dprintf(D_CACHE,"cache_insert: RPC response copy failed\n");
				break;
			}
			if((insert_in_cache(block_scratch_RPC, block_scratch_disk))==-1) {
				Dprintf(D_CACHE,"cache_insert: insertion failed\n");
				break;
			}
			ret = 0;
			break;
		case NFSPROC_GETATTR:
			if(rpcargs != NULL)
				if((argument_getattr_v2((nfs_fh *)rpcargs, block_scratch_RPC))==-1) {
					Dprintf(D_CACHE,"cache_lookup: argument copy failed\n");
					break;
			}
			/* create cache_scratch_RPC from rpcargs */
			if((response_remote_RPC_getattr_v2((attrstat *)res,
					block_scratch_RPC))==-1) {
				Dprintf(D_CACHE,"cache_insert: RPC response copy failed\n");
				break;
			}
			if((insert_in_cache(block_scratch_RPC, block_scratch_disk))==-1) {
				Dprintf(D_CACHE,"cache_insert: insertion failed\n");
				break;
			}
			ret = 0;
			break;
		default:
			return -1;
		} /* switch(proc_index) */
		break; /* case 2 */
	case 3:
		switch(proc_index) {
		case NFSPROC3_GETATTR:
			if(rpcargs != NULL)
				if((argument_getattr_v3((nfs_fh3 *)rpcargs, block_scratch_RPC))==-1) {
					Dprintf(D_CACHE,"cache_lookup: argument copy failed\n");
					break;
			}
			/* create cache_scratch_RPC from rpcargs */
			if((response_remote_RPC_getattr_v3((getattr3res *)res,
					block_scratch_RPC))==-1) {
				Dprintf(D_CACHE,"cache_insert: RPC response copy failed\n");
				break;
			}
			if((insert_in_cache(block_scratch_RPC, block_scratch_disk))==-1) {
				Dprintf(D_CACHE,"cache_insert: insertion failed\n");
				break;
			}
			ret = 0;
			break;
		case NFSPROC3_LOOKUP:
			if(rpcargs != NULL)
				if((argument_lookup_v3((diropargs3 *)rpcargs, block_scratch_RPC))==-1) {
					Dprintf(D_CACHE,"cache_lookup: argument copy failed\n");
					break;
			}
			/* create cache_scratch_RPC from rpcargs */
			if((response_remote_RPC_lookup_v3((lookup3res *)res,
					block_scratch_RPC))==-1) {
				Dprintf(D_CACHE,"cache_insert: RPC response copy failed\n");
				break;
			}
			if((insert_in_cache(block_scratch_RPC, block_scratch_disk))==-1) {
				Dprintf(D_CACHE,"cache_insert: insertion failed\n");
				break;
			}
			ret = 0;
			break;
		case NFSPROC3_ACCESS:
			/* create cache_scratch_RPC from rpcargs */
			block_scratch_RPC->nfsprocid = NFSPROC3_ACCESS;
			if((response_remote_RPC_access_v3((access3res *)res,
					block_scratch_RPC))==-1) {
				Dprintf(D_CACHE,"cache_insert: RPC response copy failed\n");
				break;
			}
			if((insert_in_cache(block_scratch_RPC, block_scratch_disk))==-1) {
				Dprintf(D_CACHE,"cache_insert: insertion failed\n");
				break;
			}
			ret = 0;
			break;
		default:
			break;
		} /* switch(proc_index) */
		break;
	default:
		break;
	} /* switch(prog_index) */

	Dprintf(D_CACHE, "acache insert: (%i,%i) = %s\n", prog_index, proc_index,
		    ret == 0 ? "SUCCESS" : "FAILED");
	return ret;
}

int cache_update(int prog_index, int proc_index, void *rpcargs, void *res) {
	int match=0;

	switch(prog_index) {
	case 2:
		switch(proc_index) {
		case NFSPROC_GETATTR:
			/* create cache_scratch_RPC from rpcargs */
			if((argument_getattr_v2((nfs_fh *)rpcargs, block_scratch_RPC))==-1) {
				Dprintf(D_CACHE,"cache_update: argument copy failed\n");
				return -1;
			}
			if((response_remote_RPC_getattr_v2((attrstat *)res,
					block_scratch_RPC))==-1) {
				Dprintf(D_CACHE,"cache_update: RPC response copy failed\n");
				return -1;
			}
			if((match=tag_match_and_inv(block_scratch_RPC,
					block_scratch_disk, 2))==-1) {
				Dprintf(D_CACHE,"cache_update: tag match failed\n");
				return -1;
			}
			break;
		case NFSPROC_LOOKUP:
			/* create cache_scratch_RPC from rpcargs */
			if((argument_lookup_v2((diropargs *)rpcargs, block_scratch_RPC))==-1) {
				Dprintf(D_CACHE,"cache_update: argument copy failed\n");
				return -1;
			}
			if((response_remote_RPC_lookup_v2((diropres *)res,
					block_scratch_RPC))==-1) {
				Dprintf(D_CACHE,"cache_update: RPC response copy failed\n");
				return -1;
			}
			if((match=tag_match_and_inv(block_scratch_RPC,
					block_scratch_disk, 2))==-1) {
				Dprintf(D_CACHE,"cache_update: tag match failed\n");
				return -1;
			}
			break;
		default:
			return -1;
		}
		break;

	case 3:
		switch(proc_index) {
		case NFSPROC3_GETATTR:
			/* create cache_scratch_RPC from rpcargs */
			if((argument_getattr_v3((nfs_fh3 *)rpcargs, block_scratch_RPC))==-1) {
				Dprintf(D_CACHE,"cache_update: argument copy failed\n");
				return -1;
			}
			if((response_remote_RPC_getattr_v3((getattr3res *)res,
					block_scratch_RPC))==-1) {
				Dprintf(D_CACHE,"cache_update: RPC response copy failed\n");
				return -1;
			}
			if((match=tag_match_and_inv(block_scratch_RPC,
					block_scratch_disk, 2))==-1) {
				Dprintf(D_CACHE,"cache_update: tag match failed\n");
				return -1;
			}
			break;
		case NFSPROC3_ACCESS:
			if((argument_access_v3((nfs_fh3 *)rpcargs, block_scratch_RPC))==-1) {
				Dprintf(D_CACHE,"cache_update: argument copy failed\n");
				return -1;
			}
			if((response_remote_RPC_access_v3((access3res *)res,
					block_scratch_RPC))==-1) {
				Dprintf(D_CACHE,"cache_update: RPC response copy failed\n");
				return -1;
			}
			if((match=tag_match_and_inv(block_scratch_RPC,
					block_scratch_disk, 2))==-1) {
				Dprintf(D_CACHE,"cache_update: tag match failed\n");
				return -1;
			}
			break;
		case NFSPROC3_LOOKUP:
			/* create cache_scratch_RPC from rpcargs */
			if((argument_lookup_v3((diropargs3 *)rpcargs, block_scratch_RPC))==-1) {
				Dprintf(D_CACHE,"cache_update: argument copy failed\n");
				return -1;
			}
			if((response_remote_RPC_lookup_v3((lookup3res *)res,
					block_scratch_RPC))==-1) {
				Dprintf(D_CACHE,"cache_update: RPC response copy failed\n");
				return -1;
			}
			if((match=tag_match_and_inv(block_scratch_RPC,
					block_scratch_disk, 2))==-1) {
				Dprintf(D_CACHE,"cache_update: tag match failed\n");
				return -1;
			}
			break;
		default:
			return -1;
		} /* switch(proc_index) */
		break;
	} /* switch(prog_index) */

	Dprintf(D_CACHE, "acache update: (%i,%i) = %s\n", prog_index, proc_index,
		    match ? "HIT" : "MISS");
	return match;
}

/*************************************************************************
 * Lookup entry for diroparg *arg in attribute cache.
 * if found, invalidate it
 * return values:                                                        *
 * 0 : no invalidation needed                                            *
 * 1 : successful invalidation                                           *
 * -1: error                                                             *
 *************************************************************************/
int invalidate_diroparg_v2(diropargs *arg) {
	int match;

	if(arg==NULL) {
		Dprintf(D_CACHE,"invalidate_diroparg_v2: arg==NULL\n");
		return -1;
	}

	/* copy diropargs to block_scratch_RPC */
	if((argument_lookup_v2(arg, block_scratch_RPC))==-1) {
		Dprintf(D_CACHE,"invalidate_diroparg_v2: argument copy failed\n");
		return -1;
	}

	if((match=tag_match_and_inv(block_scratch_RPC,
						block_scratch_disk, 1))==-1) {
		Dprintf(D_CACHE,"invalidate_diroparg_v2: tag invalidate failed\n");
		return -1;
	}
	Dprintf(D_CACHE,"invalidate_diroparg_v2: %s\n", match ? "CLEARED" : "NOT FOUND");
	return match;
}

int invalidate_diroparg_v3(diropargs3 *arg) {
	int match;

	if(arg==NULL) {
		Dprintf(D_CACHE,"invalidate_diroparg_v3: arg==NULL\n");
		return -1;
	}

	/* copy diropargs to block_scratch_RPC */
	if((argument_lookup_v3(arg, block_scratch_RPC))==-1) {
		Dprintf(D_CACHE,"invalidate_diroparg_v3: argument copy failed\n");
		return -1;
	}

	if((match=tag_match_and_inv(block_scratch_RPC,
						block_scratch_disk, 1))==-1) {
		Dprintf(D_CACHE,"invalidate_diroparg_v3: tag invalidate failed\n");
		return -1;
	}
	Dprintf(D_CACHE,"invalidate_diroparg_v3: %s\n", match ? "CLEARED" : "NOT FOUND");
	return match;
}

/*************************************************************************
 * Lookup entry for nfs_fh *arg in attribute cache.                      *
 * if found, invalidate it
 * return values:                                                        *
 * 0 : no invalidation needed                                            *
 * 1 : successful invalidation                                           *
 * -1: error                                                             *
 *************************************************************************/
int invalidate_nfs_fh_v2(nfs_fh *arg) {
	int match;

	if(arg==NULL) {
		Dprintf(D_CACHE,"invalidate_nfs_fh_v2: arg==NULL\n");
		return -1;
	}

	/* copy diropargs to block_scratch_RPC */
	if((argument_getattr_v2(arg, block_scratch_RPC))==-1) {
		Dprintf(D_CACHE,"invalidate_nfs_fh_v2: argument copy failed\n");
		return -1;
	}

	if((match=tag_match_and_inv(block_scratch_RPC,
						block_scratch_disk, 1))==-1) {
		Dprintf(D_CACHE,"invalidate_nfs_fh_v2: tag invalidate failed\n");
		return -1;
	}
	Dprintf(D_CACHE,"invalidate_nfs_fh_v2: %s\n", match ? "CLEARED" : "NOT FOUND");
	return match;
}

int invalidate_nfs_fh_v3(nfs_fh3 *arg) {
	int match;

	if(arg==NULL) {
		Dprintf(D_CACHE,"invalidate_nfs_fh_v3: arg==NULL\n");
		return -1;
	}

	/* copy diropargs to block_scratch_RPC */
	if((argument_getattr_v3(arg, block_scratch_RPC))==-1) {
		Dprintf(D_CACHE,"invalidate_nfs_fh_v3: argument copy failed\n");
		return -1;
	}

	if((match=tag_match_and_inv(block_scratch_RPC,
						block_scratch_disk, 1))==-1) {
		Dprintf(D_CACHE,"invalidate_nfs_fh_v3: tag invalidate failed\n");
		return -1;
	}
	Dprintf(D_CACHE,"invalidate_nfs_fh_v3: %s\n", match ? "CLEARED" : "NOT FOUND");
	return match;
}

/*************************************************************************
 * check if need to invalidate block in cache                            *
 * prog_index, proc_index: type of RPC call                              *
 * argument = (void *)rpcargs from incoming RPC call                     *
 * return values:                                                        *
 * 0 : no invalidation needed                                            *
 * 1 : successful invalidation                                           *
 * -1: error                                                             *
 *                                                                       *
 * The following invalidations are performed:                            *
 *  NFS-V2:                                                              *
 *    CREATE, MKDIR: invalidate (diropargs *)rpcargs->where d            *
 *    REMOVE, RMDIR: invalidate (diropargs *)rpcargs                     *
 *    RENAME : invalidate (diropargs *)rpcargs->from, rpcargs->to        *
 *    SYMLINK: invalidate (diropargs *)rpcargs->from                     *
 *    LINK:    invalidate (diropargs *)rpcargs->to                       *
 *    SETATTR: invalidate (sattrstat *)rpcargs->file                     *
 *    WRITE:   invalidate (sattrstat *)rpcargs->file
 * side-effects:                                                         *
 * block_scratch_RPC contains copy of rpcargs                            *
 *************************************************************************/
int cache_inval(int prog_index, int proc_index, void *rpcargs) {
	createargs *crargs;
	renameargs *renargs;
	symlinkargs *symargs;
	linkargs *largs;
	sattrargs *sargs;
	writeargs *wargs;
	diropargs  *invalidate_arg1,*invalidate_arg2;
	int invres1=0, invres2=0;

	switch(prog_index) {
	case 2: /* NFS-V2 */
		/* first, set up invalidate_arg1 and invalidate_arg2 */
		/* most calls will only require one invalidation; in these
			 cases, arg2 will be set to NULL. The exception is RENAME */
		switch(proc_index) {
		case NFSPROC_CREATE:
		case NFSPROC_MKDIR:
			crargs = (createargs *)rpcargs;
			invalidate_arg1 = (diropargs *)&crargs->where;
			invalidate_arg2 = NULL;
			invres1 = invalidate_nfs_fh_v2(&invalidate_arg1->dir);
			if(invres1==-1) {
				Dprintf(D_CACHE,"Error invalidating attribute cache block\n");
				return -1;
			}
			Dprintf(D_CACHE,"cache_inval: invalidate fh %i\n", invres1);
			break;
		case NFSPROC_REMOVE:
		case NFSPROC_RMDIR:
			invalidate_arg1 = (diropargs *)rpcargs;
			invalidate_arg2 = NULL;
			invres1 = invalidate_nfs_fh_v2((nfs_fh *)&invalidate_arg1->dir);
			if(invres1==-1) {
				Dprintf(D_CACHE,"Error invalidating attribute cache block\n");
				return -1;
			}
			Dprintf(D_CACHE,"cache_inval: invalidate fh %i\n", invres1);
			break;
		case NFSPROC_RENAME:
			renargs = (renameargs *)rpcargs;
			invalidate_arg1 = (diropargs *)&renargs->from;
			invalidate_arg2 = (diropargs *)&renargs->to;
			invres1 = invalidate_nfs_fh_v2((nfs_fh *)&invalidate_arg1->dir);
			if(invres1==-1) {
				Dprintf(D_CACHE,"Error invalidating attribute cache block\n");
				return -1;
			}
			Dprintf(D_CACHE,"cache_inval: invalidate fh %i\n", invres1);
			if(memcmp(&invalidate_arg1->dir,&invalidate_arg2->dir,NFS_FHSIZE) != 0) {
				invres2 = invalidate_nfs_fh_v2((nfs_fh *)&invalidate_arg2->dir);
				if(invres2==-1) {
					Dprintf(D_CACHE,"Error invalidating attribute cache block\n");
					return -1;
				}
				Dprintf(D_CACHE,"cache_inval: invalidate fh %i\n", invres2);
			}
			break;
		case NFSPROC_SYMLINK:
			symargs = (symlinkargs *)rpcargs;
			invalidate_arg1 = (diropargs *)&symargs->from;
			invalidate_arg2 = NULL;
			invres1 = invalidate_nfs_fh_v2((nfs_fh *)&invalidate_arg1->dir);
			if(invres1==-1) {
				Dprintf(D_CACHE,"Error invalidating attribute cache block\n");
				return -1;
			}
			Dprintf(D_CACHE,"cache_inval: invalidate fh %i\n", invres1);
			break;
		case NFSPROC_LINK:
			largs = (linkargs *)rpcargs;
			invalidate_arg1 = (diropargs *)&largs->to;
			invalidate_arg2 = NULL;
			invres1 = invalidate_nfs_fh_v2((nfs_fh *)&largs->from);
			if(invres1==-1) {
				Dprintf(D_CACHE,"Error invalidating attribute cache block\n");
				return -1;
			}
			invres1 = invalidate_nfs_fh_v2((nfs_fh *)&invalidate_arg1->dir);
			if(invres1==-1) {
				Dprintf(D_CACHE,"Error invalidating attribute cache block\n");
				return -1;
			}
			Dprintf(D_CACHE,"cache_inval: invalidate fh %i\n", invres1);
			break;
		case NFSPROC_SETATTR:
			/* SETATTR, WRITE are unlike the above calls; they invalidate files
			   based on the file handle rpcargs->file
			   They are handled separately here */
			sargs = (sattrargs *)rpcargs;
			invres1 = invalidate_nfs_fh_v2((nfs_fh *)&sargs->file);
			if(invres1==-1) {
				Dprintf(D_CACHE,"Error invalidating attribute cache block\n");
				return -1;
			}
			Dprintf(D_CACHE,"cache_inval: invalidate fh %i\n", invres1);
			return invres1;
			break;
		case NFSPROC_WRITE:
			wargs = (writeargs *)rpcargs;
			invres1 = invalidate_nfs_fh_v2((nfs_fh *)&wargs->file);
			if(invres1==-1) {
				Dprintf(D_CACHE,"Error invalidating attribute cache block\n");
				return -1;
			}
			Dprintf(D_CACHE,"cache_inval: invalidate fh %i\n", invres1);
			return invres1;
			break;
		default:
			/* no invalidations necessary for other calls */
			invalidate_arg1 = NULL;
			invalidate_arg2 = NULL;
		}
		/* now perform actual invalidation */
		if(invalidate_arg1!=NULL) {
			invres1 = invalidate_diroparg_v2(invalidate_arg1);
			if(invres1==-1) {
				Dprintf(D_CACHE,"Error invalidating attribute cache block\n");
				return -1;
			}
			Dprintf(D_CACHE,"cache_inval: result of argument1: %i\n", invres1);
			if(invalidate_arg2!=NULL) {
				invres2 = invalidate_diroparg_v2(invalidate_arg2);
				if(invres2==-1) {
					Dprintf(D_CACHE,"Error invalidating attribute cache block\n");
					return -1;
				}
				Dprintf(D_CACHE,"cache_inval: result of argument2: %i\n", invres2);
			}
			return((invres1+invres2)>0);
		}
		break; /* case 2 */
	case 3: /* NFS-V3 */
		break;
	default:
		return -1; /* unknown NFS version */
	}
	return -1;
}
