/***************************************************************************
 * Copyright (c) 2002-2005, In-VIGO Team
 * Advanced Computing and Information Systems (ACIS) Laboratory
 * University of Florida
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * cache_attrib.h
 *
 * Author: Renato J. Figueiredo, Ming Zhao
 *
 * Data structures for the gvfs proxy attribute cache.
 * The attribute cache is stored in the local disk.
 * The attribute cache is structured as follows:
 *
 * - The smallest allocation unit is a cacheblock
 *   - the cache_block is sufficiently large to hold the largest RPC
 *     reply that can be received by NFS, plus a tag, and state bits
 *   - The number of cache blocks is "cache_size"
 *
 * - Cache blocks are stored in file "banks" in the disk.
 *   - The name of a bank X in the disk is given by:
 *     BaseCachePath/SessionID/attr_cache_bank_X
 *     (where BaseCachePath is common to multiple proxies, and SessionID
 *      is unique to each proxy instance)
 *   - The number of file banks is "cache_banks"
 *
 * - The associativity of the cache is given by "cache_assoc"
 *
 * - Example: a 4-way associative cache holding a max. of 32 attributes,
 *   with two file banks. This cache would consist of files attr_cache_bank_0
 *   and attr_cache_bank_1; each bank would have 4 sets, each set having
 *   4 blocks.
 ***************************************************************************/

#ifndef CACHE_ATTRIB_H
#define CACHE_ATTRIB_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <math.h>
#include "nfs3_prot.h"
/*
 * Constants (sizes)
 * MAX_CACHESIZE : maximum number of blocks in disk cache
 * MAX_CACHE_ASSOC : maximum cache associativity
 * MAX_STR : maximum string size for NFS paths
 */

#define MAX_CACHE_SIZE 1048576
#define MAX_CACHE_ASSOC 1024
#define MAX_CACHE_BANKS 1024
#define MAX_STR NFS_MAXPATHLEN

/* default cache parameters */
#define DEFAULT_ATTR_CACHE_SIZE 4096
#define DEFAULT_ATTR_CACHE_ASSOC 8
#define DEFAULT_ATTR_CACHE_BANKS 64

/*
 * Constants (cache states)
 */
#define ST_INV 0       /* invalid           */
#define ST_RO_CLEAN 1  /* read-only, clean  */
#define ST_RW_DIRTY 2  /* read/write, dirty */

/*
 * Data structures
 */

struct cacheblock_lookup_v2 {
  /* store contents of a NFS-V2 lookup transaction */
  /* storage is static (no mallocs) because an entire block goes to
     disk in a single write operation */

  /* TAG */
  /* NFS diropargs (lookup arguments) */
  nfs_fh   dir;
  char     name[MAX_STR];

  /* CONTENTS */
  /* NFS diropokres (lookup results) */
  nfs_fh file;
  fattr attributes;
  /* NFS diropres (lookup result status) */
  nfsstat status;
};
typedef struct cacheblock_lookup_v2 cacheblock_lookup_v2;

struct cacheblock_getattr_v2 {
  /* TAG */
  /* arguments to NFS getattr call */
  nfs_fh   handle;

  /* CONTENTS */
  /* results from NFS getattr call */
  nfsstat status;
  fattr attributes;
};
typedef struct cacheblock_getattr_v2 cacheblock_getattr_v2;

typedef struct nfs3_fh {
  u_int length;
  char handle[NFS3_FHSIZE];
} nfs3_fh;

struct cacheblock_lookup_v3 {
  /* store contents of a NFS-V3 lookup transaction */
  /* storage is static (no mallocs) because an entire block goes to
     disk in a single write operation */

  /* TAG */
  /* NFS diropargs (lookup arguments) */
  nfs3_fh dir;
  char name[MAX_STR];

  /* CONTENTS */
  /* NFS diropokres (lookup results) */
  nfs3_fh file;
  /* NFS diropres (lookup result status) */
  nfsstat3 status;
};
typedef struct cacheblock_lookup_v3 cacheblock_lookup_v3;

struct cacheblock_getattr_v3 {
  /* TAG */
  /* arguments to NFS getattr call */
  nfs3_fh handle;
  uint32 access;
  /* CONTENTS */
  /* results from NFS getattr call */
  nfsstat3 status;
  fattr3 attributes;
};
typedef struct cacheblock_getattr_v3 cacheblock_getattr_v3;

struct cacheblock {  /* stores a cache entry for NFS transactions */

  int state;               /* block's state */
  int nfsprogid;           /* NFS program version for this block */
  int nfsprocid;           /* NFS procedure number for this block */

  union {
                           /* contents: depend on transaction type */
    cacheblock_lookup_v2 lookup_v2;
    cacheblock_getattr_v2 getattr_v2;
    cacheblock_lookup_v3 lookup_v3;
    cacheblock_getattr_v3 getattr_v3;
  } contents;
};
typedef struct cacheblock cacheblock;

/* The top-level functions of cache_attrib.c */

/* Lookup a block in disk cache, for a given NFS RPC call (prog_index,
   proc_index) and arguments (rpcargs). Returns hit (1) or miss (0); if
   an error occurs, returns -1. In case of a hit, return RPC results
   are stored in *res */
extern int cache_lookup(int prog_index, int proc_index,
			void *rpcargs, void *res);

/* Insert results *res received from a given NFS RPC call (prog_index,
   proc_index) into the disk cache (creating a new cache bank in disk
   if necessary). Returns 0 on success, -1 on error. */
extern int cache_insert(int prog_index, int proc_index, void *rpcargs, void *res);
extern int cache_update(int prog_index, int proc_index, void *rpcargs, void *res);
/* Check and perform necessary invalidations of attrib. cache blocks */
extern int cache_inval(int prog_index, int proc_index, void *rpcargs);
extern int invalidate_diroparg_v2(diropargs *arg);
extern int invalidate_diroparg_v3(diropargs3 *arg);
extern int invalidate_nfs_fh_v3(nfs_fh3 *arg);

/* Initialize attribute cache parameters and base path directory
   Returns 0 if successfull, -1 if an error occurred */
extern int init_attr_cache(unsigned long int size, int assoc,
			   int banks, char *basepath, char *sessionid);
extern int flush_attr_cache(void);

#endif /* CACHE_ATTRIB_H */
