/***************************************************************************
 * Copyright (c) 2002-2005, In-VIGO Team
 * Advanced Computing and Information Systems (ACIS) Laboratory
 * University of Florida
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * cache_data.c
 *
 * Author:   Renato J. Figueiredo, Ming Zhao
 * Updated: Jan., 2004
 *               Implemented write-back with write-allocate for data cache;
 * Updated: Jan. 27, 2004
 *               Corrected the problem that the resources allocated RPC results
 *               are not freed in several places.
 *
 * Functions that implement the gvfs proxy data cache.
 * The data cache is stored in the local disk.
 * Refer to cache_data.h for a description of the data cache
 * organization
 *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <unistd.h>
#include <fcntl.h>
#include <math.h>
#include <time.h>

#include "nfs_prot.h"
#include "logging.h"
#include "cache_data.h"

/*
 * Global variables
 */

unsigned long int dcache_size = MAX_CACHE_SIZE;
unsigned int dcache_assoc = MAX_CACHE_ASSOC;
unsigned int dcache_banks = MAX_CACHE_BANKS;
unsigned long int dcache_sets; /* total number of cache sets */
unsigned long int dsets_per_bank;
char *dBaseCachePath;
char *dSessionID;
char *dscratch_path;
struct stat *statbuf;
dcacheblock *dblock_scratch_RPC;
dcacheblock *dblock_scratch_disk;
dirty_block *db_head, *db_tail; /* head and tail of dirty block list*/
fh_node *stale_fh_hash[FH_HASH_SIZE]; /* stale file handle hash*/

/* indices used in search for a block replacement; RANDOM_PERMUTE
	 elements are randomly permuted every time a new search is performed */
char *randomrepl;
#define RANDOM_PERMUTE 4

extern int nfsvers;
extern CLIENT *nfsdclient;

long int dblock_file_path_and_set(dcacheblock *block, char *path);


/*************************************************************************
 * Add a dirty dcache block into dirty_block_list
 *************************************************************************/
int add_dirty_dcache_block(dcacheblock *block, int seekset) {
	dirty_block *p=(dirty_block*)malloc(sizeof(dirty_block));

	if(p == NULL) return -1;

	p->id = DCACHE_DIRTY_BLOCK;
	p->nfsprogid = block->nfsprogid;
	p->nfsprocid = block->nfsprocid;
	switch(p->nfsprogid) {
	case 2:
		p->file.handle = block->contents.read_v2.file;
		p->offset = block->contents.read_v2.offset;
		break;
	case 3:
		p->file.handle3.length = block->contents.read_v3.file.length;
		memset(p->file.handle3.handle, 0, NFS3_FHSIZE);
		if((memcpy(p->file.handle3.handle,
							block->contents.read_v3.file.handle, block->contents.read_v3.file.length))==NULL) {
			Dprintf(D_CACHE,"add_dirty_dcache_block: memcpy failed\n");
			return -1;
		}
		p->offset = block->contents.read_v3.offset;
		break;
	default:
		return -1;
	}
	p->seekset = seekset;
	p->next = NULL;
	Dprintf(D_CACHE, "add_dirty_dcache_block: offset %d data_len %d\n",
				(block->nfsprogid == 2 ?
		block->contents.read_v2.offset : block->contents.read_v3.offset), (block->nfsprogid == 2 ?
				block->contents.read_v2.data_len : block->contents.read_v3.data_len));
	if(db_head == NULL) db_head = p;
	else db_tail->next = p;
	db_tail = p;

	return 0;
}

/*************************************************************************
 * Write back a dirty dcache block
 * return 1 if write back is successful, otherwise return -1 or error code
 *************************************************************************/
int dwrite_back_block(dcacheblock *block) {
	attrstat res;
	write3res res3;
	writeargs arg;
	write3args arg3;

	Dprintf(D_CACHE, "write back dcache block (offest %d, len %d)\n",
				(block->nfsprogid == 2 ?
		block->contents.read_v2.offset : block->contents.read_v3.offset), (block->nfsprogid == 2 ?
				block->contents.read_v2.data_len : block->contents.read_v3.data_len));

	switch(block->nfsprogid) {
	case 2:
		arg.file = block->contents.read_v2.file;
		arg.offset = block->contents.read_v2.offset;
		arg.beginoffset = block->contents.read_v2.offset;
		arg.totalcount = block->contents.read_v2.data_len;
		arg.data.data_len = block->contents.read_v2.data_len;
		hexdump(arg.file.data, NFS_FHSIZE, D_CACHE);
		/* a whole block is written back, which means multiple write to the
				same block is forwarded to server only once. */
		if((arg.data.data_val = (char *)malloc(arg.data.data_len)) == NULL)
			return -1;
		if((memcpy(arg.data.data_val, block->contents.read_v2.data_val,
			block->contents.read_v2.data_len)) == NULL) {
			Dprintf(D_CACHE, "dwrite_back_block: memcopy for data error\n");
			return -1;
		}
		if((nfsproc_write_2(&arg, &res, nfsdclient)) != RPC_SUCCESS ||
				res.status != NFS_OK) {
			Dprintf(D_CACHE, "write back error\n");
			free(arg.data.data_val);
			return -1;
		}
		free(arg.data.data_val);
		break;
	case 3:
		arg3.file.data.data_len = block->contents.read_v3.file.length;
		if((arg3.file.data.data_val = (char *)malloc(arg3.file.data.data_len)) == NULL) return -1;
		if((memcpy(arg3.file.data.data_val, block->contents.read_v3.file.handle,
							arg3.file.data.data_len))==NULL) {
			Dprintf(D_CACHE,"dwrite_back_block: memcpy failed\n");
			return -1;
		}
		hexdump(arg3.file.data.data_val, arg3.file.data.data_len, D_CACHE);
		arg3.offset = block->contents.read_v3.offset;
		arg3.count = block->contents.read_v3.count;
		arg3.data.data_len = arg3.count;
		if((arg3.data.data_val = (char *)malloc(arg3.data.data_len)) == NULL)
			return -1;
		if((memcpy(arg3.data.data_val, block->contents.read_v3.data_val,
			block->contents.read_v3.data_len)) == NULL) {
			Dprintf(D_CACHE, "dwrite_back_block: memcopy for data error\n");
			return -1;
		}
		if((nfsproc3_write_3(&arg3, &res3, nfsdclient)) != RPC_SUCCESS ||
				res3.status != NFS3_OK) {
			Dprintf(D_CACHE, "write back error\n");
			free(arg3.data.data_val);
			return -1;
		}
		free(arg3.data.data_val);
		break;
	default:
		return -1;
	}

	return 0;
}

/*************************************************************************
 * Hash file handle
 * return bucket number
 *************************************************************************/
unsigned long  int hash_fh(nfs_fh *file) {
	int i;
	double A=0.6180339887;
	unsigned long int B=0;

	/* accumulate file handle into B */
	for(i=0;i<NFS_FHSIZE;i++) {
		B+=(unsigned long int)(file->data[i]);
	}

	A = A * (double)B;
	A = A - floor(A);
	A = A * FH_HASH_SIZE;
	return (unsigned long int) (A) % FH_HASH_SIZE;
}

unsigned long  int hash_fh3(char *file) {
	int i;
	double A=0.6180339887;
	unsigned long int B=0;

	/* accumulate file handle into B */
	for(i=0; i<NFS3_FHSIZE; i++) {
			B+=(unsigned long int)(file[i]);
	}

	A = A * (double)B;
	A = A - floor(A);
	A = A * FH_HASH_SIZE;
	return (unsigned long int) (A) % FH_HASH_SIZE;
}

/*************************************************************************
 * Insert file handle into fh hash
 * each bucket in hash contains a linked list. file handles hashed into
 * the same bucket are chained in the list (no duplicated file handles).
 * return 1 if file handle is succesfully inserted
 * return 2 if file handle already exists and it's updated succesfully
 *************************************************************************/
int fh_hash_insert(fh_node *hash[], fh_node *node) {
	unsigned long int bucket;
	fh_node *p;
	int match;

	if(!hash) {
		Dprintf(D_CACHE, "fh_hash_insert: hash table is invalid\n");
		return -1;
	}

	if(nfsvers == 2 ) bucket = hash_fh(&node->file.handle);
	else bucket = hash_fh3(node->file.handle3.handle);

	if(hash[bucket] == NULL) { /* insert node at the head */
		hash[bucket] = node;
		node->next = NULL;
	} else { /* insert node at the tail */
		if(nfsvers == 2)
			match = memcmp(hash[bucket]->file.handle.data, node->file.handle.data, NFS_FHSIZE);
		else
			match = memcmp(hash[bucket]->file.handle3.handle, node->file.handle3.handle, node->file.handle3.length) != 0;
		if(match) {
			node->next = hash[bucket]->next;
			hash[bucket] = node;
			Dprintf(D_CACHE, "fh_hash_insert: updated fh node\n");
			return 2;
		}

		p = hash[bucket];
		while(p->next != NULL) {
			if(nfsvers == 2)
				match = memcmp(p->next->file.handle.data, node->file.handle.data, NFS_FHSIZE);
			else
				match = memcmp(p->next->file.handle3.handle, node->file.handle3.handle, node->file.handle3.length) != 0;
			if(match) {
				node->next = p->next->next;
				p->next = node;
				Dprintf(D_CACHE, "fh_hash_insert: updated fh node\n");
				return 2;
			}
			p = p->next;
		}

		p->next = node;
		node->next = NULL;
	}

	Dprintf(D_CACHE, "fh_hash_insert: fh node inserted at %d\n", bucket);

	return 1;
}

/*************************************************************************
 * Lookup file handle from fh hash
 * return 1 if file handle is found, otherwise return 0
 *************************************************************************/
fh_node* fh_hash_lookup(fh_node *hash[], nfs_fh *file) {
	unsigned long int bucket;
	fh_node *p;

	if(!hash) {
		Dprintf(D_CACHE, "hash table is invalid\n");
		return NULL;
	}

	bucket = hash_fh(file);
	Dprintf(D_CACHE, "fh_hash_lookup: bucket %lu", bucket);
	p = hash[bucket];

	while(p != NULL) {
		if(memcmp(p->file.handle.data, file->data, NFS_FHSIZE) == 0) {
			break;
		}
		p = p->next;
	}

	return p;
}

fh_node* fh_hash_lookup3(fh_node *hash[], nfs3_fh *file) {
	unsigned long int bucket;
	fh_node *p;

	if(!hash) {
		Dprintf(D_CACHE, "hash table is invalid\n");
		return NULL;
	}

	bucket = hash_fh3(file->handle);
	p = hash[bucket];

	while(p != NULL) {
		if(p->file.handle3.length != file->length || memcmp(p->file.handle3.handle, file->handle, file->length) == 0) {
			break;
		}
		p = p->next;
	}

	return p;
}

/*************************************************************************
 * Free file handle hash
 *************************************************************************/
int fh_hash_free(fh_node *hash[]) {
	int i;
	fh_node *p, *q;

	for(i=0; i<FH_HASH_SIZE; i++) {
		p = q = hash[i];
		while(q != NULL) {
			q = p->next;
			free(p);
			p = q;
		}
	}

	return 1;
}

/*************************************************************************
 * Flush dirty dcache blocks
 * return 1 if write back is successful, otherwise return -1
 *************************************************************************/
int flush_dirty_blocks(void)
{
	int fd, seekr, reads, err, lkupres;
	dirty_block *p = db_head, *p_next;
	time_t now;
	struct tm *tm;
	unsigned long int set_number;
	attrstat res_attr;
	getattr3res res_attr3;
	fh_node *node;
	nfs_fh3 fh3;

	Dprintf(D_CACHE, "### flush dirty dblocks\n");
	(void) time(&now);
	tm = localtime(&now);
	Dprintf(L_WARNING, "flush begins at %02d/%02d/%02d %02d:%02d:%02d",
	 tm->tm_mon + 1, tm->tm_mday, tm->tm_year,
	 tm->tm_hour, tm->tm_min, tm->tm_sec);
	while(p != NULL) {
		Dprintf(D_CACHE, "check block(offset: %d)\n", p->offset);
		if(nfsvers == 2) node = fh_hash_lookup(stale_fh_hash, &p->file.handle);
		else node = fh_hash_lookup3(stale_fh_hash, &p->file.handle3);
		if(node != NULL) {
			if(node->status == FH_STALE) {
				Dprintf(D_CACHE, "this file has been removed\n");
				p_next = p->next; free(p); p = p_next;
				continue;
			}
			else if(node->status == FH_TRUNC) {
				Dprintf(D_CACHE, "this file has been truncated\n");
				/* TODO */
				if(nfsvers == 2) {
					lkupres = cache_lookup(2, NFSPROC_GETATTR, (void *)&p->file.handle, (void *)&res_attr);
					if(lkupres != 1) {
						Dprintf(D_CACHE, "file size unknown; fetch attr from server\n");
							if((nfsproc_getattr_2(&p->file.handle, &res_attr, nfsdclient)) != RPC_SUCCESS || res_attr.status != NFS_OK) {
							Dprintf(D_CACHE, "error fetch file attributes\n");
							return -1;
						}
						cache_insert(2, NFSPROC_GETATTR, NULL, (void *)&res_attr);
					}
					if(p->offset >= res_attr.attrstat_u.attributes.size) {
							Dprintf(D_CACHE, "no need to write back (offset: %d, file size: %d)\n", p->offset, res_attr.attrstat_u.attributes.size);
							p_next = p->next; free(p); p = p_next;
							continue;
					}
				} else {
					fh3.data.data_len = p->file.handle3.length;
					fh3.data.data_val = p->file.handle3.handle;
					lkupres = cache_lookup(3, NFSPROC3_GETATTR, (void *)&fh3, (void *)&res_attr3);
					if(lkupres != 1) {
						Dprintf(D_CACHE, "file size unknown; fetch attr from server\n");
						if((nfsproc3_getattr_3(&fh3, &res_attr3, nfsdclient)) != RPC_SUCCESS || res_attr3.status != NFS3_OK) {
							Dprintf(D_CACHE, "error fetch file attributes\n");
							return -1;
						}
						cache_insert(3, NFSPROC3_GETATTR, NULL, (void *)&res_attr3);
					}
					if(p->offset >= res_attr3.getattr3res_u.attributes.size) {
						Dprintf(D_CACHE, "no need to write back (offset: %d, file size: %d)\n", p->offset, res_attr3.getattr3res_u.attributes.size);
						p_next = p->next; free(p); p = p_next;
						continue;
					}
				}
			}
		}

		if(p->nfsprogid == 2) {
			dblock_scratch_RPC->contents.read_v2.file = p->file.handle;
		dblock_scratch_RPC->contents.read_v2.offset = p->offset;
		} else {
			dblock_scratch_RPC->contents.read_v3.file = p->file.handle3;
			dblock_scratch_RPC->contents.read_v3.offset = p->offset;
		}

		set_number = dblock_file_path_and_set(dblock_scratch_RPC, dscratch_path);
		/* open cache bank in R+W mode and acquire an exclusive lock on it */
		if((fd = open(dscratch_path, O_RDWR)) == -1) {
			Dprintf(D_CACHE,"flush dirty dcache: could not open cache bank %s for reading\n",
			 dscratch_path);
			flock(fd,LOCK_UN);
			close(fd);
			return -1;
		}
		if ((flock(fd,LOCK_EX))==-1) {
			Dprintf(D_CACHE,"flush dirty dcache: could not lock dcache bank %s for r/w\n",
				dscratch_path);
			flock(fd,LOCK_UN);
			close(fd);
			return -1;
		}
		/* seek to the dirty block */
		seekr = lseek(fd, p->seekset, SEEK_SET);
		if(seekr != p->seekset) {
			Dprintf(D_CACHE,"flush dirty dcache: could not seek %i in %s\n",
				p->seekset, dscratch_path);
			flock(fd,LOCK_UN);
			close(fd);
			return -1;
		}
		/* read the dirty block and write it back to server if it's still dirty */
		if(p->id == DCACHE_DIRTY_BLOCK) {
			reads = read(fd,dblock_scratch_disk,sizeof(dcacheblock));
			if(reads != sizeof(dcacheblock)) {
				Dprintf(D_CACHE,"flush dirty dcache: could not read scan block %i in %s\n",
					p->seekset, dscratch_path);
				flock(fd,LOCK_UN);
				close(fd);
				return -1;
			}

			if((p->nfsprogid != dblock_scratch_disk->nfsprogid) ||
				(p->nfsprocid != dblock_scratch_disk->nfsprocid)) {
				Dprintf(D_CACHE, "progid/procid doesnot match!\n");
				goto next;
			}
			if(p->nfsprogid == 2) {
				if(memcmp(&p->file.handle, &dblock_scratch_disk->contents.read_v2.file,
					 NFS_FHSIZE) != 0) {
					Dprintf(D_CACHE, "file handle doesnot match!\n");
					goto next;
				}
				if(p->offset != dblock_scratch_disk->contents.read_v2.offset) {
					Dprintf(D_CACHE, "offset doesnot match(%d != %d)!\n", p->offset,
							dblock_scratch_disk->contents.read_v2.offset);
					goto next;
				}
			} else {
				if(p->file.handle3.length != dblock_scratch_disk->contents.read_v3.file.length || memcmp(p->file.handle3.handle, &dblock_scratch_disk->contents.read_v3.file.handle, p->file.handle3.length) != 0) {
					Dprintf(D_CACHE, "file handle doesnot match!\n");
					goto next;
				}
				if(p->offset != dblock_scratch_disk->contents.read_v3.offset) {
					Dprintf(D_CACHE, "offset doesnot match(%d != %d)!\n", p->offset,
													dblock_scratch_disk->contents.read_v3.offset);
					goto next;
				}
			}

			if(dblock_scratch_disk->state != ST_RW_DIRTY) {
				Dprintf(D_CACHE, "dcache block is not dirty\n");
				goto next;
			}

			if((err=dwrite_back_block(dblock_scratch_disk)) != 0) {
				Dprintf(D_CACHE, "flush dirty dcache: write back error(%d)\n", err);
				goto next;
				flock(fd,LOCK_UN);
				close(fd);
				return -1;
			}
			/* mark the block as clean and update cache */
			dblock_scratch_disk->state = ST_RO_CLEAN;
			seekr = lseek(fd, p->seekset, SEEK_SET);
			if(seekr != p->seekset) {
				Dprintf(D_CACHE,"flush dirty dcache: could not seek %i in %s\n",
					p->seekset, dscratch_path);
				flock(fd,LOCK_UN);
				close(fd);
				return -1;
			}
			if((write(fd,dblock_scratch_disk,sizeof(dcacheblock)))!=sizeof(dcacheblock)){
				Dprintf(D_CACHE,"flush dirty dcache: could not clean block in %s\n",
				dscratch_path);
				flock(fd,LOCK_UN);
				close(fd);
				return -1;
			}
		}
		else return -1;

next: /* release lock, close cache bank, free dirty block node, and go to the next node */
		if((flock(fd,LOCK_UN))==-1) {
			Dprintf(D_CACHE, "flush dirty dcache: could not unlock cache bank %s for reading\n",
					dscratch_path);
			close(fd);
			return -1;
		}

		p_next = p->next;
		free(p);
		p = p_next;
		close(fd);
	}

	fh_hash_free(stale_fh_hash);

	(void) time(&now);
	tm = localtime(&now);
	Dprintf(L_WARNING, "flush ends at %02d/%02d/%02d %02d:%02d:%02d",
	 tm->tm_mon + 1, tm->tm_mday, tm->tm_year,
	 tm->tm_hour, tm->tm_min, tm->tm_sec);
	return 0;
}

int flush_data_cache(void) {
	unsigned long int i;
	int staterr;

	Dprintf(D_CACHE, "Flushing data cache...\n");
	for(i=0; i<dcache_banks; i++) {
		/* Not yet checking for modified blocks!!! */
		if((strlen(dBaseCachePath)+1+strlen(dSessionID)+
				1+strlen("data_cache_bank_")+16)>MAX_STR) {
			Dprintf(L_ERROR,"cache block path too long\n");
			return -1;
		}
		sprintf(dscratch_path,"%s/%s/data_cache_bank_%lu",dBaseCachePath,
			dSessionID,i);
		staterr = stat(dscratch_path, statbuf);
		if(staterr==0) { /* file exists; remove it */
			if((unlink(dscratch_path))==-1) {
				Dprintf(L_ERROR, "Could not flush data cache bank!\n");
				return -1;
			}
		}
	}
	Dprintf(D_CACHE, "Done flushing data cache.\n");
	return 0;
}

/*************************************************************************
 * NFS-V2-READ cache entry manipulation functions
 *************************************************************************/

/*************************************************************************
 * Copy NFS-V2 read arguments into dcacheblock data structure
 * that is passed as argument. Returns 0 if successful, -1 if
 * error
 *************************************************************************/
int argument_read_v2(readargs *arg, dcacheblock *block) {

	if(arg==NULL) {
		Dprintf(D_CACHE,"argument_read_v2: arg==NULL\n");
		return -1;
	}

	/* zero out cache block */
	bzero(block,sizeof(*block));

	/* mark block as NFS V2 read */
	block->nfsprogid = 2;
	block->nfsprocid = NFSPROC_READ;

	/* copy contents of argument to cache block */
	if((memcpy(&block->contents.read_v2.file,&arg->file,NFS_FHSIZE))==NULL) {
		Dprintf(D_CACHE,"argument_read_v2: memcpy failed\n");
		return -1;
	}
	block->contents.read_v2.offset = arg->offset;
	block->contents.read_v2.count = arg->count;
	block->contents.read_v2.totalcount = arg->totalcount;
	return 0;
}

/*************************************************************************
 * Copy NFS-V3 read arguments into dcacheblock data structure
 * that is passed as argument. Returns 0 if successful, -1 if
 * error
 *************************************************************************/
int argument_read_v3(read3args *arg, dcacheblock *block) {

	if(arg==NULL) {
		Dprintf(D_CACHE,"argument_read_v3: arg==NULL\n");
		return -1;
	}

	/* zero out cache block */
	bzero(block,sizeof(*block));

	/* mark block as NFS V2 read */
	block->nfsprogid = 3;
	block->nfsprocid = NFSPROC3_READ;

	/* copy contents of argument to cache block */
	block->contents.read_v3.file.length = arg->file.data.data_len;
	memset(block->contents.read_v3.file.handle, 0, NFS3_FHSIZE);
	if((memcpy(block->contents.read_v3.file.handle,
			 arg->file.data.data_val, arg->file.data.data_len))==NULL) {
		Dprintf(D_CACHE,"argument_read_v3: memcpy failed\n");
		return -1;
	}
	block->contents.read_v3.offset = (int)arg->offset;
	block->contents.read_v3.count = arg->count;
	return 0;
}

/*************************************************************************
 * Copy NFS-V2 write arguments into dcacheblock data structure
 * that is passed as argument. Returns 0 if successful, -1 if
 * error
 *************************************************************************/
int argument_write_v2(writeargs *arg, dcacheblock *block) {

	if(arg==NULL) {
		Dprintf(D_CACHE,"argument_write_v2: arg==NULL\n");
		return -1;
	}

	/* zero out cache block */
	bzero(block,sizeof(*block));

	/* mark block as NFS V2 read */
	block->nfsprogid = 2;
	block->nfsprocid = NFSPROC_WRITE;

	/* copy contents of argument to cache block */
	block->contents.read_v2.file = arg->file;
	block->contents.read_v2.offset = arg->offset;
	block->contents.read_v2.count = arg->data.data_len;
	block->contents.read_v2.totalcount = arg->totalcount;
	block->contents.read_v2.data_len = arg->data.data_len;
	if((memcpy(block->contents.read_v2.data_val, arg->data.data_val,
		arg->data.data_len)) == NULL) {
		Dprintf(D_CACHE, "argument_write_v2: memcopy for data error\n");
		return -1;
	}

	return 0;
}

int argument_write_v3(write3args *arg, dcacheblock *block) {

	if(arg==NULL) {
		Dprintf(D_CACHE,"argument_write_v3: arg==NULL\n");
		return -1;
	}

	/* zero out cache block */
	bzero(block,sizeof(*block));

	/* mark block as NFS V2 read */
	block->nfsprogid = 3;
	block->nfsprocid = NFSPROC3_WRITE;

	/* copy contents of argument to cache block */
	block->contents.read_v3.file.length = arg->file.data.data_len;
	memset(block->contents.read_v3.file.handle, 0, NFS3_FHSIZE);
	if((memcpy(block->contents.read_v3.file.handle,
			 arg->file.data.data_val, arg->file.data.data_len))==NULL) {
		Dprintf(D_CACHE,"argument_write_v3: memcpy failed\n");
		return -1;
	}
	block->contents.read_v3.offset = arg->offset;
	block->contents.read_v3.count = arg->data.data_len;
	if((memcpy(block->contents.read_v3.data_val, arg->data.data_val,
		arg->data.data_len)) == NULL) {
		Dprintf(D_CACHE, "argument_write_v3: memcopy for data error\n");
		return -1;
	}

	return 0;
}

/*************************************************************************
 * Copy NFS-V2 read results into dcacheblock data structure
 * that is passed as argument. Returns 0 if successful, -1 if
 * error
 *************************************************************************/
int response_remote_RPC_read_v2(readres *res,
				dcacheblock *block) {
	u_int data_len;

	if(res==NULL) {
		Dprintf(D_CACHE,"response_remote_RPC_read_v2: res==NULL\n");
		return -1;
	}

	block->contents.read_v2.status = res->status;
	data_len = res->readres_u.reply.data.data_len;
	block->contents.read_v2.data_len = data_len;

	if(data_len > 0) /*server return zero size block if the block doesn't exist*/
		if((memcpy(&block->contents.read_v2.data_val[0],
				 &res->readres_u.reply.data.data_val[0],data_len))==NULL) {
			Dprintf(D_CACHE,"response_remote_RPC_read_v2: memcpy failed\n");
			return -1;
		}

	block->contents.read_v2.attributes = res->readres_u.reply.attributes;
	return 0;
}

/*************************************************************************
 * Copy NFS-V3 read results into dcacheblock data structure
 * that is passed as argument. Returns 0 if successful, -1 if
 * error
 *************************************************************************/
int response_remote_RPC_read_v3(read3res *res,
				dcacheblock *block) {
	u_int data_len;

	if(res==NULL) {
		Dprintf(D_CACHE,"response_remote_RPC_read_v3: res==NULL\n");
		return -1;
	}

	block->contents.read_v3.status = res->status;
	data_len = res->read3res_u.resok.data.data_len;
	block->contents.read_v3.data_len = data_len;

	if(data_len > 0) /*server return zero size block if the block doesn't exist*/
		if((memcpy(&block->contents.read_v3.data_val[0],
				 &res->read3res_u.resok.data.data_val[0], data_len))==NULL) {
			Dprintf(D_CACHE,"response_remote_RPC_read_v3: memcpy failed\n");
			return -1;
		}

	block->contents.read_v3.attributes = res->read3res_u.resok.file_attributes.post_op_attr_u.attributes;
	block->contents.read_v3.eof = res->read3res_u.resok.eof;
	return 0;
}

int response_remote_RPC_write_v2(readres *res,
				dcacheblock *block) {
	u_int data_len;

	if(res==NULL) {
		Dprintf(D_CACHE,"response_remote_RPC_read_v2: res==NULL\n");
		return -1;
	}

	block->contents.read_v2.status = res->status;
	data_len = res->readres_u.reply.data.data_len;
	block->contents.read_v2.data_len = data_len;
	block->contents.read_v2.count = data_len;
	block->contents.read_v2.totalcount = data_len;
	block->nfsprogid = 2;
	block->nfsprocid = NFSPROC_WRITE;

	if(data_len > 0) /*server return zero size block if the block doesn't exist*/
		if((memcpy(&block->contents.read_v2.data_val[0],
				 &res->readres_u.reply.data.data_val[0],data_len))==NULL) {
			Dprintf(D_CACHE,"response_remote_RPC_read_v2: memcpy error\n");
			return -1;
		}

	block->contents.read_v2.attributes = res->readres_u.reply.attributes;
	return 0;
}

/*************************************************************************
 * Copy NFS-V2 read cached data from dcacheblock data structure to RPC
 * response that is passed as argument. Returns 0 if successful, -1 if
 * error
 *************************************************************************/
int response_from_disk_cache_read_v2(dcacheblock *block, readargs *args,
						 readres *res) {
	int offset_index;
	int data_len;

	if(res==NULL) {
		Dprintf(D_CACHE,"response_from_disk_cache_read_v2: res==NULL\n");
		return -1;
	}
	if(args==NULL) {
		Dprintf(D_CACHE,"response_from_disk_cache_read_v2: args==NULL\n");
		return -1;
	}
	if(block->state == ST_RW_DIRTY) {
		Dprintf(D_CACHE, "block is dirty\n");
		res->status = 100;
	} else
		res->status = block->contents.read_v2.status;
	res->readres_u.reply.attributes = block->contents.read_v2.attributes;
	/* calculate where the offset is within the cache block */
	offset_index = args->offset % MAX_NFS_BLOCKSIZE;

	/* returned data may be more than required */
	if((block->contents.read_v2.data_len - offset_index) >= args->count)
		data_len = args->count;
	else  data_len = block->contents.read_v2.data_len - offset_index;

	res->readres_u.reply.data.data_len = data_len;
	if(data_len > 0)
		if((memcpy(&res->readres_u.reply.data.data_val[0],
					 &block->contents.read_v2.data_val[offset_index],
					 data_len))==NULL) {
			Dprintf(D_CACHE,"response_from_disk_cache_read_v2: memcpy failed\n");
			return -1;
		}

	return 0;
}

/*************************************************************************
 * Copy NFS-V3 read cached data from dcacheblock data structure to RPC
 * response that is passed as argument. Returns 0 if successful, -1 if
 * error
 *************************************************************************/
int response_from_disk_cache_read_v3(dcacheblock *block, read3args *args,
						 read3res *res) {
	u_int offset_index;
	u_int data_len;

	if(res==NULL) {
		Dprintf(D_CACHE,"response_from_disk_cache_read_v3: res==NULL\n");
		return -1;
	}
	if(args==NULL) {
		Dprintf(D_CACHE,"response_from_disk_cache_read_v3: args==NULL\n");
		return -1;
	}
	if(block->state == ST_RW_DIRTY) {
		Dprintf(D_CACHE, "block is dirty\n");
		res->status = 100;
	} else
		res->status = block->contents.read_v3.status;
	res->read3res_u.resok.file_attributes.post_op_attr_u.attributes = block->contents.read_v3.attributes;
	res->read3res_u.resok.file_attributes.present = TRUE;

	/* calculate where the offset is within the cache block */
	offset_index = (u_int)args->offset % MAX_NFS3_BLOCKSIZE;
	/* returned data may be more than required */
	if(block->contents.read_v3.data_len < offset_index) data_len = 0;
	else if((block->contents.read_v3.data_len - offset_index) > args->count) {
		data_len = args->count;
		res->read3res_u.resok.eof = FALSE;
	} else {
		data_len = block->contents.read_v3.data_len - offset_index;
		res->read3res_u.resok.eof = block->contents.read_v3.eof;
	}

	res->read3res_u.resok.data.data_len = data_len;
	res->read3res_u.resok.count = data_len;
	if(data_len > 0)
		if((memcpy(&res->read3res_u.resok.data.data_val[0],
					 &block->contents.read_v3.data_val[offset_index],
					 data_len))==NULL) {
			Dprintf(D_CACHE,"response_from_disk_cache_read_v3: memcpy failed\n");
			return -1;
		}

	return 0;
}

/*************************************************************************
 * Copy NFS-V2 read cached data from dcacheblock data structure to RPC
 * response that is passed as argument. Returns 0 if successful, -1 if
 * error
 *************************************************************************/
int response_from_disk_cache_write_v2(dcacheblock *block, writeargs *args, attrstat *res) {
	if(res==NULL) {
		Dprintf(D_CACHE,"response_from_disk_cache_write_v2: res==NULL\n");
		return -1;
	}
	res->status = block->contents.read_v2.status;
	res->attrstat_u.attributes = block->contents.read_v2.attributes;
	return 0;
}

int response_from_disk_cache_write_v3(dcacheblock *block, write3args *args, write3res *res) {
	if(res==NULL) {
		Dprintf(D_CACHE,"response_from_disk_cache_write_v3: res==NULL\n");
		return -1;
	}
	res->status = block->contents.read_v3.status;
	res->write3res_u.resok.file_wcc.after.post_op_attr_u.attributes = block->contents.read_v3.attributes;
	res->write3res_u.resok.file_wcc.after.present = TRUE;
	res->write3res_u.resok.file_wcc.before.present = FALSE;
	res->write3res_u.resok.count = args->count;
	res->write3res_u.resok.committed = FILE_SYNC;
	memset(res->write3res_u.resok.verf, 9, NFS3_WRITEVERFSIZE);
	return 0;
}

/*************************************************************************
 * Original hash function; block of (offset+MAX_NFS_BLOCKSIZE) maps to   *
 * same set, but two consecutive blocks will map to different set/bank   *
 *************************************************************************/
unsigned long int hash_readargs_v2_0(nfs_fh *fhandle, u_int offset) {
	int i;
	double A=0.6180339887;
	unsigned long int B=0;

	/* accumulate file handle into B */
	for(i=0;i<NFS_FHSIZE;i++) {
		B+=(unsigned long int)(fhandle->data[i]);
	}
	B+=offset-(offset%MAX_NFS_BLOCKSIZE);

	A = A * (double)B;
	A = A - floor(A);
	A = A * dcache_sets;

	return (unsigned long int) (A) % dcache_sets;
}

/*************************************************************************
 * Attempt to improve locality for sequential file reading;              *
 * HASH_CONSECUTIVEBLOCKS__SETS block of (offset+MAX_NFS_BLOCKSIZE) map  *
 * to same set; it seems to work well for few                            *
 * file handles right now, but it is important not to have               *
 * HASH_CONSECUTIVE_SETS larger that cache's associativity, otherwise    *
 * will thrash cache                                                     *
 *************************************************************************/
unsigned long int hash_readargs_v2_1(nfs_fh *fhandle, u_int offset) {
	int i;
	double A=0.6180339887;
	unsigned long int B=0;

	/* accumulate file handle into B */
	for(i=0;i<NFS_FHSIZE;i++) {
		B+=(unsigned long int)(fhandle->data[i]);
	}
	/* effective offset is the request's offset (modulo) the product
		 of the NFS block size and number of consecutive blocks that
		 map to a same set */
	B+=offset-(offset%(MAX_NFS_BLOCKSIZE*HASH_CONSECUTIVE_BLOCKS_SET));

	A = A * (double)B;
	A = A - floor(A);
	A = A * dcache_sets;
	return (unsigned long int) (A) % dcache_sets;
}

/*************************************************************************
 * Another attempt to improve locality for sequential file reading;      *
 * HASH_CONSECUTIVE_BLOCKS_SETS block of (offset+MAX_NFS_BLOCKSIZE) map  *
 * to same set; HASH_CONSECUTIVE_SETS of those map to consecutive sets   *
 * file handles right now, but it is important not to have               *
 * HASH_CONSECUTIVE_SETS larger that cache's associativity, otherwise    *
 * will thrash cache                                                     *
 *************************************************************************/
unsigned long int hash_readargs_v2_2(nfs_fh *fhandle, u_int offset) {
	int i;
	double A=0.6180339887,D;
	unsigned long int B=0,C;

	/* accumulate file handle into B */
	for(i=0;i<NFS_FHSIZE;i++) {
		B+=(unsigned long int)(fhandle->data[i]);
	}
	/* effective offset is the request's offset (modulo) the product
		 of the NFS block size and number of consecutive blocks that
		 map to a same set */
	B+=offset-(offset%(MAX_NFS_BLOCKSIZE*HASH_CONSECUTIVE_BLOCKS_SET*
				 HASH_CONSECUTIVE_SETS));

	A = A * (double)B;
	A = A - floor(A);
	/* C is a set in the range [0,dcache_sets/HASH_CONSECUTIVE_SETS-1] */
	A = A * (dcache_sets/HASH_CONSECUTIVE_SETS);
	C = (unsigned long int) (A) % (dcache_sets/HASH_CONSECUTIVE_SETS);
	/* Expand C to range [0,dcache_sets-1], based on value of offset */
	D = floor((offset/(MAX_NFS_BLOCKSIZE*HASH_CONSECUTIVE_BLOCKS_SET)));
	C = (C * HASH_CONSECUTIVE_SETS) + ( ((unsigned long int)(D))%
							HASH_CONSECUTIVE_SETS );
	return (C % dcache_sets);
}

/*************************************************************************
 * Hash V2 readargs (fh + offset)                                        *
 * Return set number                                                     *
 * hook to multiple possible hash functions                              *
 *************************************************************************/
unsigned long int hash_readargs_v2(nfs_fh *fhandle, u_int offset) {
	switch(HASH_FUNCTION_TO_USE) {
	case 1:
		return hash_readargs_v2_1(fhandle,offset);
		break;
	case 2:
		return hash_readargs_v2_2(fhandle,offset);
		break;
	case 0:
	default:
		return hash_readargs_v2_0(fhandle,offset);
		break;
	}
}

unsigned long int hash_readargs_v3_0(char *fhandle, u_int offset) {
	int i;
	double A=0.6180339887;
	unsigned long int B=0;
	Dprintf(D_CACHE, "hash v0\n");
	/* accumulate file handle into B */
	for(i=0;i<NFS3_FHSIZE;i++) {
		B+=(unsigned long int)(fhandle[i]);
	}
	B+=offset-(offset%MAX_NFS3_BLOCKSIZE);

	A = A * (double)B;
	A = A - floor(A);
	A = A * dcache_sets;

	return (unsigned long int) (A) % dcache_sets;
}

unsigned long int hash_readargs_v3_2(char *fhandle, u_int offset) {
	int i;
	double A=0.6180339887,D;
	unsigned long int B=0,C;

	/* accumulate file handle into B */
	for(i=0;i<NFS3_FHSIZE;i++) {
		B+=(unsigned long int)(fhandle[i]);
	}
	/* effective offset is the request's offset (modulo) the product
		 of the NFS block size and number of consecutive blocks that
		 map to a same set */
	B+=offset-(offset%(MAX_NFS3_BLOCKSIZE*HASH_CONSECUTIVE_BLOCKS_SET*HASH_CONSECUTIVE_SETS));

	A = A * (double)B;
	A = A - floor(A);
	/* C is a set in the range [0,dcache_sets/HASH_CONSECUTIVE_SETS-1] */
	A = A * (dcache_sets/HASH_CONSECUTIVE_SETS);
	C = (unsigned long int) (A) % (dcache_sets/HASH_CONSECUTIVE_SETS);
	/* Expand C to range [0,dcache_sets-1], based on value of offset */
	D = floor((offset/(MAX_NFS3_BLOCKSIZE*HASH_CONSECUTIVE_BLOCKS_SET)));
	C = (C * HASH_CONSECUTIVE_SETS) + ( ((unsigned long int)(D))%
							HASH_CONSECUTIVE_SETS );
	return (C % dcache_sets);
}

unsigned long int hash_readargs_v3(char *fhandle, u_int offset) {
	switch(HASH_FUNCTION_TO_USE) {
//   case 1:
//     return hash_readargs_v3_1(fhandle,offset);
//     break;
	case 2:
		return hash_readargs_v3_2(fhandle,offset);
		break;
	case 0:
	default:
		return hash_readargs_v3_0(fhandle,offset);
		break;
	}
}

/*************************************************************************
 * Determine the file name that a cache block maps to                    *
 * Returns set number,                                                   *
 *  and the path to the cache bank file if successfull, -1 otherwise     *
 *************************************************************************/
long int dblock_file_path_and_set(dcacheblock *block, char *path) {
	unsigned long int hash, bank_number;

	if(block==NULL) {
		Dprintf(D_CACHE,"dblock_file_path_and_set: block==NULL\n");
		return -1;
	}

	switch(block->nfsprogid) {
	case 2:
		hash = hash_readargs_v2(&(block->contents.read_v2.file),
					 (block->contents.read_v2.offset));
		bank_number = (unsigned long int) (hash/dsets_per_bank);
		if((strlen(dBaseCachePath)+1+strlen(dSessionID)+
			1+strlen("data_cache_bank_")+16)>MAX_STR) {
			Dprintf(L_ERROR,"cache block path too long\n");
			return -1;
		}
		sprintf(path,"%s/%s/data_cache_bank_%lu",dBaseCachePath,dSessionID,
			bank_number);
		Dprintf(D_CACHE,"dcache: HASH(%lu) = %lu (bank %lu)\n",
			block->contents.read_v2.offset,hash,bank_number);
		return (unsigned long int) hash;
		break; /* case 2 */
	case 3:
			hash = hash_readargs_v3(block->contents.read_v3.file.handle,
						 (block->contents.read_v3.offset));
			bank_number = (unsigned long int) (hash/dsets_per_bank);
			if((strlen(dBaseCachePath)+1+strlen(dSessionID)+
		1+strlen("data_cache_bank_")+16)>MAX_STR) {
				Dprintf(L_ERROR,"cache block path too long\n");
				return -1;
			}
			sprintf(path,"%s/%s/data_cache_bank_%lu",dBaseCachePath,dSessionID,
				bank_number);
			Dprintf(D_CACHE,"dcache: HASH(%lu) = %lu (bank %lu)\n",
				block->contents.read_v3.offset,hash,bank_number);
			return (unsigned long int) hash;
		break;
	default:
		return -1;
	}
	return -1;
}


/*************************************************************************
 * Check whether tag of NFS arguments matches cached data on disk,       *
 *  possibly invalidating cached data (if invalidate==1)                 *
 * block_RPC : has NFS arguments                                         *
 * block_disk: returns contents of a match on a hit                      *
 * return values:                                                        *
 * 0 : no match                                                          *
 * 1 : tag match                                                         *
 * 2 : tag match but request crosses cache block boundary                *
 * -1: error                                                             *
 *************************************************************************/
int dtag_match_and_inv(dcacheblock *block_RPC, dcacheblock *block_disk,
					 int invalidate) {
	int fd,reads,framescan;
	int file_handle_match = 0;
	int offset_match = 0;
	int overall_match = 0;
	int seekset, seekr, seekcur;
	unsigned long int set_number;
	int cachedoffset, requestoffset, staterr, baseoffset, requestcnt, org_state;
	int new_size, old_size, new_block_len, old_block_len;

	Dprintf(D_CACHE, "dtag_match(%d, %d)\n", block_RPC->nfsprogid,block_RPC->nfsprocid);
	if(block_RPC==NULL) {
		Dprintf(D_CACHE,"dtag_match: block_RPC==NULL\n");
		return -1;
	}

	/* obtain path and set number */
	set_number = dblock_file_path_and_set(block_RPC, dscratch_path);
	if(set_number==-1) {
		Dprintf(D_CACHE,"dtag_match: could not obtain set\n");
		return -1;
	}

	/* open & seek into 1st block of appropriate set in cache file */
	staterr = stat(dscratch_path, statbuf);
	if(staterr != 0) {
		Dprintf(D_CACHE, "dtag_match: bank does not exist\n");
		return 0;
	}
	if((fd = open(dscratch_path, O_RDONLY)) == -1) {
		Dprintf(D_CACHE,"dtag_match: could not open cache bank %s for reading\n",
			dscratch_path);
		return -1;
	}

	/* obtain a shared lock for cache lookup */
	if ((flock(fd,LOCK_SH))==-1) {
		Dprintf(D_CACHE,"dtag_match: could not lock cache bank %s for reading\n",
			dscratch_path);
		return -1;
	}

	seekset = (set_number % dsets_per_bank)*dcache_assoc*sizeof(*block_RPC);
	Dprintf(D_CACHE,
		"dtag_match: seeking %lu -> %lu\n",set_number,seekset);
	seekr = lseek(fd,seekset,SEEK_SET);
	if(seekr != seekset) {
		Dprintf(D_CACHE,"dtag_match: could not seek initial block %i in %s\n",
			seekset, dscratch_path);
		flock(fd,LOCK_UN);
		close(fd);
		return -1;
	}

	/* associative lookup; traverse all sets */
	/* stops before traversing the sets if a match was found */
	for(framescan=0; (framescan<dcache_assoc && !overall_match); framescan++) {
		reads = read(fd,block_disk,sizeof(*block_disk));
		if(reads != sizeof(*block_disk)) { /* read was not right */
			Dprintf(D_CACHE,"dtag_match: could not read scan block %i in %s\n",
				framescan, dscratch_path);
			flock(fd,LOCK_UN);
			close(fd);
			return -1;
		}
		/* different NFS calls require diff. tag matching */
		switch(block_RPC->nfsprogid) {
		case 2:
			switch(block_RPC->nfsprocid) {
			case NFSPROC_READ:
				if (block_disk->state == ST_INV) /* invalid block; no match */
					break;
				/* check if blocks' prog/proc ids match */
				if(!((block_RPC->nfsprogid==block_disk->nfsprogid) &&
						(block_RPC->nfsprocid==block_disk->nfsprocid)))
					break;
				/* check if offsets match */
				/* a write may begin in an offset different from where a read
					 was issued; must invalidate in this case. Hence, must check
					 offset % BLOCK_SIZE. */
				cachedoffset = block_disk->contents.read_v2.offset;
				requestoffset = block_RPC->contents.read_v2.offset;
				offset_match=(((cachedoffset-(cachedoffset%MAX_NFS_BLOCKSIZE))==
								 (requestoffset-(requestoffset%MAX_NFS_BLOCKSIZE))));
				if(!offset_match) break;
				/* check if file handles match */
				file_handle_match = memcmp(&block_RPC->contents.read_v2.file,
								 &block_disk->contents.read_v2.file,
								 NFS_FHSIZE);
							if(file_handle_match != 0) break;
				overall_match = 1;

				/* invalidate block in disk if required */
				if(invalidate) {
					/* keep current seek point in seekcur; will need it after
						 fd is reopened */
					seekcur = lseek(fd, 0, SEEK_CUR);
					/* release shared lock  */
					if ((flock(fd,LOCK_UN))==-1) {
						Dprintf(D_CACHE,
							"dtag_match: could not unlock cache bank %s for reading\n",
							dscratch_path);
						return -1;
					}
					/* reopen bank file in read/write mode  for invalidation */
					close(fd);
					if((fd = open(dscratch_path, O_RDWR)) == -1) {
						Dprintf(D_CACHE,"dtag_match: could not open cache bank %s R/W\n",
							dscratch_path);
						return -1;
					}
					/* obtain exclusive lock for invalidation */
					if ((flock(fd,LOCK_EX))==-1) {
						Dprintf(D_CACHE,
							"dtag_match: could not lock cache bank %s for writing\n",
							dscratch_path);
						return -1;
					}
					/* seek back one block from seekcur */
					seekset = seekcur-sizeof(*block_RPC); /* point back one block */
					seekr = lseek(fd, seekset, SEEK_SET);
					if(seekr!=seekset) {
						Dprintf(D_CACHE,"dtag_match: could not seek %i in file bank %s\n",
							seekset, dscratch_path);
						flock(fd,LOCK_UN);
						close(fd);
						return -1;
					}
					/* write an invalid block (all zeroes) */
					bzero(block_disk,sizeof(*block_disk));
					if((write(fd,block_disk,sizeof(*block_disk)))!=sizeof(*block_disk)) {
						Dprintf(D_CACHE,"dtag_match: could not invalidate block in %s\n",
							dscratch_path);
						flock(fd,LOCK_UN);
						close(fd);
						return -1;
					}
				}/* endif invalidate */
				break;

			case NFSPROC_WRITE: /* look up and update data cache*/
				if (block_disk->state == ST_INV) /* invalid block; no match */
					break;
				/* check if blocks' prog/proc ids match (proc id is stored as read) */
				if(!((block_RPC->nfsprogid==block_disk->nfsprogid) &&
					(block_RPC->nfsprocid==block_disk->nfsprocid+2))) {
					break;
				}
				/* check if base offsets match (requested write may not begin from NFS
						block edge, so mush check the base offset */
				cachedoffset = block_disk->contents.read_v2.offset;
				requestoffset = block_RPC->contents.read_v2.offset;
				offset_match=(((cachedoffset-(cachedoffset%MAX_NFS_BLOCKSIZE))==
								 (requestoffset-(requestoffset%MAX_NFS_BLOCKSIZE))));
				if(!offset_match) break;
				/* checks if file handles match */
				file_handle_match = memcmp(&block_RPC->contents.read_v2.file,
								 &block_disk->contents.read_v2.file,
								 NFS_FHSIZE);
							if(file_handle_match != 0) break;
				overall_match = 1;

				seekcur = lseek(fd, 0, SEEK_CUR); /* save seek offset for later use */

				if ((flock(fd,LOCK_UN))==-1) {
					Dprintf(D_CACHE,
							"dtag_match: could not unlock cache bank %s for reading\n",
							dscratch_path);
					return -1;
				} /* release shared lock */
				close(fd);
				if((fd = open(dscratch_path, O_RDWR)) == -1) {
					Dprintf(D_CACHE,"dtag_match: could not open cache bank %s R/W\n",
							dscratch_path);
					return -1;
				} /* reopen file in r/w mode */
				if ((flock(fd,LOCK_EX))==-1) {
					Dprintf(D_CACHE,
							"dtag_match: could not lock cache bank %s for writing\n",
							dscratch_path);
					return -1;
				} /* acquire exclusive lock */

				seekset = seekcur-sizeof(*block_RPC); /* seek back one block */
				seekr = lseek(fd, seekset, SEEK_SET);
				if(seekr!=seekset) {
					Dprintf(D_CACHE,"dtag_match: could not seek %i in file bank %s\n",
							seekset, dscratch_path);
					flock(fd,LOCK_UN);
					close(fd);
					return -1;
				}

				baseoffset = block_RPC->contents.read_v2.offset % MAX_NFS_BLOCKSIZE;
				requestcnt = block_RPC->contents.read_v2.data_len;
				old_block_len = block_disk->contents.read_v2.count;
				/* if there is a gap between previous block's data length and the new
					 write's offset, then it should be filled with zeros. */
				if(baseoffset > old_block_len)
					memset(&block_disk->contents.read_v2.data_val[old_block_len], 0,
						baseoffset - old_block_len);
				if((memcpy(&block_disk->contents.read_v2.data_val[baseoffset],
						block_RPC->contents.read_v2.data_val, requestcnt))
					== NULL) {
					Dprintf(D_CACHE, "memcpy for write_back update error\n");
					flock(fd,LOCK_UN);
					close(fd);
					return -1;
				} /* rewrite requested region in data cache block */

				org_state = block_disk->state;
				block_disk->state = ST_RW_DIRTY; /* set block state to dirty */

				/* update file attributes in cache, which are also returned to client
					 for the write procedure */
							new_size = block_RPC->contents.read_v2.offset +
					block_RPC->contents.read_v2.data_len;
				/* current file attrs are stored in the block_RPC argument. the attrs in
					 the block_disk are probably stale */
				old_size = block_RPC->contents.read_v2.attributes.size;
				new_block_len = baseoffset + requestcnt;
				/* Should be careful here: new_block_len may be smaller than
					 old_block_len because the file has been truncated but stale data
					 is left behind in this block. So we cannot determin the new block's
					 length by comaring it with the old block length. However, if the
					 write is appending to the file, then new_block_len should be used.
					 Otherwise, the previous block data cannot be stale, and we can
					 decide the new block length by comparing new_block_len and
					 old_block_len.
				*/
				if(new_size > old_size) { /* appending to file*/
					block_disk->contents.read_v2.attributes.size = new_size;
					block_disk->contents.read_v2.data_len = new_block_len;
				} else {
					block_disk->contents.read_v2.attributes.size = old_size;
					if(new_block_len > old_block_len)
						block_disk->contents.read_v2.data_len = new_block_len;
							}
				block_disk->contents.read_v2.count =
						block_disk->contents.read_v2.data_len;
				block_disk->contents.read_v2.totalcount =
						block_disk->contents.read_v2.data_len;
				Dprintf(D_CACHE, "block len %d, file size %d\n",
							block_disk->contents.read_v2.data_len,
							block_disk->contents.read_v2.attributes.size);

				/* write updated block into disk */
				if((write(fd,block_disk,sizeof(*block_disk)))!=sizeof(*block_disk)) {
					Dprintf(D_CACHE,"dtag_match: could not invalidate/update block in %s\n",
							dscratch_path);
					flock(fd,LOCK_UN);
					close(fd);
					return -1;
				}

				if(org_state != ST_RW_DIRTY)
				add_dirty_dcache_block(block_disk, seekset);
				break;
			} /*end switch(block_RPC->nfsprocid)*/
			break; /* case 2 */

		case 3:
			switch(block_RPC->nfsprocid) {
			case NFSPROC3_READ:
				if (block_disk->state == ST_INV) /* invalid block; no match */
					break;
				/* check if blocks' prog/proc ids match */
				if(!((block_RPC->nfsprogid==block_disk->nfsprogid) &&
						(block_RPC->nfsprocid==block_disk->nfsprocid)))
					break;
				/* check if offsets match */
				/* a write may begin in an offset different from where a read
					 was issued; must invalidate in this case. Hence, must check
					 offset % BLOCK_SIZE. */
				cachedoffset = block_disk->contents.read_v3.offset;
				requestoffset = block_RPC->contents.read_v3.offset;
				offset_match=(((cachedoffset-(cachedoffset%MAX_NFS3_BLOCKSIZE))==
								 (requestoffset-(requestoffset%MAX_NFS3_BLOCKSIZE))));
				if(!offset_match) break;
				/* check if file handles match */
				file_handle_match = memcmp(&block_RPC->contents.read_v3.file.handle,
								 block_disk->contents.read_v3.file.handle,
								 block_RPC->contents.read_v3.file.length);
							if(file_handle_match != 0) break;
				overall_match = 1;

				/* invalidate block in disk if required */
				if(invalidate) {
					/* keep current seek point in seekcur; will need it after
						 fd is reopened */
					seekcur = lseek(fd, 0, SEEK_CUR);
					/* release shared lock  */
					if ((flock(fd,LOCK_UN))==-1) {
						Dprintf(D_CACHE,
							"dtag_match: could not unlock cache bank %s for reading\n",
							dscratch_path);
						return -1;
					}
					/* reopen bank file in read/write mode  for invalidation */
					close(fd);
					if((fd = open(dscratch_path, O_RDWR)) == -1) {
						Dprintf(D_CACHE,"dtag_match: could not open cache bank %s R/W\n",
							dscratch_path);
						return -1;
					}
					/* obtain exclusive lock for invalidation */
					if ((flock(fd,LOCK_EX))==-1) {
						Dprintf(D_CACHE,
							"dtag_match: could not lock cache bank %s for writing\n",
							dscratch_path);
						return -1;
					}
					/* seek back one block from seekcur */
					seekset = seekcur-sizeof(*block_RPC); /* point back one block */
					seekr = lseek(fd, seekset, SEEK_SET);
					if(seekr!=seekset) {
						Dprintf(D_CACHE,"dtag_match: could not seek %i in file bank %s\n",
							seekset, dscratch_path);
						flock(fd,LOCK_UN);
						close(fd);
						return -1;
					}
					/* write an invalid block (all zeroes) */
					bzero(block_disk,sizeof(*block_disk));
					if((write(fd,block_disk,sizeof(*block_disk)))!=sizeof(*block_disk)) {
						Dprintf(D_CACHE,"dtag_match: could not invalidate block in %s\n",
							dscratch_path);
						flock(fd,LOCK_UN);
						close(fd);
						return -1;
					}
				}/* endif invalidate */
				break;

			case NFSPROC3_WRITE: /* look up and update data cache*/
				if (block_disk->state == ST_INV) {/* invalid block; no match */
				//           Dprintf(D_CACHE, "invalid block\n");
								break;
							}
				if(!((block_RPC->nfsprogid==block_disk->nfsprogid) &&
					(block_RPC->nfsprocid==block_disk->nfsprocid+1))) {
				//           Dprintf(D_CACHE, "%d <> %d\n", block_RPC->nfsprocid, block_disk->nfsprocid);
					break;
				}
				/* check if base offsets match (requested write may not begin from NFS
						block edge, so mush check the base offset */
				cachedoffset = block_disk->contents.read_v3.offset;
				requestoffset = block_RPC->contents.read_v3.offset;
				offset_match=(((cachedoffset-(cachedoffset%MAX_NFS3_BLOCKSIZE))==
								 (requestoffset-(requestoffset%MAX_NFS3_BLOCKSIZE))));
				if(!offset_match) break;
				/* checks if file handles match */
				file_handle_match = memcmp(block_RPC->contents.read_v3.file.handle,
								 block_disk->contents.read_v3.file.handle,
								 block_RPC->contents.read_v3.file.length);
				if(file_handle_match != 0) {
					break;
				}
				overall_match = 1;

				seekcur = lseek(fd, 0, SEEK_CUR); /* save seek offset for later use */

				if ((flock(fd,LOCK_UN))==-1) {
					Dprintf(D_CACHE,
							"dtag_match: could not unlock cache bank %s for reading\n",
							dscratch_path);
					return -1;
				} /* release shared lock */
				close(fd);
				if((fd = open(dscratch_path, O_RDWR)) == -1) {
					Dprintf(D_CACHE,"dtag_match: could not open cache bank %s R/W\n",
							dscratch_path);
					return -1;
				} /* reopen file in r/w mode */
				if ((flock(fd,LOCK_EX))==-1) {
					Dprintf(D_CACHE,
							"dtag_match: could not lock cache bank %s for writing\n",
							dscratch_path);
					return -1;
				} /* acquire exclusive lock */

				seekset = seekcur-sizeof(*block_RPC); /* seek back one block */
				seekr = lseek(fd, seekset, SEEK_SET);
				if(seekr!=seekset) {
					Dprintf(D_CACHE,"dtag_match: could not seek %i in file bank %s\n",
							seekset, dscratch_path);
					flock(fd,LOCK_UN);
					close(fd);
					return -1;
				}

				baseoffset = block_RPC->contents.read_v3.offset % MAX_NFS3_BLOCKSIZE;
				requestcnt = block_RPC->contents.read_v3.count;
				old_block_len = block_disk->contents.read_v3.data_len;
				/* if there is a gap between previous block's data length and the new
					 write's offset, then it should be filled with zeros. */
				Dprintf(D_CACHE, "baseoffset %d, old_block_len %d, requestcnt %d\n", baseoffset, old_block_len, requestcnt);
				if(baseoffset > old_block_len)
				memset(&block_disk->contents.read_v3.data_val[old_block_len], 0,
					baseoffset - old_block_len);
				if((memcpy(&block_disk->contents.read_v3.data_val[baseoffset],
						block_RPC->contents.read_v3.data_val, requestcnt))
					== NULL) {
					Dprintf(D_CACHE, "memcpy for write_back update error\n");
					flock(fd,LOCK_UN);
					close(fd);
					return -1;
				} /* rewrite requested region in data cache block */

				org_state = block_disk->state;
				block_disk->state = ST_RW_DIRTY; /* set block state to dirty */

				/* update file attributes in cache, which are also returned to client
				   for the write procedure */
				new_size = block_RPC->contents.read_v3.offset +
					block_RPC->contents.read_v3.count;
				/* current file attrs are stored in the block_RPC argument. the attrs in
				   the block_disk are probably stale */
				old_size = block_RPC->contents.read_v3.attributes.size;
				new_block_len = baseoffset + requestcnt;
				if(new_size > old_size) { /* appending to file*/
					block_disk->contents.read_v3.attributes.size = new_size;
					block_disk->contents.read_v3.data_len = new_block_len;
				} else {
					block_disk->contents.read_v3.attributes.size = old_size;
					if(new_block_len > old_block_len)
						block_disk->contents.read_v3.data_len = new_block_len;
							}
				block_disk->contents.read_v3.count =
						block_disk->contents.read_v3.data_len;
				Dprintf(D_CACHE, "block len %d, file size %d\n",
							block_disk->contents.read_v3.data_len,
							block_disk->contents.read_v3.attributes.size);

				/* write updated block into disk */
				if((write(fd,block_disk,sizeof(*block_disk)))!=sizeof(*block_disk)) {
					Dprintf(D_CACHE,"dtag_match: could not invalidate/update block in %s\n",
							dscratch_path);
					flock(fd,LOCK_UN);
					close(fd);
					return -1;
				}

				if(org_state != ST_RW_DIRTY)
					add_dirty_dcache_block(block_disk, seekset);
				break;
			} /*end switch(block_RPC->nfsprocid)*/
			break;
		} /*end switch(block_RPC->nfsprogid)*/
	}/* end for framescan */

	/* release shared lock  */
	if ((flock(fd,LOCK_UN))==-1) {
		Dprintf(D_CACHE,
			"dtag_match: could not unlock cache bank %s\n", dscratch_path);
		return -1;
	}
	close(fd);

	Dprintf(D_CACHE,"dtag_match(%d): offset(%lu) returns %i (frame %i, state %d)\n",
			block_RPC->nfsprocid,
			(block_RPC->nfsprogid == 2 ?
			block_RPC->contents.read_v2.offset : block_RPC->contents.read_v3.offset),
			overall_match, framescan-1, block_disk->state);
	return overall_match;
}

/*************************************************************************
 * Finds if there is a frame with specified state for a cache block      *
 * creates new file for cache bank if necessary                          *
 * return values:                                                        *
 * 0 =< x < cache_assoc        : free frame found in x                   *
 * cache_assoc:                : no free frames found                    *
 * -1: error                                                             *
 *************************************************************************/
int dfind_replace_frame(unsigned long int set_number, dcacheblock *block_disk,
			int state) {
	int fd,reads,framescan;
	int seekset, seekr;
	int idx1,idx2,i,ctmp;

	/* open & seek into 1st block of appropriate set in cache file */
	if((fd = open(dscratch_path, O_RDONLY)) == -1) {
		Dprintf(D_CACHE,
			"dfind_replace_frame: could not open cache bank %s for reading\n",
			dscratch_path);
		return -1;
	}
	/* obtain a shared lock for cache lookup */
	if ((flock(fd,LOCK_SH))==-1) {
		Dprintf(D_CACHE,
			"dfind_replace_frame: could not lock cache bank %s for reading\n",
			dscratch_path);
		return -1;
	}

	seekset = (set_number % dsets_per_bank)*dcache_assoc*sizeof(dcacheblock);
	Dprintf(D_CACHE,
		"dfind_replace_frame: seeking %lu -> %lu\n",set_number,seekset);

	seekr = lseek(fd,seekset,SEEK_SET);
	if(seekr != seekset) {
		Dprintf(D_CACHE,
			"dfind_replace_frame: could not seek initial block %i in %s\n",
			seekset, dscratch_path);
		flock(fd,LOCK_UN);
		close(fd);
		return -1;
	}

	/* randomly permute elements of randomrepl */
	for(i=0;i<RANDOM_PERMUTE;i++) {
		idx1 = random()%dcache_assoc;
		idx2 = random()%dcache_assoc;
		ctmp = randomrepl[idx1];
		randomrepl[idx1] = randomrepl[idx2];
		randomrepl[idx2] = ctmp;
	}

	/* associative lookup; traverse all sets */
	for(i=0; i<dcache_assoc; i++) {

		if(state==ST_INV) {
			/* linear scan; no need to seek */
			framescan = i;
		} else {
			/* random scan; need to seek appropriate block */
			framescan = randomrepl[i];
			seekr = lseek(fd,(seekset+framescan*sizeof(dcacheblock)),SEEK_SET);
			if(seekr != seekset+framescan*sizeof(dcacheblock)) {
				Dprintf(D_CACHE,
					"dfind_replace_frame: could not seek random block %i in %s\n",
					framescan, dscratch_path);
				flock(fd,LOCK_UN);
				close(fd);
				return -1;
			}
		}

		reads = read(fd,block_disk,sizeof(*block_disk));
		if(reads != sizeof(*block_disk)) { /* read was not right */
			Dprintf(D_CACHE,
				"find_replace_frame: could not read scan block %i in %s\n",
				framescan, dscratch_path);
			flock(fd,LOCK_UN);
			close(fd);
			return -1;
		}
		if(block_disk->state == state) {
			/* release shared lock  */
			if ((flock(fd,LOCK_UN))==-1) {
				Dprintf(D_CACHE,
					"dfind_replace_frame: could not unlock cache bank %s\n",
					dscratch_path);
				return -1;
			}
			close(fd);
			Dprintf(D_CACHE,
				"dfind_replace_frame: found %i (state %d)\n",
				framescan, block_disk->state);
			return framescan;
		}
	}

	/* release shared lock  */
	if ((flock(fd,LOCK_UN))==-1) {
		Dprintf(D_CACHE,
			"dfind_replace_frame: could not unlock cache bank %s\n",
			dscratch_path);
		return -1;
	}
	close(fd);
	return dcache_assoc;
}

/*************************************************************************
 * Insert result from RPC call in disk cache                             *
 * 0 : inserted                                                          *
 *-1 : error                                                             *
 *************************************************************************/
int dinsert_in_cache(dcacheblock *block_RPC, dcacheblock *block_disk) {
	int freeset,replaceset,foundset;
	unsigned long int set_number;
	int staterr, fd;
	int written, seekset, seekr, seekcur, err, lkupres;
	char c = '\0';
	attrstat res_attr;
	fh_node *node;

	if(block_RPC==NULL)
		return -1;

	/* determine if cache bank exists; if not, create it */
	set_number = dblock_file_path_and_set(block_RPC, dscratch_path);
	staterr = stat(dscratch_path, statbuf);

	if(staterr!=0) { /* create new bank */
		if((fd = creat(dscratch_path, S_IRUSR | S_IWUSR |
			 S_IRGRP | S_IWGRP ))==-1) {
			Dprintf(D_CACHE,"dinsert_in_cache: could not create %s for R/W\n",
				dscratch_path);
			return -1;
		}

		seekset = dsets_per_bank*dcache_assoc*sizeof(*block_disk)-1;
		seekr = lseek(fd, seekset, SEEK_SET);
		if(seekr != seekset) {
			Dprintf(D_CACHE,
				"dinsert in cache: could not seek initial block %i in %s\n",
				seekset, dscratch_path);
			close(fd);
			return -1;
		}
		if(write(fd, &c, 1) != 1) {
			Dprintf(D_CACHE,"dinsert_in_cache: could not write to %s\n",
				dscratch_path);
			close(fd);
			return -1;
		}
		close(fd);
	} else { /* bank file exists; check permissions */
		if(!((statbuf->st_mode & S_IRUSR) &&
	 (statbuf->st_mode & S_IWUSR))) {
			Dprintf(D_CACHE,"dinsert_in_cache: wrong permissions in %s\n",
				dscratch_path);
			return -1;
		}
	}

	/* first try, look for empty block */
	if((freeset = dfind_replace_frame(set_number, block_disk, ST_INV))==-1) {
		Dprintf(D_CACHE,"dinsert_in_cache: error finding replace frame ST_INV\n");
		return -1;
	}
	if(freeset == dcache_assoc) { /* second try, look for clean block */
		Dprintf(D_CACHE, "No empty blocks\n");
		if((replaceset = dfind_replace_frame(set_number, block_disk,
					 ST_RO_CLEAN))==-1) {
			Dprintf(D_CACHE,
				"dinsert_in_cache: error finding replace frame ST_RO_CLEAN\n");
			return -1;
		}
		if(replaceset == dcache_assoc) { /* last try, look for dirty block */
			Dprintf(D_CACHE, "No clean blocks\n");
			if((replaceset = dfind_replace_frame(set_number, block_disk,
										ST_RW_DIRTY)) == -1) {
				Dprintf(D_CACHE,
					"dinsert_in_cache: error finding replace frame ST_RW_DIRTY\n");
				return -1;
			}
			if(replaceset == dcache_assoc) { /* mission impossible */
				Dprintf(D_CACHE,
					"dinsert_in_cache: could not find block to evict\n");
				return -1;
			}
		}
		if(block_disk->state == ST_RW_DIRTY) {
			node = fh_hash_lookup(stale_fh_hash, &block_disk->contents.read_v2.file);
			if(node != NULL) {
		if(node->status == FH_STALE) {
			/* no need to write back if the file has been removed */
			Dprintf(D_CACHE, "dinsert_in_cache: this file has been removed\n");
		} else if(node->status == FH_TRUNC) {
			/* if the file has been truncated once, there may still exist
				 stale data blocks of the file in cache unless the file has
				 grown again and updated these blocks */
			Dprintf(D_CACHE, "dinsert_in_cache: this file has been truncated\n");
			lkupres = cache_lookup(2, NFSPROC_GETATTR,
						(void *)&block_disk->contents.read_v2.file,
					(void *)&res_attr);
				if(lkupres != 1) {
					Dprintf(D_CACHE, "file size unknown; fetch attr from server\n");
					if((nfsproc_getattr_2(&block_disk->contents.read_v2.file, &res_attr, nfsdclient))
							!= RPC_SUCCESS || res_attr.status != NFS_OK) {
						Dprintf(D_CACHE, "error fetch file attributes\n");
						return -1;
					}
				}
				if(block_disk->contents.read_v2.offset >= res_attr.attrstat_u.attributes.size) {
					Dprintf(D_CACHE, "dinsert_in_cache: no need to write back (offset: %d, file size: %d)\n",
						block_disk->contents.read_v2.offset, res_attr.attrstat_u.attributes.size);
				}
				else if((err=dwrite_back_block(block_disk)) != 0) {
					Dprintf(D_CACHE,"dinsert_in_cache: write back error(%d)\n", err);
				}
			}
		} else if((err=dwrite_back_block(block_disk)) != 0)
			Dprintf(D_CACHE,"dinsert_in_cache: write back error(%d)\n", err);
		}
		foundset = replaceset;
		Dprintf(D_CACHE,"dinsert_in_cache: offset(%lu) replaces block %i\n",
			block_RPC->contents.read_v2.offset,replaceset);
	} else {
		foundset = freeset;
		Dprintf(D_CACHE,"dinsert_in_cache: offset(%lu) found empty frame %i\n",
			(block_RPC->nfsprogid == 2 ?
			block_RPC->contents.read_v2.offset : block_RPC->contents.read_v3.offset),
			freeset);
	}

	/* open cache bank */
	if((fd = open(dscratch_path, O_RDWR))==-1) {
		Dprintf(D_CACHE, "dinsert_in_cache: could not open file bank %s\n",
			dscratch_path);
		return -1;
	}

	/* obtain exclusive lock for insertion */
	if ((flock(fd,LOCK_EX))==-1) {
		Dprintf(D_CACHE,
			"dinsert_in_cache: could not lock cache bank %s for writing\n",
			dscratch_path);
		return -1;
	}

	/* seek into appropriate insert position */
	/* first, seek into first block of set */
	seekset = (set_number % dsets_per_bank)*dcache_assoc*sizeof(*block_RPC);
	seekr = lseek(fd,seekset,SEEK_SET);
	if(seekr != seekset) {
		Dprintf(D_CACHE,
			"dinsert in cache: could not seek initial block %i in %s\n",
			seekset, dscratch_path);
		flock(fd,LOCK_UN);
		close(fd);
		return -1;
	}

	/* then, seek within set for replaced frame */
	seekcur = seekr; /* remember current seek point */
	seekset = foundset*sizeof(*block_RPC); /* determine offset within set */
	seekr = lseek(fd, seekset, SEEK_CUR); /* offset seek */
	if((seekcur+seekset) != seekr) {
		Dprintf(D_CACHE, "dinsert_in_cache: could not seek %i in file bank %s\n",
			seekset, dscratch_path);
		flock(fd,LOCK_UN);
		close(fd);
		return -1;
	}

	/* determine cache block state (RO, RW) */
	switch(block_RPC->nfsprogid) {
	case 2:
		switch(block_RPC->nfsprocid) {
		case NFSPROC_READ:
			block_RPC->state = ST_RO_CLEAN;
			break;
		case NFSPROC_WRITE:
			block_RPC->state = ST_RW_DIRTY;
			block_RPC->nfsprocid = NFSPROC_READ;
			add_dirty_dcache_block(block_RPC, seekr);
			break;
		default:
			block_RPC->state = ST_INV;
		}
		break; /* case 2 */
	case 3:
		switch(block_RPC->nfsprocid) {
		case NFSPROC3_READ:
		case NFSPROC3_WRITE:
			block_RPC->state = ST_RO_CLEAN;
			block_RPC->nfsprocid = NFSPROC3_READ;
			break;
		}
		break;
	default:
		block_RPC->state = ST_INV;
	}

	/* write block into cache */
	written = write(fd, block_RPC, sizeof(*block_RPC));
	if(written != sizeof(*block_RPC)) {
		Dprintf(L_ERROR,"Could not write into cache!\n");
		flock(fd,LOCK_UN);
		close(fd);
		return -1;
	}
	/* release exclusive lock  */
	if ((flock(fd,LOCK_UN))==-1) {
		Dprintf(D_CACHE,
			"dinsert_in_cache: could not unlock cache bank %s for writing\n",
			dscratch_path);
		return -1;
	}
	close(fd);
	return 0;
}

/*************************************************************************
 * Initialize global variables                                           *
 * path    : base path to file cache directory                           *
 * return values:                                                        *
 * 0 : successfull                                                       *
 * -1 : unsuccessful                                                     *
 *************************************************************************/
int init_data_cache_global(char *basepath, char *sessionid) {
	int staterr, mkdirerr, i;

	if (!(dscratch_path = (char *)malloc(sizeof(*dscratch_path)*MAX_STR)))
		return -1;

	if (!(dSessionID = (char *)malloc(sizeof(*dSessionID)*MAX_STR)))
		return -1;
	if(strlen(sessionid)>(sizeof(*dSessionID)*MAX_STR))
		return -1;
	strcpy(dSessionID,sessionid);

	if (!(dBaseCachePath = (char *)malloc(sizeof(*dBaseCachePath)*MAX_STR)))
		return -1;
	if(strlen(basepath)>(sizeof(*dBaseCachePath)*MAX_STR))
		return -1;
	strcpy(dBaseCachePath,basepath);

	if (!(statbuf = (struct stat *) malloc (sizeof(*statbuf))))
		return -1;

	if (!(dblock_scratch_RPC=(dcacheblock *)
	malloc(sizeof(*dblock_scratch_RPC))))
		return -1;

	if (!(dblock_scratch_disk=(dcacheblock *)
	malloc(sizeof(*dblock_scratch_disk))))
		return -1;

	if ((randomrepl=malloc(sizeof(*randomrepl)*dcache_assoc)) == NULL)
		return -1;
	for(i=0;i<dcache_assoc;i++)
		randomrepl[i]=i;

	/* check the existence of base path and session ID directories */
	staterr = stat(dBaseCachePath, statbuf);
	if(!((staterr==0) && (statbuf->st_mode & S_IRUSR) &&
			 (statbuf->st_mode & S_IWUSR))) {
		Dprintf(L_ERROR, "dBaseCachePath directory %s does not exist\n",
			dBaseCachePath);
		return -1;
	}
	if((strlen(dBaseCachePath)+1+strlen(dSessionID)+
			1+strlen("data_cache_bank_")+16)>MAX_STR) {
		Dprintf(L_ERROR,"cache block path too long\n");
		return -1;
	}
	sprintf(dscratch_path,"%s/%s",dBaseCachePath,dSessionID);
	staterr = stat(dscratch_path, statbuf);
	if(!((staterr==0) && (statbuf->st_mode & S_IRUSR) &&
			 (statbuf->st_mode & S_IWUSR))) {
		mkdirerr = mkdir(dscratch_path,S_IRWXU | S_IRWXG);
		if(mkdirerr==-1) {
			Dprintf(L_ERROR, "Could not create dSessionID directory %s\n",
				dscratch_path);
			return -1;
		}
	}
	return 0;
}

void dump_data_cache_parameters(FILE *fp) {
	fprintf(fp,"Data cache parameters:\n");
	fprintf(fp,"\tCache size: %lu blocks\n",dcache_size);
	fprintf(fp,"\tCache assoc: %du\n",dcache_assoc);
	fprintf(fp,"\t# cache banks: %du\n",dcache_banks);
	fprintf(fp,"\t# cache sets: %lu\n",dcache_sets);
	fprintf(fp,"\t# sets per bank: %lu\n",dsets_per_bank);
	fprintf(fp,"\tSize of cache block: %i\n",sizeof(*dblock_scratch_RPC));
	fprintf(fp,"\tSize of cache bank file: %lu\n",
		dsets_per_bank*dcache_assoc*sizeof(*dblock_scratch_RPC));
}

/*************************************************************************
 * Initialize cache parameters                                           *
 * size    : cache size (# entries, pwr of 2)                            *
 * assoc   : cache associativity (pwr of 2)                              *
 * return values:                                                        *
 * 0 : successfull                                                       *
 *-1 : unsuccessful                                                      *
 *************************************************************************/
int init_data_cache(unsigned long int size, int assoc, int banks,
				char *basepath, char *sessionid) {

	unsigned long int localsize, localassoc, localbanks;

	if(basepath==NULL || sessionid==NULL)
		return -1;

	if(size==0 || size > MAX_CACHE_SIZE) {
		Dprintf(L_ERROR,"Dcache size cannot be larger than %d\n", MAX_CACHE_SIZE);
		return -1;
	}

	if (assoc==0 || assoc > MAX_CACHE_ASSOC) {
		Dprintf(L_ERROR,"Dcache associativity cannot be larger than %d\n", MAX_CACHE_ASSOC);
		return -1;
	}

	if(init_data_cache_global(basepath,sessionid) == -1)
		return -1;

	localsize=size;
	while (localsize>1) {
		if(localsize%2) {
			Dprintf(L_ERROR,"Dcache size not a power of two\n");
			return -1;
		}
		localsize>>=1;
	}

	localassoc=assoc;
	while (localassoc>1) {
		if(localassoc%2) {
			Dprintf(L_ERROR,"Dcache associativity not a power of two\n");
			return -1;
		}
		localassoc>>=1;
	}

	localbanks=banks;
	while (localbanks>1) {
		if(localbanks%2) {
			Dprintf(L_ERROR,"Dcache banks not a power of two\n");
			return -1;
		}
		localbanks>>=1;
	}

	if((banks*assoc)>size) {
			Dprintf(L_ERROR,"Dcache size < banks * associativity \n");
		return -1;
	}

	db_head = db_tail = NULL;
	memset(stale_fh_hash, 0, FH_HASH_SIZE*sizeof(fh_node *));

	dcache_size = size;
	dcache_assoc = assoc;
	dcache_banks = banks;
	dcache_sets = dcache_size/dcache_assoc;
	dsets_per_bank = dcache_size/(dcache_assoc*dcache_banks);
	Dprintf(L_NOTICE, "data cache init successful\n"
										"size: %d, assoc: %d, banks: %d "
				"sets: %d, sets per bank: %d\n",
				dcache_size, dcache_assoc, dcache_banks,
				dcache_sets, dsets_per_bank);

	return 0;
}

/*************************************************************************
 * lookup block in cache                                                 *
 * prog_index, proc_index: type of RPC call                              *
 * argument = (void *)rpcargs from incoming RPC call                     *
 * return values:                                                        *
 * 0 : miss                                                              *
 * 1 : hit                                                               *
 * -1: error                                                             *
 * side-effects:                                                         *
 *  block_scratch_RPC contains copy of diropargs                         *
 *  block_scratch_disk contains copy of results in case of a hit         *
 *  res contains RPC result looked up from cache                         *
 *************************************************************************/
int dcache_lookup(int prog_index, int proc_index, void *rpcargs, void *res) {
	int match = 0;

	switch(prog_index) {
	case 2:
		switch(proc_index) {
		case NFSPROC_READ:
			/* create cache_scratch_RPC from rpcargs */
			if((argument_read_v2((readargs *)rpcargs, dblock_scratch_RPC))==-1) {
				Dprintf(D_CACHE,"dcache_lookup: argument copy failed\n");
				return -1;
			}
			if((match=dtag_match_and_inv(dblock_scratch_RPC,
					 dblock_scratch_disk, 0))==-1) {
				Dprintf(D_CACHE,"dcache_lookup: tag match failed\n");
				return -1;
			}
			if (match == 1) { /* hit; copy data */
				if((response_from_disk_cache_read_v2(dblock_scratch_disk,
										 (readargs *)rpcargs,
										 (readres *)res))==-1) {
					Dprintf(D_CACHE,"dcache_lookup: response copy failed\n");
					return -1;
				}
			}
			break;
		} /* switch (proc_index) */
		break; /* case 2 */
	case 3:
		switch(proc_index) {
		case NFSPROC3_READ:
			/* create cache_scratch_RPC from rpcargs */
			if((argument_read_v3((read3args *)rpcargs, dblock_scratch_RPC))==-1) {
				Dprintf(D_CACHE,"dcache_lookup: argument copy failed\n");
				return -1;
			}
			if((match=dtag_match_and_inv(dblock_scratch_RPC,
					 dblock_scratch_disk, 0))==-1) {
				Dprintf(D_CACHE,"dcache_lookup: tag match failed\n");
				return -1;
			}
			if (match == 1) { /* hit; copy data */
				if((response_from_disk_cache_read_v3(dblock_scratch_disk,
										 (read3args *)rpcargs,
										 (read3res *)res))==-1) {
					Dprintf(D_CACHE,"dcache_lookup: response copy failed\n");
					return -1;
				}
			}
			break;
		} /* switch (proc_index) */
		break;
	} /* switch(prog_index) */

	Dprintf(D_CACHE, "dcache lookup: (%i,%i) = %s\n", prog_index, proc_index,
		    match ? "HIT" : "MISS");
	return(match);
}

int dcache_update(int prog_index, int proc_index, void *rpcargs, void *res) {
	int match = 0;

	switch(prog_index) {
	case 2:
		switch(proc_index) {
		case NFSPROC_WRITE:
			if((argument_write_v2((writeargs *)rpcargs, dblock_scratch_RPC))==-1) {
				Dprintf(D_CACHE,"dcache_lookup for write: argument copy error\n");
				return -1;
			}
			/* copy current file attributes */
			dblock_scratch_RPC->contents.read_v2.attributes =
					((attrstat *)res)->attrstat_u.attributes;

			if((match = dtag_match_and_inv(dblock_scratch_RPC,
			dblock_scratch_disk, 0)) == -1) {
				Dprintf(D_CACHE,"dcache_lookup for write: tag match error\n");
				return -1;
			}

			if (match == 1) {
				if((response_from_disk_cache_write_v2(dblock_scratch_disk,
					(writeargs *)rpcargs,
					(attrstat *)res))==-1) {
					Dprintf(D_CACHE,"dcache_lookup: response copy error\n");
					return -1;
				}
			}
			break;
		} /* switch (proc_index) */
		break; /* case 2 */
	case 3:
		switch(proc_index) {
		case NFSPROC3_WRITE:
			if((argument_write_v3((write3args *)rpcargs, dblock_scratch_RPC))==-1) {
				Dprintf(D_CACHE,"dcache_lookup for write: argument copy error\n");
				return -1;
			}
			/* copy current file attributes */
			dblock_scratch_RPC->contents.read_v3.attributes = ((write3res *)res)->write3res_u.resok.file_wcc.after.post_op_attr_u.attributes;

			if((match = dtag_match_and_inv(dblock_scratch_RPC,
					dblock_scratch_disk, 0)) == -1) {
				Dprintf(D_CACHE,"dcache_lookup for write: tag match error\n");
				return -1;
			}

			if (match == 1) {
				if((response_from_disk_cache_write_v3(dblock_scratch_disk,
					(write3args *)rpcargs, (write3res *)res))==-1) {
					Dprintf(D_CACHE,"dcache_lookup: response copy error\n");
					return -1;
				}
			}
			break;
		} /* switch (proc_index) */
		break;
	} /* switch(prog_index) */

	Dprintf(D_CACHE, "dcache update: (%i,%i) = %s\n", prog_index, proc_index,
		    match ? "HIT" : "MISS");
	return(match);
}


/*************************************************************************
 * insert block in cache                                                 *
 * prog_index, proc_index: type of RPC call                              *
 * argument = (void *)rpcargs from incoming RPC call                     *
 * return values:                                                        *
 * 0 : inserted                                                          *
 *-1 : did not insert                                                    *
 * side-effects:                                                         *
 *  block_scratch_RPC contains copy of results                           *
 *  block_scratch_disk contains copy of results in case of a hit         *
 *************************************************************************/
int dcache_insert(void *argp, int prog_index, int proc_index, void *res) {
	readargs *readarg;
	readres *rdres;
	int base_offset,req_offset,req_count;
	int total_count;
	read3args *read3arg;
	read3res *rd3res;

	switch(prog_index) {
	case 2:
		rdres = (readres *)res;
		readarg = (readargs *)argp;
		/* calculate initial offset to fetch from */
		base_offset = (readarg->offset)-((readarg->offset)%MAX_NFS_BLOCKSIZE);
		/* calculate offset for building result structure */
		req_offset = (readarg->offset)%MAX_NFS_BLOCKSIZE;
		req_count = readarg->count;
		total_count = rdres->readres_u.reply.data.data_len;
		Dprintf(D_CACHE, "dcache_insert: offset %d, count %d\n",
			base_offset, total_count);

		switch(proc_index) {
		case NFSPROC_READ:
			/* add res. to cache_scratch_RPC*/
			if((response_remote_RPC_read_v2(rdres,
							dblock_scratch_RPC))==-1) {
				Dprintf(D_CACHE,"dcache_insert: RPC response copy failed\n");
				return -1;
			}
			/* prepare dblock_scratch_RPC for insertion. The request that
			 actually went out to the server was for base_offset + MAX_BLOCK;
			 and has returned a block of size total_count<=MAX_BLOCK */
			dblock_scratch_RPC->contents.read_v2.offset = base_offset;
			dblock_scratch_RPC->contents.read_v2.file = readarg->file;
			if((dinsert_in_cache(dblock_scratch_RPC, dblock_scratch_disk))==-1) {
				Dprintf(D_CACHE,"dcache_insert: insertion failed\n");
				return -1;
			}
			/* prepare result structure; we need to overwrite *res so that its
					first data matches the requested data, and the data length
					matches the requested count
					two situations can arise: req_count + offset is <= than the count
					returned by the server (total_count), in which case all requested
					bytes have been returned and we simply return req_count bytes.
					Or, the total bytes returned is not sufficient to cover
					req_offset -> req_offset_req_count.
					In this case, the total number of bytes returned by the server
					is total_count, but only (total_count-req_offset) are
					visible to the client
			 */
			if((req_count+req_offset) <= total_count)
				rdres->readres_u.reply.data.data_len = req_count;
			else if((total_count-req_offset) > 0)
				rdres->readres_u.reply.data.data_len = total_count-req_offset;
			else rdres->readres_u.reply.data.data_len = 0;

			if(rdres->readres_u.reply.data.data_len > 0)
				if((memcpy(&rdres->readres_u.reply.data.data_val[0],
					&dblock_scratch_RPC->contents.read_v2.data_val[req_offset],
					rdres->readres_u.reply.data.data_len))==NULL) {
					Dprintf(D_CACHE,"dcache_insert: memcpy error\n");
					return -1;
			}
			return 0;

		case NFSPROC_WRITE:
			/* add res. to cache_scratch_RPC*/
			if((response_remote_RPC_write_v2(rdres,
							dblock_scratch_RPC))==-1) {
				Dprintf(D_CACHE,"dcache_insert: RPC response copy error\n");
				return -1;
			}
			/* prepare dblock_scratch_RPC for insertion. The request that
			 actually went out to the server was for base_offset + MAX_BLOCK;
			 and has returned a block of size total_count<=MAX_BLOCK */
			dblock_scratch_RPC->contents.read_v2.offset = base_offset;
			dblock_scratch_RPC->contents.read_v2.file = readarg->file;
			if((dinsert_in_cache(dblock_scratch_RPC, dblock_scratch_disk))==-1) {
				Dprintf(D_CACHE,"dcache_insert: insertion error\n");
				return -1;
			}
			return 0;
		default:
			break;
		} /* switch(proc_index) */
		break;  /* case 2 */
	case 3:
		rd3res = (read3res *)res;
		read3arg = (read3args *)argp;
		/* calculate initial offset to fetch from */
		base_offset = (read3arg->offset)-((read3arg->offset)%MAX_NFS3_BLOCKSIZE);
		/* calculate offset for building result structure */
		req_offset = (read3arg->offset)%MAX_NFS3_BLOCKSIZE;
		req_count = read3arg->count;
		total_count = rd3res->read3res_u.resok.data.data_len;
		Dprintf(D_CACHE, "dcache_insert: offset %d, count %d\n",
			base_offset, total_count);

		switch(proc_index) {
		case NFSPROC3_READ:
			/* add res. to cache_scratch_RPC*/
			if((response_remote_RPC_read_v3(rd3res,
							dblock_scratch_RPC))==-1) {
				Dprintf(D_CACHE,"dcache_insert: RPC response copy failed\n");
				return -1;
			}

			dblock_scratch_RPC->contents.read_v3.offset = base_offset;
			dblock_scratch_RPC->contents.read_v3.file.length = read3arg->file.data.data_len;
			memcpy(dblock_scratch_RPC->contents.read_v3.file.handle, read3arg->file.data.data_val, read3arg->file.data.data_len);
			if((dinsert_in_cache(dblock_scratch_RPC, dblock_scratch_disk))==-1) {
				Dprintf(D_CACHE,"dcache_insert: insertion failed\n");
				return -1;
			}

			if((req_count+req_offset) <= total_count)
				rd3res->read3res_u.resok.data.data_len = req_count;
			else if((total_count-req_offset) > 0)
				rd3res->read3res_u.resok.data.data_len = total_count - req_offset;
			else rd3res->read3res_u.resok.data.data_len = 0;
				rd3res->read3res_u.resok.count = rd3res->read3res_u.resok.data.data_len;

			if(rd3res->read3res_u.resok.data.data_len > 0)
				if((memcpy(&rd3res->read3res_u.resok.data.data_val[0],
					&dblock_scratch_RPC->contents.read_v3.data_val[req_offset],
					rd3res->read3res_u.resok.data.data_len))==NULL) {
					Dprintf(D_CACHE,"dcache_insert: memcpy error\n");
					return -1;
			}
			return 0;
		case NFSPROC3_WRITE:
			/* add res. to cache_scratch_RPC*/
			if((response_remote_RPC_read_v3(rd3res,
							dblock_scratch_RPC))==-1) {
				Dprintf(D_CACHE,"dcache_insert: RPC response copy error\n");
				return -1;
			}
//      dblock_scratch_RPC->nsfprocid = NFSPROC3_WRITE;
			dblock_scratch_RPC->contents.read_v3.offset = base_offset;
			dblock_scratch_RPC->contents.read_v3.file.length = read3arg->file.data.data_len;
			memcpy(dblock_scratch_RPC->contents.read_v3.file.handle, read3arg->file.data.data_val, read3arg->file.data.data_len);
			if((dinsert_in_cache(dblock_scratch_RPC, dblock_scratch_disk))==-1) {
				Dprintf(D_CACHE,"dcache_insert: insertion error\n");
				return -1;
			}
			return 0;
		default:
			break;
		} /* switch(proc_index) */
		break;
	} /* switch(prog_index) */
	Dprintf(D_CACHE,"dcache_insert: unknown RPC call\n");
	return -1;
}

/*************************************************************************
 * Lookup entry for readarg *arg in data cache.
 * if found, invalidate it
 * return values:                                                        *
 * 0 : no invalidation needed                                            *
 * 1 : successful invalidation                                           *
 * -1: error                                                             *
 *************************************************************************/
int invalidate_readarg_v2(readargs *arg) {
	int match1 = 0, match2 = 0;
	readargs args;
	int crossed_boundary = 0;

	if(arg==NULL) {
		Dprintf(D_CACHE,"invalidate_readarg_v2: arg==NULL\n");
		return -1;
	}

	args = *arg;
	Dprintf(D_CACHE, "invalidate_readarg_v2: offset %d, count %d\n",
				args.offset, args.count);
	if(((args.offset % MAX_NFS_BLOCKSIZE) + args.count) >
						MAX_NFS_BLOCKSIZE) {
		Dprintf(D_CACHE, "invalidate_readarg: crossed block boundary\n");
		crossed_boundary = 1;
		args.offset = args.offset - (args.offset % MAX_NFS_BLOCKSIZE);
		args.count = MAX_NFS_BLOCKSIZE;
		args.totalcount = MAX_NFS_BLOCKSIZE;
	}
	/* copy diropargs to block_scratch_RPC */
	if((argument_read_v2(&args, dblock_scratch_RPC))==-1) {
		Dprintf(D_CACHE,"invalidate_readarg_v2: argument copy failed\n");
		return -1;
	}
	Dprintf(D_CACHE, "invalidate_readarg_v2: invalidate offset %d\n", args.offset);
	if((match1=dtag_match_and_inv(dblock_scratch_RPC,
						dblock_scratch_disk, 1))==-1) {
		Dprintf(D_CACHE,"invalidate_readarg_v2: tag invalidate failed\n");
		return -1;
	}

	if(crossed_boundary) { /* invalidate second block if this access crosses
								block boundary*/
		args.offset = arg->offset + arg->count;
		args.offset = args.offset - (args.offset % MAX_NFS_BLOCKSIZE);
		if((argument_read_v2(&args, dblock_scratch_RPC))==-1) {
			Dprintf(D_CACHE,"invalidate_readarg_v2: argument copy error\n");
			return -1;
		}
		Dprintf(D_CACHE, "invalidate: offset %d\n", args.offset);
		if((match2=dtag_match_and_inv(dblock_scratch_RPC,
				 dblock_scratch_disk, 1))==-1) {
			Dprintf(D_CACHE,"invalidate_readarg_v2: tag match error\n");
			return -1;
		}
	}
	Dprintf(D_CACHE,"invalidate_readarg_v2: %s\n", match1+match2 ? "CLEARED" : "NOT FOUND");
	return (match1+match2);
}

int invalidate_readarg_v3(read3args *arg) {
	int match1 = 0, match2 = 0;
	read3args args;
	int crossed_boundary = 0;

	if(arg==NULL) {
		Dprintf(D_CACHE,"invalidate_readarg_v3: arg==NULL\n");
		return -1;
	}

	args = *arg;
	Dprintf(D_CACHE, "invalidate_readarg_v3: offset %d, count %d\n",
				(int)args.offset, args.count);
	if(((args.offset % MAX_NFS3_BLOCKSIZE) + args.count) >
						MAX_NFS3_BLOCKSIZE) {
		Dprintf(D_CACHE, "invalidate_readarg_v3: crossed block boundary\n");
		crossed_boundary = 1;
		args.offset = args.offset - (args.offset % MAX_NFS3_BLOCKSIZE);
		args.count = MAX_NFS3_BLOCKSIZE;
	}
	/* copy diropargs to block_scratch_RPC */
	if((argument_read_v3(&args, dblock_scratch_RPC))==-1) {
		Dprintf(D_CACHE,"invalidate_readarg_v3: argument copy failed\n");
		return -1;
	}
	if((match1=dtag_match_and_inv(dblock_scratch_RPC,
						dblock_scratch_disk, 1))==-1) {
		Dprintf(D_CACHE,"invalidate_readarg_v3: tag invalidate failed\n");
		return -1;
	}

	if(crossed_boundary) { /* invalidate second block if this access crosses
								block boundary*/
		args.offset = arg->offset + arg->count;
		args.offset = args.offset - (args.offset % MAX_NFS3_BLOCKSIZE);
		if((argument_read_v3(&args, dblock_scratch_RPC))==-1) {
			Dprintf(D_CACHE,"invalidate_readarg_v3: argument copy error\n");
			return -1;
		}
		Dprintf(D_CACHE, "invalidate: offset %d\n", (int)args.offset);
		if((match2=dtag_match_and_inv(dblock_scratch_RPC,
				 dblock_scratch_disk, 1))==-1) {
			Dprintf(D_CACHE,"invalidate_readarg_v3: tag match error\n");
			return -1;
		}
	}
	Dprintf(D_CACHE,"invalidate_readarg_v3: %s\n", match1+match2 ? "CLEARED" : "NOT FOUND");
	return (match1+match2);
}

/*************************************************************************
 * check if need to invalidate block in cache                            *
 * prog_index, proc_index: type of RPC call                              *
 * argument = (void *)rpcargs from incoming RPC call                     *
 * return values:                                                        *
 * 0 : no invalidation needed                                            *
 * 1 : successful invalidation                                           *
 * -1: error                                                             *
 * side-effects:                                                         *
 * block_scratch_RPC contains copy of rpcargs                            *
 *************************************************************************/
int dcache_inval(int prog_index, int proc_index, void *rpcargs) {
	writeargs *wargs;
	readargs  invalidate_arg, *invalidate_arg_ptr;
	read3args invalidate_arg3;
	int invres=0;

	switch(prog_index) {
	case 2: /* NFS-V2 */
		switch(proc_index) {
		case NFSPROC_READ:
			invalidate_arg_ptr = (readargs *)rpcargs;
			break;
		case NFSPROC_WRITE:
			wargs = (writeargs *)rpcargs;
			invalidate_arg_ptr = &invalidate_arg;
			invalidate_arg_ptr->file = wargs->file;
			invalidate_arg_ptr->offset = wargs->offset;
			invalidate_arg_ptr->count = wargs->data.data_len;
			invalidate_arg_ptr->totalcount = wargs->data.data_len;
			break;
		default:
			/* no invalidations necessary for other calls */
			invalidate_arg_ptr = NULL;
		}
		/* now perform actual invalidation */
		if(invalidate_arg_ptr!=NULL) {
			invres = invalidate_readarg_v2(invalidate_arg_ptr);
			if(invres == -1) {
				Dprintf(D_CACHE,"Error invalidating data cache block\n");
				return -1;
			}
			return(invres>0);
		}
		break; /* case 2 */
	case 3: /* NFS-V3 */
		switch(proc_index) {
		case NFSPROC3_WRITE:
			invalidate_arg3.file = ((write3args *)rpcargs)->file;
			invalidate_arg3.offset = ((write3args *)rpcargs)->offset;
			invalidate_arg3.count = ((write3args *)rpcargs)->data.data_len;

			invres = invalidate_readarg_v3(&invalidate_arg3);
			if(invres == -1) {
				Dprintf(D_CACHE,"Error invalidating data cache block\n");
				return -1;
			}
			return(invres>0);

			break;
		}
		break; /* case 3 */
	default:
		return -1; /* unknown NFS version */
	}
	return -1;
}
