/***************************************************************************
 * Copyright (c) 2002-2005, In-VIGO Team
 * Advanced Computing and Information Systems (ACIS) Laboratory
 * University of Florida
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * cache_data.h
 *
 * Author: Renato J. Figueiredo, Ming Zhao
 *
 * Data structures for the gvfs proxy data cache.
 * The attribute cache is stored in the local disk.
 * The attribute cache is structured as follows:
 *
 * - The smallest allocation unit is a cacheblock
 *   - the cache_block is sufficiently large to hold the largest RPC
 *     reply that can be received by NFS, plus a tag, and state bits
 *   - The number of cache blocks is "cache_size"
 *
 * - Cache blocks are stored in file "banks" in the disk.
 *   - The name of a bank X in the disk is given by:
 *     BaseCachePath/SessionID/data_cache_bank_X
 *     (where BaseCachePath is common to multiple proxies, and SessionID
 *      is unique to each proxy instance)
 *   - The number of file banks is "cache_banks"
 *
 * - The associativity of the cache is given by "cache_assoc"
 *
 * - Example: a 4-way associative cache holding a max. of 32 attributes,
 *   with two file banks. This cache would consist of files data_cache_bank_0
 *   and data_cache_bank_1; each bank would have 4 sets, each set having
 *   4 blocks.
 *
 * Updated: Jan., 2004
 *          Added data structures for recording special types of file handles;
 *          Added data structures for recording dirty cache blocks.
 ***************************************************************************/

#ifndef CACHE_DATA_H
#define CACHE_DATA_H

#include "cache_attrib.h"

/* default cache parameters */
#define DEFAULT_DATA_CACHE_SIZE 65536
#define DEFAULT_DATA_CACHE_ASSOC 16
#define DEFAULT_DATA_CACHE_BANKS 128

/* special types of file handles */
#define FH_HASH_SIZE 1024
#define FH_NULL 0
#define FH_STALE 1
#define FH_DIRTY 2
#define FH_TRUNC 3

#define ACACHE_DIRTY_BLOCK 0
#define DCACHE_DIRTY_BLOCK 1

/* NFS block size supported */
#define MAX_NFS_BLOCKSIZE 8192
#define MAX_NFS3_BLOCKSIZE 32768

/* The following defines concern with the hash function */
/* How many consecutive NFS blocks will map to a same set in
   cache. Must be careful in choosing this parameter in
   conjunction with associativity; should not allow associativity
   to be smaller than this value for good performance */
#define HASH_CONSECUTIVE_BLOCKS_SET_LOG2 3
#define HASH_CONSECUTIVE_BLOCKS_SET 8
#define HASH_CONSECUTIVE_SETS_LOG2 6
#define HASH_CONSECUTIVE_SETS 64
#define HASH_FUNCTION_TO_USE 2

/*
 * Data structures
 */

struct dcacheblock_read_v2 {
  /* store contents of a NFS-V2 read transaction */
  /* storage is static (no mallocs); need to improve this by
     allowing max NFS read sizes to be configured via command line,
     otherwise will waste a lot of space */

  /* TAG */
  /* readargs */
  nfs_fh   file;
  u_int    offset;
  u_int    count;
  u_int    totalcount;

  /* result */
  /* readokres  */
  fattr attributes;
  u_int data_len;
  char data_val[MAX_NFS_BLOCKSIZE];
  /* readres */
  nfsstat status;
};
typedef struct dcacheblock_read_v2 dcacheblock_read_v2;

typedef struct dcacheblock_read_v3 {
  /* store contents of a NFS-V3 read transaction */
  /* storage is static (no mallocs); need to improve this by
     allowing max NFS read sizes to be configured via command line,
     otherwise will waste a lot of space */

  /* TAG */
  /* readargs */
  nfs3_fh   file;
  u_int    offset;
  u_int    count;

  /* result */
  /* readokres  */
  fattr3 attributes;
  bool_t eof;
  u_int data_len;
  char data_val[MAX_NFS3_BLOCKSIZE];
  /* readres */
  nfsstat3 status;
} dcacheblock_read_v3;

struct dirty_block {
  int id;
  int nfsprogid;
  int nfsprocid;
  union {
    nfs_fh handle;
    nfs3_fh handle3;
  } file;
  u_int offset;

  int seekset;
  struct dirty_block *next;
};
typedef struct dirty_block dirty_block;

struct fh_node {
  int prog;
  union {
    nfs_fh handle;
    nfs3_fh handle3;
  } file;
  int status;
  char path[PATH_MAX];
  struct fh_node *next;
};
typedef struct fh_node fh_node;

struct dcacheblock {  /* stores a cache entry for NFS transactions */

  int state;               /* block's state */
  int nfsprogid;           /* NFS program version for this block */
  int nfsprocid;           /* NFS procedure number for this block */

  union {
                           /* contents; depend on transaction type */
    dcacheblock_read_v2 read_v2;
    dcacheblock_read_v3 read_v3;
  } contents;
};
typedef struct dcacheblock dcacheblock;

/* The top-level functions of cache_data.c */

/* Lookup a block in disk cache, for a given NFS RPC call (prog_index,
   proc_index) and arguments (rpcargs). Returns hit (1) or miss (0); if
   an error occurs, returns -1. In case of a hit, return RPC results
   are stored in *res */
extern int dcache_lookup(int prog_index, int proc_index,
			 void *rpcargs, void *res);

extern int dcache_update(int prog_index, int proc_index,
			 void *rpcargs, void *res);

/* Insert results *res received from a given NFS RPC call (prog_index,
   proc_index) into the disk cache (creating a new cache bank in disk
   if necessary). Returns 0 on success, -1 on error. */
extern int dcache_insert(void *argp, int prog_index, int proc_index,
			 void *res);

/* Check and perform necessary invalidations of attrib. cache blocks */
extern int dcache_inval(int prog_index, int proc_index, void *rpcargs);

/* Initialize attribute cache parameters and base path directory
   Returns 0 if successfull, -1 if an error occurred */
extern int init_data_cache(unsigned long int size, int assoc,
			   int banks, char *basepath, char *sessionid);
extern int flush_data_cache(void);
extern int fh_hash_insert(fh_node *hash[],fh_node *node);
extern int flush_dirty_blocks(void);
extern fh_node *stale_fh_hash[FH_HASH_SIZE];
extern int invalidate_diroparg_v2(diropargs *arg);
extern int invalidate_nfs_fh_v2(nfs_fh *arg);

#endif /* CACHE_DATA_H */
