/*
 * invalidation.c	This module implemented client polling invalidation.
 *
 * Author:	Ming Zhao
 */

#include "invalidation.h"
#include "sk.h"

#define PAGESIZE 8192 /* size of a page for invalidation list */
#define CACHESIZE 10 /* size of file handle cache */

inv_node *	inval_list = NULL;
int list_size = 0;
int inval_index = 0;
int inval_base = 0;
u_int inval_tstamp = 0;
uint32 getinv_calls = 0;
inv_cache_ent cached_fh[MAX_CLIENT_NUM*CACHESIZE];
int cache_next[MAX_CLIENT_NUM];
u_int client_tstamps[MAX_CLIENT_NUM];
int sync_interval = 30; /* defines client polling interval */
int inv_min_interval = 1;
int inv_max_interval = 0;

extern int inval_enabled;
extern int attr_cache_enabled;
extern int data_cache_enabled;
extern int refresh_authentication_handle_uid;
extern int refresh_authentication_handle_gid;
extern CLIENT *nfsdclient3;

int inval_init() {
	 if((inval_list = (inv_node *)malloc(PAGESIZE*sizeof(inv_node))) == NULL) {
		 fprintf(stderr, "error: cannot allocate memory\n");
		 return -1;
	 }
	 inval_index = 0;
	 inval_base = 0;
	 list_size = PAGESIZE;
	 bzero(cached_fh, MAX_CLIENT_NUM*CACHESIZE*sizeof(inv_cache_ent));
	 bzero(cache_next, MAX_CLIENT_NUM*sizeof(int));
	 bzero(client_tstamps, MAX_CLIENT_NUM*sizeof(int));
	 Dprintf(D_CACHE, "initialize invalidation data structures succesfully\n");
	 return 0;
}

void inval_free() {
	int i;

	for(i=0; i<PAGESIZE; i++) {
		if(inval_list[i].flag) free(inval_list[i].info.attr.file.data.data_val);
		else {
			free(inval_list[i].info.diro.dir.data.data_val);
			free(inval_list[i].info.diro.name);
		}
	}
	free(inval_list);
}

int inval_insert(nfs_fh3 file, char *name, int client_id) {
	Dprintf(D_CACHE, "client_id:%d\n", client_id);

	if(inval_index == list_size) {
		if(addapage() != 0) return -1;
	}

	if(name != NULL) {
		Dprintf(D_CACHE, "insert diro into invalidation list at %d+%d\n",
		inval_index, inval_base);
		inval_list[inval_index].flag = 0;
		inval_list[inval_index].info.diro.dir.data.data_len = file.data.data_len;
		inval_list[inval_index].info.diro.dir.data.data_val = (char *)malloc(file.data.data_len);
		memcpy(inval_list[inval_index].info.diro.dir.data.data_val, file.data.data_val,
		       file.data.data_len);
		inval_list[inval_index].info.diro.name = (char *)malloc(strlen(name) + 1);
		strcpy(inval_list[inval_index].info.diro.name, name);
		inval_list[inval_index].client_id = client_id;

		inval_index ++;

		hexdump(inval_list[inval_index-1].info.diro.dir.data.data_val, inval_list[inval_index-1].info.diro.dir.data.data_len, D_CACHE);
		Dprintf(D_CACHE, "%s\n", inval_list[inval_index-1].info.diro.name);
	} else if(inval_index == 0 || lookup_fh_cache(file, client_id) == 0) {
		Dprintf(D_CACHE, "insert attr into invalidation list at %d+%d\n",
		inval_index, inval_base);
		inval_list[inval_index].flag = 1;
		inval_list[inval_index].client_id = client_id;
		inval_list[inval_index].info.attr.file.data.data_len = file.data.data_len;
		inval_list[inval_index].info.attr.file.data.data_val = (char *)malloc(file.data.data_len);
		memcpy(inval_list[inval_index].info.attr.file.data.data_val, file.data.data_val,
		       file.data.data_len);

		memset(cached_fh[(client_id-1)*CACHESIZE+cache_next[client_id-1]].file,
		       0, NFS3_FHSIZE);
		memcpy(cached_fh[(client_id-1)*CACHESIZE+cache_next[client_id-1]].file,
		       file.data.data_val, file.data.data_len);
		cached_fh[(client_id-1)*CACHESIZE+cache_next[client_id-1]].tstamp =
		inval_index+inval_base;
		cache_next[client_id-1] = (cache_next[client_id-1]+1) % CACHESIZE;

		inval_index ++;

		hexdump(inval_list[inval_index-1].info.attr.file.data.data_val, file.data.data_len, D_CACHE);
	}

	 return 0;
}

int lookup_fh_cache(nfs_fh3 file, int client_id) {
	int i;
	u_int max_tstamp=0;

	for(i=0; i<sk_no; i++)
		if(client_tstamps[i] > max_tstamp) max_tstamp = client_tstamps[i];

	for(i=0; i<CACHESIZE; i++) {
	if(cached_fh[(client_id-1)*CACHESIZE+i].tstamp != 0 &&
		max_tstamp-1 <= cached_fh[(client_id-1)*CACHESIZE+i].tstamp)
		if(memcmp(file.data.data_val, cached_fh[(client_id-1)*CACHESIZE+i].file, file.data.data_len) == 0) {
		Dprintf(D_CACHE, "found the file handle in cache\n");
		return 1;
		}
	}

	return 0;
}

int addapage() {
	inv_node *p;
	int i, flag=0;

	Dprintf(D_CACHE, "add a page to invalidation list\n");

	for(i=0; i<sk_no; i++) {
		Dprintf(D_CACHE, "client %d's timestamp: %u, inval_base: %d\n",
					i+1, client_tstamps[i], inval_base);
		if(client_tstamps[i] != 0 && client_tstamps[i]-1 < inval_base + PAGESIZE) {
			flag = 1;
			break;
		}
	}
	if(flag == 0) {
		Dprintf(D_CACHE, "a page can be released\n");
		if(list_size > PAGESIZE) {
			if((p = (inv_node *)malloc(list_size*sizeof(inv_node))) == NULL) {
				Dprintf(L_ERROR, "allocate memory error\n");
				return -1;
			}
			if(memcpy(p, &inval_list[PAGESIZE], (list_size-PAGESIZE)*sizeof(inv_node)) == NULL) {
				Dprintf(L_ERROR, "memory copy error\n");
				return -1;
			}
			inval_free();
			inval_list = p;
		}
		inval_base += PAGESIZE;
		inval_index -= PAGESIZE;
	} else {
		if((p = (inv_node *)malloc((list_size+PAGESIZE)*sizeof(inv_node))) == NULL) {
			Dprintf(L_ERROR, "allocate memory error\n");
			return -1;
		}
		if(memcpy(p, inval_list, list_size*sizeof(inv_node)) == NULL) {
			Dprintf(L_ERROR, "memory copy error\n");
			return -1;
		}
		free(inval_list);
		inval_list = p;
		list_size += PAGESIZE;
	}

	Dprintf(D_CACHE, "current list size = %d, inval_base = %d, inval_index = %d\n",
			list_size, inval_base, inval_index);

	return 0;
}

int getinv()
{
	u_int data_index, invres;
	read3res res;
	nfs_fh3 file;
	diropargs3 dir_arg;
	char machname[1024];
	uid_t uid = 0;
	gid_t gid = 0;
	gid_t gids[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	int len = 0;
	char data_buf[MAX_NFS3_BLOCKSIZE];
	res.read3res_u.resok.data.data_val = data_buf;

	if((gethostname(machname,1024))==-1) {
		Dprintf(L_ERROR, "cannot obtain host name\n");
		inval_enabled = 0;
		return -1;
	}
	sk_pack(machname);
	refresh_authentication_handle_uid = uid;
	refresh_authentication_handle_gid = gid;
	auth_destroy(nfsdclient3->cl_auth);
	nfsdclient3->cl_auth = authunix_create(machname, uid, gid, len, gids);

call:
	Dprintf(D_CACHE, "*** GETINV: %d\n", getinv_calls);
	if (((nfsproc3_getinv_3(&inval_tstamp, &res, nfsdclient3)) != RPC_SUCCESS) || (res.status != NFS3_OK && res.status != 3)) {
		Dprintf(D_CACHE, "call getinv error (%d)\n", res.status);
		inval_enabled = 0;
		return -1;
	}
	Dprintf(D_CACHE, "timestamp: %d->%d\n", inval_tstamp, res.read3res_u.resok.file_attributes.post_op_attr_u.attributes.size);
	if(inval_tstamp == 0) {
		inval_tstamp = res.read3res_u.resok.file_attributes.post_op_attr_u.attributes.size;
		if(inv_max_interval > 0) sync_interval = inv_min_interval;
		goto out;
	}
	getinv_calls ++;
	if(res.status == 3) {
		Dprintf(D_CACHE, "reinitialize ... \n");
		inval_tstamp = 0;
		flush_attr_cache();
	}

	data_index = 0;
	while(data_index < res.read3res_u.resok.data.data_len) {
		Dprintf(D_CACHE, "%d %c\n", data_index, res.read3res_u.resok.data.data_val[data_index]);
		if(res.read3res_u.resok.data.data_val[data_index++] == 'g') {
			memcpy(&file.data.data_len,
										&res.read3res_u.resok.data.data_val[data_index],
										sizeof(u_int));
			data_index += sizeof(u_int);
			file.data.data_val = &res.read3res_u.resok.data.data_val[data_index];
			hexdump(file.data.data_val, file.data.data_len, D_CACHE);
			invres = invalidate_nfs_fh_v3(&file);
			if(invres == -1) {
				Dprintf(D_CACHE,"Error invalidating attribute cache block\n");
			}
			data_index += file.data.data_len;
		} else {
			memcpy(&dir_arg.dir.data.data_len,
			       &res.read3res_u.resok.data.data_val[data_index], sizeof(u_int));
			data_index += sizeof(u_int);
			dir_arg.dir.data.data_val = &res.read3res_u.resok.data.data_val[data_index];
			hexdump(dir_arg.dir.data.data_val, dir_arg.dir.data.data_len, D_CACHE);
			data_index += dir_arg.dir.data.data_len;
			dir_arg.name = &res.read3res_u.resok.data.data_val[data_index];
			Dprintf(D_CACHE, "%s\n", dir_arg.name);
			invres = invalidate_diroparg_v3(&dir_arg);
			if(invres == -1) {
				Dprintf(D_CACHE,"Error invalidating attribute cache block\n");
			}
			data_index += strlen(dir_arg.name) + 1;
		}
	}

	inval_tstamp = res.read3res_u.resok.file_attributes.post_op_attr_u.attributes.size;
	Dprintf(D_CACHE, "call getinv: done, timestamp: %d\n", inval_tstamp);
	if(res.read3res_u.resok.file_attributes.post_op_attr_u.attributes.mode == 1)
		goto call;

	if(inv_max_interval > 0) {
		if(data_index > 0) sync_interval = inv_min_interval;
		else if(inv_max_interval >= sync_interval*2) sync_interval *= 2;
		else sync_interval = inv_max_interval;
	}

out:
	Dprintf(D_CACHE, "setup alarm(%d)\n", sync_interval);
	alarm(sync_interval);

	return 0;
}

