/*
 * mountd	This program handles RPC "NFS" mount requests. for
 *              versions 1, 2, and 3 of the NFS mount protocol.
 *
 * Renato J. Figueiredo. This code is mostly based on the original
 *                       UNFSD code by:
 *
 * Usage:	[rpc.]mountd [-dhnpv] [-f authfile]
 *
 * Authors:	Mark A. Shand, May 1988
 *		Donald J. Becker, <becker@super.org>
 *		Rick Sladkey, <jrs@world.std.com>
 *		Fred N. van Kempen, <waltje@uWalt.NL.Mugnet.ORG>
 *		Olaf Kirch, <okir@monad.swb.de>
 *
 *		Copyright 1988 Mark A. Shand
 *		This software maybe be used for any purpose provided
 *		the above copyright notice is retained.  It is supplied
 *		as is, with no warranty expressed or implied.
 *
 * Updated:	Corrected the problem that error code is not returned in
 *		several functions; when the RPC call to server fails, the
 *		proxy returns a NFSERR_ACCES to client.
 *		Ming June, 2005
 */

 /*

    WANT_LOG_MOUNTS guarded patch is added Aug 18, 1995 by Alex Yuriev
    		   (alex@bach.cis.temple.edu), CIS Laboratories,
    		   TEMPLE UNIVERSITY, USA

    This version is ugly and can be really improved but it actually
    pinpointed a couple of intruders :)

*/

#include "mountd.h"
#include "getopt.h"
#include "rpcmisc.h"
#include "signals.h"
#include "sk.h"
#include <rpc/pmap_clnt.h>

#ifdef SOLARIS
#define PORTMAP
#include <rpc/rpc.h>
#include <rpc/svc_soc.h>
#endif /* SOLARIS */

#ifndef svc_getcaller
#define svc_getcaller(x) ((struct sockaddr_in *) (x)->xp_rtaddr.buf)
#endif

static void	usage(FILE *, int);
static void	terminate(void);
static RETSIGTYPE sigterm(int sig);

/*
 * Option table for mountd
 */
static struct option longopts[] =
{
      { "debug",		required_argument,	0,	'd' },
      { "foreground",           0,                      0,      'F' },
      { "exports-file",		required_argument,	0,	'f' },
      { "help",			0,			0,	'h' },
      { "allow-non-root",	0,			0,	'n' },
      { "host-server",          required_argument,      0,      'H' },
      { "portin",               required_argument,      0,      'P' },
      { "portout",              required_argument,      0,      'O' },
      { "sock",                 required_argument,      0,      'S' },
      { "transport",            required_argument,      0,      't' },
      { "mountversion",      	required_argument,	0,	'V' },
      { "cookiefile",      	required_argument,	0,	'C' },
      { "promiscous",		0,			0,	'p' },
      { "sessionkeysend",	0,			0,	's' },
      { "sessionkeymatch",	0,			0,	'm' },
      { "version",		0,			0,	'v' },
      { NULL,			0,			0,	0 }
};
static const char *	shortopts = "d:Ff:hnH:P:O:S:t:V:C:ps:m:v::";

CLIENT *mountdclient1;
CLIENT *mountdclient2;
CLIENT *authclient1;
CLIENT *authclient2;
CLIENT *mountdclient3;
CLIENT *authclient3;
int refresh_authentication_handle_uid = 0;
int refresh_authentication_handle_gid = 0;

int mountvers = 2;
int authvers;

char		argbuf[MNTPATHLEN + 1];
char		*auth_file = NULL;
static char	*program_name;
int		need_reinit = 0;
int		need_flush = 0;
extern char	version[];

extern CLIENT *open_mountdclient(char *, int, int, int, int);
extern void close_mountdclient(CLIENT *);

/***************************************************************/
/*********************** V1 HANDLERS ***************************/
/***************************************************************/

/*
 * NULL
 * Do nothing
 */
void *
mountproc_null_1_svc(void *argp, struct svc_req *rqstp)
{
  return ((void *) &result);
}

/*
 * MOUNT
 * This is what the whole protocol is all about
 */
fhstatus *
mountproc_mnt_1_svc(dirpath *argp, struct svc_req *rqstp)
{
  fhstatus	*res;
  nfs_client	*cp;
  nfs_mount	*mp;
#ifdef WANT_LOG_MOUNTS
  struct in_addr	addr;
#endif	/* WANT_LOG_MOUNTS */

  res = (struct fhstatus *)&result;

  if (**argp == '\0') {
    strcpy(argbuf, "/");
  } else {
    /* don't trust librpc */
    strncpy(argbuf, *argp, MNTPATHLEN);
    argbuf[MNTPATHLEN] = '\0';
  }

#ifdef WANT_LOG_MOUNTS
  addr = svc_getcaller(rqstp->rq_xprt)->sin_addr;
  Dprintf(D_CALL, "NFSv1 mount of %s attempted from %s\n",
	  argbuf, inet_ntoa(addr));
#endif /* WANT_LOG_MOUNTS */

  /* Now authenticate the intruder... */
  if (((cp = auth_clnt(rqstp)) == NULL)
      || (mp = auth_path(cp, rqstp, argbuf)) == NULL
      || mp->o.noaccess
      || (mp->o.secure_port &&
	  !SECURE_PORT(svc_getcaller(rqstp->rq_xprt)->sin_port))) {
    res->fhs_status = NFSERR_ACCES;
#ifdef WANT_LOG_MOUNTS
    Dprintf(D_AUTH, "Blocked attempt of %s to mount %s\n",
	    inet_ntoa(addr), argbuf);
#endif /* WANT_LOG_MOUNTS */
    Dprintf (D_CALL, "\tmount res = %d\n", res->fhs_status);
    return (res);
  }

#ifdef WANT_LOG_MOUNTS
  Dprintf(D_CALL, "%s has been mounted by %s\n",
	  argbuf, inet_ntoa(addr));
#endif /* WANT_LOG_MOUNTS */
  Dprintf (D_CALL, "\tmount res = %d\n", res->fhs_status);

  /* After all proxy authentication takes place, perform
     actual request forwarding to native client */

  if ((res = mountproc_mnt_1(argp, mountdclient1)) == NULL) {
    clnt_perror(mountdclient1, "mountproc_mnt_1");
    result.fstatus.fhs_status = NFSERR_ACCES;
    return(&result.fstatus);
  }

  /* Return file handle obtained from native server */
  result.fstatus = *res;
  return(&result.fstatus);
}

/*
 * DUMP
 * Dump the contents of rmtab on the caller. We don't keep rmtabs;
 * return NULL
 */
mountlist *
mountproc_dump_1_svc(void *argp, struct svc_req *rqstp)
{
  return (NULL);
}

/*
 * UMNTALL
 * Remove a mounted fs's rmtab entry, FWIW.
 */
void *
mountproc_umnt_1_svc(dirpath *argp, struct svc_req *rqstp)
{
  return ((void*) &result);
}

/*
 * UMNTALL
 * Remove a client's rmtab entry.
 */
void *
mountproc_umntall_1_svc(void *argp, struct svc_req *rqstp)
{
  return ((void*) &result);
}

/*
 * EXPORT
 * Return list of all exported file systems.
 */
exports *
mountproc_export_1_svc(void *argp, struct svc_req *rqstp)
{
  return (&export_list);
}

/*
 * EXPORTALL
 * Same as EXPORT
 */
exports *
mountproc_exportall_1_svc(void *argp, struct svc_req *rqstp)
{
  return (&export_list);
}

/***************************************************************/
/*********************** V2 HANDLERS ***************************/
/***************************************************************/

/*
 * NULL
 * Do nothing
 */
void *
mountproc_null_2_svc(void *argp, struct svc_req *rqstp)
{
  return ((void *) &result);
}

/*
 * MOUNT
 * This is what the whole protocol is all about
 */
fhstatus *
mountproc_mnt_2_svc(dirpath *argp, struct svc_req *rqstp)
{
  fhstatus	*res;
  nfs_client	*cp;
  nfs_mount	*mp;
#ifdef WANT_LOG_MOUNTS
  struct in_addr	addr;
#endif	/* WANT_LOG_MOUNTS */
  struct authunix_parms *unix_cred;
  char machname[1024];
  uid_t uid;
  gid_t gid;
  gid_t gids[16];
  int len;

  res = (struct fhstatus *)&result;

  if (**argp == '\0') {
    strcpy(argbuf, "/");
  } else {
    /* don't trust librpc */
    strncpy(argbuf, *argp, MNTPATHLEN);
    argbuf[MNTPATHLEN] = '\0';
  }

#ifdef WANT_LOG_MOUNTS
  addr = svc_getcaller(rqstp->rq_xprt)->sin_addr;
  Dprintf(D_CALL, "NFSv2 mount of %s attempted from %s\n",
	  argbuf, inet_ntoa(addr));
#endif /* WANT_LOG_MOUNTS */

  /* Now authenticate the intruder... */
  if (((cp = auth_clnt(rqstp)) == NULL)
      || (mp = auth_path(cp, rqstp, argbuf)) == NULL
      || (match_sk_enabled && auth_sk(rqstp)==0)
      || mp->o.noaccess) {
    res->fhs_status = NFSERR_ACCES;
#ifdef WANT_LOG_MOUNTS
    Dprintf(D_AUTH, "Blocked attempt of %s to mount %s\n",
	    inet_ntoa(addr), argbuf);
#endif /* WANT_LOG_MOUNTS */
    Dprintf (D_CALL, "\tmount res = %d\n", res->fhs_status);
    return (res);
  }

#ifdef WANT_LOG_MOUNTS
  Dprintf(D_CALL, "%s has been mounted by %s\n",
	  argbuf, inet_ntoa(addr));
#endif /* WANT_LOG_MOUNTS */

  Dprintf (D_CALL, "\tmount res = %d\n", res->fhs_status);

  if(send_sk_enabled) {
    unix_cred = (struct authunix_parms *) rqstp->rq_clntcred;

    uid  = unix_cred->aup_uid;
    gid  = unix_cred->aup_gid;
    len  = unix_cred->aup_len;
    memcpy(gids, unix_cred->aup_gids, len*sizeof(gid_t));

    if((gethostname(machname,1024))==-1) {
      Dprintf(L_ERROR, "create authenticator: cannot obtain host name\n");
      result.fstatus.fhs_status = NFSERR_ACCES;
      return(&result.fstatus);
    }

    if(sk_pack(machname) == -1) {
      result.fstatus.fhs_status = NFSERR_ACCES;
      return(&result.fstatus);
    }

    auth_destroy(authclient2->cl_auth);
    authclient2->cl_auth = authunix_create(machname, uid, gid, len, gids);
  }

  /* After all proxy authentication takes place, perform
     actual request forwarding to native client */

  if ((res = mountproc_mnt_2(argp, mountdclient2)) == NULL) {
    clnt_perror(mountdclient2, "mountproc_mnt_2");
    /* Modified by vineet, Ming */
    result.fstatus.fhs_status = NFSERR_ACCES;
    return(&result.fstatus);
  }

  /* Return file handle obtained from native server */
  result.fstatus = *res;
  return(&result.fstatus);
}

/*
 * DUMP
 * Dump the contents of rmtab on the caller.
 */
mountlist *
mountproc_dump_2_svc(void *argp, struct svc_req *rqstp)
{
  return (NULL);
}

/*
 * UMNTALL
 * Remove a mounted fs's rmtab entry, FWIW.
 */
void *
mountproc_umnt_2_svc(dirpath *argp, struct svc_req *rqstp)
{
  return ((void*) &result);
}

/*
 * UMNTALL
 * Remove a client's rmtab entry.
 */
void *
mountproc_umntall_2_svc(void *argp, struct svc_req *rqstp)
{
  return ((void*) &result);
}

/*
 * EXPORT
 * Return list of all exported file systems.
 */
exports *
mountproc_export_2_svc(void *argp, struct svc_req *rqstp)
{
  return (&export_list);
}

/*
 * EXPORTALL
 * Same as EXPORT
 */
exports *
mountproc_exportall_2_svc(void *argp, struct svc_req *rqstp)
{
  return (&export_list);
}

/*
 * PATHCONF
 * Since the protocol doesn't include a status field, Sun apparently
 * considers it good practice to let anyone snoop on your system, even if
 * it's pretty harmless data such as pathconf. We don't.
 *
 * Besides, many of the pathconf values don't make much sense on NFS volumes.
 * FIFOs and tty device files represent devices on the *client*, so there's
 * no point in getting the *server's* buffer sizes etc. Wonder what made the
 * Sun people choose these.
 */
ppathcnf *
mountproc_pathconf_2_svc(dirpath *argp, struct svc_req *rqstp)
{
  ppathcnf	*res = (ppathcnf *) &result;
  nfs_client	*cp;
  nfs_mount	*mp;
#ifdef WANT_LOG_MOUNTS
  struct in_addr	addr;
#endif	/* WANT_LOG_MOUNTS */

  memset(res, 0, sizeof(*res));

  if (**argp == '\0') {
    strcpy(argbuf, "/");
  } else {
    /* don't trust librpc */
    strncpy(argbuf, *argp, MNTPATHLEN);
    argbuf[MNTPATHLEN] = '\0';
  }

  /* Now authenticate the intruder... */
  if (((cp = auth_clnt(rqstp)) == NULL)
      || (mp = auth_path(cp, rqstp, argbuf)) == NULL
      || mp->o.noaccess) {
#ifdef WANT_LOG_MOUNTS
    Dprintf(D_AUTH, "Blocked attempt of %s to pathconf(%s)\n",
	    inet_ntoa(addr), dir);
#endif /* WANT_LOG_MOUNTS */
  }

  /* After all proxy authentication takes place, perform
     actual request forwarding to native client */

  if ((res = mountproc_pathconf_2(argp, mountdclient2)) == NULL) {
    clnt_perror(mountdclient2, "mountproc_pathcnf_2");
  }

  /* Return result obtained from native server */
  result.pathconf = *res;
  return(&result.pathconf);
}


/***************************************************************/
/*********************** V3 HANDLERS ***************************/
/***************************************************************/

/*
 * NULL
 * Do nothing
 */
void *
mountproc3_null_3_svc(void *argp, struct svc_req *rqstp)
{
  return ((void *) &result);
}

/*
 * MOUNT
 * This is what the whole protocol is all about
 */
mountres3 *
mountproc3_mnt_3_svc(dirpath *argp, struct svc_req *rqstp)
{
  mountres3	*res;
  nfs_client	*cp;
  nfs_mount	*mp;
#ifdef WANT_LOG_MOUNTS
  struct in_addr	addr;
#endif	/* WANT_LOG_MOUNTS */

  res = (struct mountres3 *)&result;

  if (**argp == '\0') {
    strcpy(argbuf, "/");
  } else {
    /* don't trust librpc */
    strncpy(argbuf, *argp, MNTPATHLEN);
    argbuf[MNTPATHLEN] = '\0';
  }

#ifdef WANT_LOG_MOUNTS
  addr = svc_getcaller(rqstp->rq_xprt)->sin_addr;
  Dprintf(D_CALL, "NFSv3 mount of %s attempted from %s\n",
	  argbuf, inet_ntoa(addr));
#endif /* WANT_LOG_MOUNTS */

  /* Now authenticate the intruder... */
  if (((cp = auth_clnt(rqstp)) == NULL)
      || (mp = auth_path(cp, rqstp, argbuf)) == NULL
      || mp->o.noaccess
      || (mp->o.secure_port &&
	  !SECURE_PORT(svc_getcaller(rqstp->rq_xprt)->sin_port))) {
    res->fhs_status = NFSERR_ACCES;
#ifdef WANT_LOG_MOUNTS
    Dprintf(D_AUTH, "Blocked attempt of %s to mount %s\n",
	    inet_ntoa(addr), argbuf);
#endif /* WANT_LOG_MOUNTS */
    Dprintf (D_CALL, "\tmount res = %d\n", res->fhs_status);
    return (res);
  }

#ifdef WANT_LOG_MOUNTS
  Dprintf(D_CALL, "%s has been mounted by %s\n",
	  argbuf, inet_ntoa(addr));
#endif /* WANT_LOG_MOUNTS */

  /* After all proxy authentication takes place, perform
     actual request forwarding to native client */

  if ((res = mountproc3_mnt_3(argp, mountdclient3)) == NULL) {
    clnt_perror(mountdclient3, "mountproc_mnt_3");
    result.mountres3.fhs_status = NFSERR_ACCES;
    return(&result.mountres3);
  }

  Dprintf (D_CALL, "\tmount res = %d\n", res->fhs_status);
  /* Return file handle obtained from native server */
  result.mountres3 = *res;
  return(&result.mountres3);
}

/*
 * DUMP
 * Dump the contents of rmtab on the caller.
 */
mountlist *
mountproc3_dump_3_svc(void *argp, struct svc_req *rqstp)
{
  return (NULL);
}

/*
 * UMNTALL
 * Remove a mounted fs's rmtab entry, FWIW.
 */
void *
mountproc3_umnt_3_svc(dirpath *argp, struct svc_req *rqstp)
{
  return ((void*) &result);
}

/*
 * UMNTALL
 * Remove a client's rmtab entry.
 */
void *
mountproc3_umntall_3_svc(void *argp, struct svc_req *rqstp)
{
  return ((void*) &result);
}

/*
 * EXPORT
 * Return list of all exported file systems.
 */
exports *
mountproc3_export_3_svc(void *argp, struct svc_req *rqstp)
{
  return (&export_list);
}

int
main(int argc, char **argv)
{
  int foreground = 0;
  int c;
  int     nfsportin = 0;
  int     nfsportout = 0;
  int     rootsock = 0;
  int     transport = NFS_OVER_UDP;
  char  localhostname[HOSTNAMESIZE];
  char  *cookiefile = NULL;
  char  *send_sk_path = NULL;
  char  *match_sk_path = NULL;

  program_name = argv[0];
  localhostname[0] = '\x0';

  /* Parse the command line options and arguments. */
  opterr = 0;
  while ((c = getopt_long(argc, argv, shortopts, longopts, NULL)) >= 0 )
    switch (c) {
    case 'F':
      foreground = 1;
      break;
    case 'h':
      usage(stdout, 0);
      break;
    case 'd':
      enable_logging(optarg);
      break;
    case 'f':
      auth_file = optarg;
      break;
    case 'n':
      allow_non_root = 1;
      break;
    case 'H':
      strncpy(&localhostname[0], optarg, HOSTNAMESIZE);
      break;
    case 'C':
#ifndef COOKIE_AUTH
      fprintf(stderr, "Cookie authentication has not been"
	      " enabled in this executable.\n");
      exit(-1);
#endif /* COOKIE_AUTH */
      cookiefile = optarg;
      break;
    case 'P':
      nfsportin = atoi(optarg);
      if (nfsportin < 0) {
	fprintf(stderr,
		"mountd: bad port number: %s\n",
		optarg);
	usage(stderr, -1);
      }
      break;
    case 'O':
      nfsportout = atoi(optarg);
      if (nfsportout <= 0) {
	fprintf(stderr,
		"mountd: bad port number: %s\n",
		optarg);
	usage(stderr, -1);
      }
      break;
    case 'V':
      mountvers = atoi(optarg);
      if(mountvers<1 || mountvers>3) {
	fprintf(stderr, "Mount version %i not supported\n",mountvers);
	exit(-1);
      }
      break;
    case 'S':
      rootsock = atoi(optarg);
      if (rootsock <= 0) {
	fprintf(stderr,
		"mountd: bad socket number: %s\n",
		optarg);
	usage(stderr, -1);
      }
      break;
    case 't':
      if(strcmp(optarg,"tcp")==0) {
	transport = NFS_OVER_TCP;
      } else {
	if(strcmp(optarg,"udp")==0) {
	  transport = NFS_OVER_UDP;
	} else {
	  fprintf(stderr, "mountd: bad transport: %s\n",
		  optarg);
	  usage(stderr, -1);
	}
      }
      break;
    case 's':
      if((send_sk_path = optarg) == NULL) {
        fprintf(stderr,"Must specify path to get session key\n");
        usage(stderr, -1);;
      }
      send_sk_enabled = 1;
      break;
    case 'm':
      if((match_sk_path = optarg) == NULL) {
        fprintf(stderr,"Must specify path to match session key\n");
        usage(stderr, -1);;
      }
      match_sk_enabled = 1;
      break;
    case 'p':
      promiscuous = 1;
      break;
    case 'v':
      printf("%s\n", version);
      exit(0);
    case 0:
      break;
    case '?':
    default:
      usage(stderr, -1);
    }

  /* No more arguments allowed. */
  if (optind != argc)
    usage(stderr, -1);

  if(auth_file == NULL) {
    fprintf(stderr,"Must specify exports file\n");
    exit(-1);
  }

  /* needed for auth_clnt.c */
  authvers = mountvers;

  /* Initialize logging. */
  log_open("mountd", foreground);

  /* Create services and register with portmapper */
  rpc_init("mountd", MOUNTPROG, mountvers, mount_dispatch,
	   nfsportin, NFS_MAXDATA);

  /* Obtain local host name for default forwarding address if
     not given as command-line argument */
  if(localhostname[0] == '\x0') {
    if(gethostname(localhostname,HOSTNAMESIZE)!=0) {
      fprintf(stderr,"gethostname(): couldn't obtain local host name\n");
      exit(-1);
    }
  }

  /* initialize v1 client */
  if(mountvers==1) {
    if((mountdclient1=open_mountdclient(localhostname,nfsportout,1,
					rootsock,transport))==NULL) {
      fprintf(stderr,"Couldn't create v1 client to %s:%i; bailing out\n",
	      localhostname,nfsportout);
      exit(-1);
    }
    /* needed in auth_clnt.c  */
    authclient1 = mountdclient1;
  }
  /* initialize v2 client */
  if(mountvers==2) {
    if((mountdclient2=open_mountdclient(localhostname,nfsportout,2,
					rootsock,transport))==NULL) {
      fprintf(stderr,"Couldn't create v2 client to %s:%i; bailing out\n",
	      localhostname,nfsportout);
      exit(-1);
    }
    /* needed in auth_clnt.c  */
    authclient2 = mountdclient2;
  }
  /* initialize v3 client */
  if(mountvers==3) {
    if((mountdclient3=open_mountdclient(localhostname,nfsportout,3,
					rootsock,transport))==NULL) {
      fprintf(stderr,"Couldn't create v3 client to %s:%i; bailing out\n",
	      localhostname,nfsportout);
      exit(-1);
    }
    /* needed in auth_clnt.c  */
    authclient3 = mountdclient3;
  }

  if (!foreground && !_rpcpmstart) {
#ifndef RPC_SVC_FG
    /* We first fork off a child. */
    if ((c = fork()) > 0)
      exit(0);
    if (c < 0) {
      Dprintf(L_FATAL, "mountd: cannot fork: %s\n",
	      strerror(errno));
    }
    /* No more logging to stderr */
    background_logging();

    /* Now we remove ourselves from the foreground. */
    (void) close(0);
    (void) close(1);
    (void) close(2);
#ifdef TIOCNOTTY
    if ((c = open("/dev/tty", O_RDWR)) >= 0) {
      (void) ioctl(c, TIOCNOTTY, (char *) NULL);
      (void) close(c);
    }
#else
    setsid();
#endif
#endif /* not RPC_SVC_FG */
	}

  /* Initialize the AUTH module. */
  auth_init(auth_file);

  if(send_sk_enabled || match_sk_enabled)
    if(sk_init(send_sk_path, match_sk_path) != 0) {
      Dprintf(L_ERROR, "Session key initialization failed\n");
      exit(-1);
  }

  /* Enable the LOG toggle with a signal. */
  install_signal_handler(SIGUSR1, toggle_logging);

  /* Enable rereading of exports file */
  install_signal_handler(SIGHUP, reinitialize);

  /* Graceful shutdown */
  install_signal_handler(SIGTERM, sigterm);

  atexit(terminate);
  fprintf(stdout,"Proxy started successfully.\n");
  fflush(stdout);
  svc_run ();

  Dprintf (L_ERROR, "Ack! Gack! svc_run returned!\n");
  exit (-1);
}

static void
usage(FILE *fp, int n)
{
  fprintf(fp,
	  "\nUsage: %s -f exports-file\n"
	  "       [-d kind] [-H servername] [-S socketd] [-t tcp|udp]\n"
          "       [-V mountvers] [-C cookiefile] [-P inputport]\n"
          "       [-O outputport] [-F] [-h] [-n] [-p] [-v]\n"
          "       [-s sendskpath] [-m matchskpath]\n"
	  , program_name);
  fprintf(fp,
          "\nOptions: \n"
	  "       -f: specifies path of exports file (required)\n"
	  "       -d: enables debugging mode(auth,all,call,devtab,acache,dcache,general,rmtab,stale,ugid)\n"
	  "       -H: specifies server name (local hostname is used if not specified)\n"
	  "       -S: specifies socket\n"
	  "       -t: specifies transport protocol (tcp or udp <default>)\n"
	  "       -V: specifies mount version (1, 2 <default>, or 3)\n"
	  "       -C: specifies cookie file\n"
	  "       -P: specifies input port number (a random port is chosen and printed out if not specified)\n"
	  "       -O: specifies output port number (a random port is chosen and printed out if not specified)\n"
	  "       -F: puts process to foreground\n"
	  "       -h: prints help\n"
	  "       -n: allows non root\n"
	  "       -p: enables promiscuous mode\n"
	  "       -v: prints version number\n"
	  "       -s: specifies path of the session key to be sent\n"
	  "       -m: specifies path of the session key to match with\n");
  exit(n);
}

static RETSIGTYPE
sigterm(int sig)
{
  /* marked as atexit--don't need to call:
     terminate();
  */
  exit(-1);
}

static void
terminate(void)
{
  rpc_exit(MOUNTPROG, mountvers);

  close_mountdclient(mountdclient1);
  close_mountdclient(mountdclient2);
  close_mountdclient(mountdclient3);
}

RETSIGTYPE
reinitialize(int sig)
{
  static volatile int	inprogress = 0;

  if (_rpcsvcdirty) {
    need_reinit = 1;
    return;
  }
  if (inprogress++)
    return;
  /* We're not handling file-handles in gvfs.mountd */
  /* fh_flush(1); */
  auth_init(NULL);
  inprogress = 0;
  need_reinit = 0;

}

/*
 * Don't look. This is an awful hack to overcome a link problem with
 * auth_clnt temporarily.
 */
uid_t
luid(uid_t uid, nfs_mount *mp, struct svc_req *rqstp)
{
  return -2;
}

gid_t
lgid(gid_t gid, nfs_mount *mp, struct svc_req *rqstp)
{
  return -2;
}

void
ugid_free_map(struct ugid_map *map)
{
  /* NOP */
}

void
ugid_map_uid(nfs_mount *mp, uid_t from, uid_t to)
{
  /* NOP */
}

void
ugid_map_gid(nfs_mount *mp, gid_t from, gid_t to)
{
  /* NOP */
}

void
ugid_squash_uids(nfs_mount *mp, uid_t from, uid_t to)
{
  /* NOP */
}

void
ugid_squash_gids(nfs_mount *mp, gid_t from, gid_t to)
{
  /* NOP */
}
