/***************************************************************************
 * nfsd
 *		This is the main program for the grid virtual file
 *		system proxy gvfs.nfsd. This proxy receives, modifies
 *		and forwards NFS RPC calls.
 *
 * Author:	Renato J. Figueiredo, Ming Zhao
 *
 *
 * Updated:	Jan., 2004
 *		Implemented write-back with write-allocate for data cache;
 *		Implemented write-through with write-allocate for attributes
 *		cache;
 *		Added a command line option -W for enabling write-back;
 *		Added cache support for CREATE, REMOVE, RENAME, and
 *		LINK procedures;
 *		Added cache support situations that requested data in READ
 *		or WRITE cross block boudaries;
 *		Corrected the invalidation of data cache when requested data
 *		crosses block boundaries;
 *		Corrected the problem that seekset is not evaluated properly
 *		in dfind_replace_frame function.
 *		Added support for session key authentication.
 *		Added two command line options -s and -m for enabling sending
 *		session key within RPC message, and matching session key from
 *		RPC message respectively.
 * Updated:	Aug., 2004
 *		Fixed a bug that attributes fetched by LOOKUP are not
 *		correctly updated in LOOKUP and GETATTR cache;
 *		Fixed a bug that cache hit is not correctly logged in WRITE;
 * Updated:	Feb., 2005
 *		Corrected the problem that id mapping is not done within the
 *		arguments of setattr call.
 *
 * Updated:	June., 2006
 *		Added caching support for NFS V3
 *		Added invalidation-polling based cache consistency
 *
 * Based on User-level NFSD code by:
 *
 * Authors:	Mark A. Shand, May 1988
 *		Donald J. Becker, <becker@super.org>
 *		Rick Sladkey, <jrs@world.std.com>
 *		Fred N. van Kempen, <waltje@uWalt.NL.Mugnet.ORG>
 *		Eric Kasten, <tigger@tigger.cl.msu.edu>
 *		Olaf Kirch, <okir@monad.swb.de>
 *
 *		Copyright 1988 Mark A. Shand
 *		This software maybe be used for any purpose provided
 *		the above copyright notice is retained.  It is supplied
 *		as is, with no warranty expressed or implied.
 ***************************************************************************/

#include "nfsd.h"
#include <rpc/pmap_clnt.h>
#include "getopt.h"
#include "rpcmisc.h"
#include "signals.h"
#include "cache_attrib.h"
#include "cache_data.h"
#include "sk.h"
#include "invalidation.h"
#ifdef __linux__ /* XXX - MvS: for UNIX sockets. */
#  include <sys/un.h>
#endif
#ifdef HAVE_SYSLOG_H
# include <syslog.h>
#endif

#ifdef SOLARIS
#define PORTMAP
#include <rpc/rpc.h>
#include <rpc/svc_soc.h>
#endif /* SOLARIS */

#define MULTIPLE_SERVERS	/* load balancing; not used in gvfs yet */

/* Flags for auth_fh */
#define CHK_READ		0
#define CHK_WRITE		1
#define CHK_NOACCESS	2

extern char version[];
static char *program_name;

/*
 * Option table for command-line arguments
 */
static struct option longopts[] = {
			{ "debug",				required_argument,	0,	'd' },
			{ "foreground",			0,					0,	'F' },
			{ "exports-file",		required_argument,	0,	'f' },
			{ "help",				0,					0,	'h' },
			{ "host-server",		required_argument,	0,	'H' },
			{ "portin",				required_argument,	0,	'P' },
			{ "portout",			required_argument,	0,	'O' },
			{ "sock",				required_argument,	0,	'S' },
			{ "transport",			required_argument,	0,	't' },
			{ "nfsversion",			required_argument,	0,	'V' },
			{ "cookiefile",			required_argument,	0,	'C' },
			{ "cachesessionid",		required_argument,	0,	'i' },
			{ "enableattrcache",	0,					0,	'A' },
			{ "enabledatacache",	0,					0,	'D' },
			{ "enablewriteback",	0,					0,	'W' },
			{ "cachebasepath",		required_argument,	0,	'c' },
			{ "attrcachesize",		required_argument,	0,	'z' },
			{ "attrcacheassoc",		required_argument,	0,	'a' },
			{ "attrcachebanks",		required_argument,	0,	'b' },
			{ "datacachesize",		required_argument,	0,	'e' },
			{ "datacacheassoc",		required_argument,	0,	'o' },
			{ "datacachebanks",		required_argument,	0,	'k' },
			{ "syncinterval",		0,					0,	'T' },
			{ "mininvalintervl",	0,					0,	'x' },
			{ "maxinvalinterval",	0,					0,	'y' },
			{ "enableinvalidation",	0,					0,	'I' },
			{ "sessionkeysend",		0,					0,	's' },
			{ "sessionkeymatch",	0,					0,	'm' },
			{ "promiscuous",		0,					0,	'p' },
			{ "version",			0,					0,	'v' },
			{ NULL,					0,					0,	0	}
};
static const char * shortopts = "d:Ff:hH:P:O:S:t:V:C:i:ADWc:z:a:b:e:o:k:T:x:y:I:s:m:gpv::";

/* Global variable used for SIGHUP handling; SIGHUPs force the proxy
   to re-read its configuration files; this variable implements a
   simple form of mutual exclusion and makes sure this
   does not happen while an RPC request is being serviced */
int need_reinit = 0;
/* Global variable used for file cache flushes; this variable
   implements a simple form of mutual exclusion and makes sure this
   does not happen while an RPC request is being serviced */
int need_flush = 0;
/* Global variable used for setting alarm for client polling invalidation;
   this variable implements a simple form of mutual exclusion and makes sure
   this does not happen while an RPC request is being serviced */
int need_alarm;
/* Global variable that determines if attribute caching is enabled */
int attr_cache_enabled = 0;
/* Global variable that determines if data caching is enabled */
int data_cache_enabled = 0;
/* Global variable that determines if write back caching is enabled */
int wb_cache_enabled = 0;
/* Global variable that determines if client polling invalidation caching is enabled */
int inval_enabled = 0;

/* Global variables that determines if auth_clnt.c must refresh client
   UNIX AUTH authenticator */
int refresh_authentication_handle_uid = -1;
int refresh_authentication_handle_gid = -1;
/* Global variable: current access hit the data or attr cache */
int data_attr_cache_hit = 0;
/* Buffer for cached data replies */
char data_cache_buffer[MAX_NFS_BLOCKSIZE];
char data_cache_buffer3[MAX_NFS3_BLOCKSIZE];
/* NFS version that this proxy will forward */
int nfsvers = 2;
int authvers;

/* Call profiling defines for functions and data structures; see
   Makefile.in for a description of the profiling #define's */

#ifdef CALL_PROFILING
extern void dump_stats(int sig);
extern struct timeval inittime;
#endif

/* NFSD client for forwarding V2 NFS protocol calls */
CLIENT *nfsdclient2;
/* NFSD client for forwarding V3 NFS protocol calls */
CLIENT *nfsdclient3;
CLIENT *nfsdclient;

int nfsportin = 0;
int nfsportout = 0;
int rootsock = 0;
int transport = NFS_OVER_TCP;
char localhostname[HOSTNAMESIZE];
int clientport = 0;

/* Function definitions */
nfs_mount *auth_fh(struct svc_req *rqstp, nfsstat *statp, CLIENT *authclient);
static void usage(FILE *fp, int n);
static void terminate(void);
static RETSIGTYPE sigterm(int sig);
static RETSIGTYPE sigusr2(int sig);
RETSIGTYPE flush_caches(int sig);
RETSIGTYPE reinitialize(int sig);
RETSIGTYPE setalarm(int sig);

/* External function definitions */
extern CLIENT *open_nfsdclient(char *nfshost, int nfsport,
						 int vers, int rootsock, int flags);
extern void close_nfsdclient(CLIENT *retclnt);
extern void log_cache_hits(int proc_index);

/* macros for handling NFS requests */
#define AUTHENTICATE_CLIENT \
				if((nfsmount=auth_fh(rqstp, &astat, nfsdclient))==NULL) { \
					result.nfsstat = astat; \
					return(result);}
#define FORWARD_REQUEST(CLIENT_FUNCTION,MESSAGE) \
				if ((CLIENT_FUNCTION(argp, &res, nfsdclient)) != RPC_SUCCESS) { \
					clnt_perror(nfsdclient, MESSAGE); \
					result.nfsstat = NFSERR_ACCES; \
					return result;}
#define READONLY \
				if(nfsmount->o.read_only) { \
					result.nfsstat = NFSERR_ROFS; \
					return result;}
#define NOTOK  \
				if (res.status!=NFS_OK) { \
					result.nfsstat = res.status; \
					return result;}
#define FORWARD_REQUEST3(CLIENT_FUNCTION,MESSAGE) \
				if ((CLIENT_FUNCTION(argp, &res, nfsdclient)) != RPC_SUCCESS) { \
					 clnt_perror(nfsdclient, MESSAGE); \
		 			result.nfsstat = NFS3ERR_ACCES; \
		 			return result; }


/****************************************************************************
 * Map user and group IDs of attributes returned by the NFS server
 * back to the user/group IDs of the client.
 * The original IDs (coming from a client request) are stored in
 * *attr; the mapping functions are defined in ugid_map.c; the
 * mapped IDs (to be forwarded by the proxy) overwrite the
 * client IDs
 ****************************************************************************/
void map_return_attributes(fattr *attr, nfs_mount *nfsmount,
				 struct svc_req *rqstp)
{
	int serveruid, servergid, clientuid, clientgid;

	serveruid = attr->uid;
	servergid = attr->gid;
	clientuid = ruid(serveruid, nfsmount, rqstp);
	clientgid = rgid(servergid, nfsmount, rqstp);
	attr->uid = clientuid;
	attr->gid = clientgid;
	Dprintf(D_UGID, "map_return_attributes: (%i,%i) -> (%i,%i)\n",
		serveruid,servergid,clientuid,clientgid);
}

void map_return_attributes3(fattr3 *attr, nfs_mount *nfsmount,
					 struct svc_req *rqstp)
{
	uint32 serveruid, servergid, clientuid, clientgid;

	serveruid = attr->uid;
	servergid = attr->gid;
	clientuid = (uint32) ruid(serveruid, nfsmount, rqstp);
	clientgid = (uint32) rgid(servergid, nfsmount, rqstp);
	attr->uid = clientuid;
	attr->gid = clientgid;
	Dprintf(D_UGID, "map_return_attributes3: (%i,%i) -> (%i,%i)\n",
		serveruid,servergid,attr->uid,attr->gid);
}

/****************************************************************************
 * Return NFS error string
 ****************************************************************************/
char *
nfs_error(enum nfsstat stat)
{
	switch (stat) {
	case NFS_OK:
		return "No error";
	case NFSERR_PERM:
		return "Not owner";
	case NFSERR_NOENT:
		return "No such file or directory";
	case NFSERR_IO:
		return "I/O error";
	case NFSERR_NXIO:
		return "No such device or address";
	case NFSERR_ACCES:
		return "Permission denied";
	case NFSERR_EXIST:
		return "File exists";
	case NFSERR_NODEV:
		return "No such device";
	case NFSERR_NOTDIR:
		return "Not a directory";
	case NFSERR_ISDIR:
		return "Is a directory";
	case NFSERR_FBIG:
		return "File too large";
	case NFSERR_NOSPC:
		return "No space left on device";
	case NFSERR_ROFS:
		return "Read-only file system";
	case NFSERR_NAMETOOLONG:
		return "File name too long";
	case NFSERR_NOTEMPTY:
		return "Directory not empty";
	case NFSERR_DQUOT:
		return "Disc quota exceeded";
	case NFSERR_STALE:
		return "Stale NFS file handle";
	case NFSERR_WFLUSH:
		return "Write cache flushed";
	default:
		return "UKNOWN NFS ERROR";
	}
}

/****************************************************************************
 * auth_fh
 *
 * This function authenticates the file handle provided by the client.
 * If the file handle is authenticated by this function, it returns
 * a '1", otherwise a "0". Requests that are authenticated are
 * forwarded by the proxy; requests not authenticated are not
 * forwarded.
 ****************************************************************************/
nfs_mount *
auth_fh(struct svc_req *rqstp, nfsstat *statp, CLIENT *authclient)
{
	nfs_mount *scan = NULL;
	nfs_client *nfsclient = NULL;

	/* First, authenticate the client. If not successfull, flag NFS
	   access error and return zero */
	if ((nfsclient = auth_clnt(rqstp)) == NULL) {
		*statp = NFSERR_ACCES;
		return 0;
	}

	/* Authenticate by the session key if it is enabled. Successful
	   authentication returns client id larger or equal than 1 */
	if(match_sk_enabled)
		if((client_id = auth_sk(rqstp)) < 1) {
			*statp = NFSERR_ACCES;
			return 0;
	}

	/* Make sure mount point is non-null */
	if (nfsclient->m == NULL) {
		Dprintf(L_ERROR, "NULL mountpoint\n");
		*statp = NFSERR_ACCES;
		return 0;
	}

	/* Authenticate user (uid/gid); return NFS access error if user
	   is not authorized */
	*statp = NFSERR_ACCES;
	for(scan = nfsclient->m; scan != NULL; scan = scan->next) {
		if(auth_user(scan, rqstp, authclient)) {
			*statp = NFS_OK;
			break;
		}
	}

	if(scan == NULL)
		return scan;

	/* Check if request comes from insecure port and not explicitly
		 allowed for this mount point (i.e. option "insecure" not set
		 in exports file) */
	if (!allow_non_root && scan->o.secure_port
			&& !SECURE_PORT(svc_getcaller(rqstp->rq_xprt)->sin_port)) {
		Dprintf(L_ERROR,
			"NFS request from %s originated on insecure port, %s\n",
			nfsclient->clnt_name,
			"psychoanalysis suggested");
		*statp = NFSERR_ACCES;
		return 0;
	}

	/* Depending on value of *statp, auth has been successful or not */
	/* return mount pointer (NULL if not successfull...)*/
	return(scan);
}


/****************************************************************************
 * The functions below are called when the server-side of the proxy
 * receives an NFS request. There are as many functions as defined
 * in the NFS protocol (for NFS-V2, and NFS-V3 if enabled). All functions
 * have a similar structure, as follows:
 *   - 1. authenticate request with auth_fh; return immediately
 *     if not authenticated.
 *   - 2. forward the request (via a client call from the proxy).
 *     Note that requests are forwarded with the uid/gid of the
 *     user that is executing the proxy.
 *   - 3. check for errors in the processing of the forwarded request;
 *     map the server return attributes, and return the mapped NFS
 *     reply to the client
 * The "wrappers" of the following functions came from `rpcgen -l nfs_prot.x`.
 * This normally generates the client routines, but it provides nice
 * prototypes for the server routines also.
 * Refer to the NFS protocol for descriptions of the procedures and
 * data structures used in these functions
 ****************************************************************************/

union result_types
nfsd_nfsproc_null_2(void *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	attrstat res;
	union result_types result;
	nfs_mount *nfsmount = NULL;

	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	/* Forward request to server */
	FORWARD_REQUEST(nfsproc_null_2, "nfsproc_null_2");
	NOTOK;

	result.nfsstat = NFS_OK;
	/* No need to map attributes in NULL call */
	return (result);
}

union result_types
nfsd_nfsproc_getattr_2(nfs_fh *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	attrstat res;
	int lkupres=0, insres;
	union result_types result;
	nfs_mount *nfsmount = NULL;

	Dprintf(D_CACHE, "*** GETATTR:\n");
	hexdump(argp->data, NFS_FHSIZE, D_CACHE);

	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	if(attr_cache_enabled) { /* look up cache; retrieve results if hit */
		lkupres = cache_lookup(rqstp->rq_vers, rqstp->rq_proc,
				 (void *)argp, (void *)&result.attrstat);
		if(lkupres==1) {
#ifdef CALL_PROFILING /* log cache hit */
			log_cache_hits(rqstp->rq_proc);
#endif
		}
	}
	if(lkupres!=1) { /* cache disabled or cache miss; forward request to server */
		/* Forward request to server */
		FORWARD_REQUEST(nfsproc_getattr_2, "nfsproc_getattr_2");
		NOTOK;
		if(attr_cache_enabled) { /* insert RPC response in cache */
			insres = cache_insert(rqstp->rq_vers, rqstp->rq_proc, NULL, (void *)&res);
		}
		result.attrstat = res;
	}

	Dprintf(D_CACHE, "current file size %d, uid %d, gid %d, mtime %d\n",
		result.attrstat.attrstat_u.attributes.size,
		result.attrstat.attrstat_u.attributes.uid,
		result.attrstat.attrstat_u.attributes.gid,
		result.attrstat.attrstat_u.attributes.mtime);

	/* Map attributes and return to client */
	map_return_attributes(&result.attrstat.attrstat_u.attributes, nfsmount, rqstp);

	return (result);
}

union result_types
nfsd_nfsproc_setattr_2(sattrargs *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	attrstat res;
	int lkupres=0, insres;
	fh_node *node;
	uid_t cuid;
	gid_t cgid;
	union result_types result;
	nfs_mount *nfsmount = NULL;

	Dprintf(D_CACHE, "*** SETATTR:\n");
	hexdump(argp->file.data, NFS_FHSIZE, D_CACHE);

	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	READONLY;

	if(wb_cache_enabled) { /* write-through with write-allocate */
		if(argp->attributes.size == 0) {
			Dprintf(D_CACHE, "nfsd_nfsproc_setattr_2: truncate file\n");
			node = (fh_node *)malloc(sizeof(fh_node));
			node->file.handle = argp->file;
			node->status = FH_TRUNC;
			fh_hash_insert(stale_fh_hash, node);
		} else /* use locally modified file size */
			lkupres = cache_lookup(rqstp->rq_vers, NFSPROC_GETATTR,
		                           (void *)argp->file.data, (void *)&result.attrstat);
	}

	/* Forward request to server */
	if(argp->attributes.uid != -1) {
		cuid = luid(argp->attributes.uid, nfsmount, rqstp);
		Dprintf(D_CACHE, "map uid in args from %d to %d\n", argp->attributes.uid, cuid);
		argp->attributes.uid = cuid;
	}
	if(argp->attributes.gid != -1) {
		cgid = lgid(argp->attributes.gid, nfsmount, rqstp);
		Dprintf(D_CACHE, "map gid in args from %d to %d\n", argp->attributes.gid, cgid);
		argp->attributes.gid = cgid;
	}
	FORWARD_REQUEST(nfsproc_setattr_2, "nfsproc_setattr_2");
	NOTOK;

	if(attr_cache_enabled) {
		if(wb_cache_enabled && lkupres)
			res.attrstat_u.attributes.size = result.attrstat.attrstat_u.attributes.size;
		lkupres = cache_update(rqstp->rq_vers, NFSPROC_GETATTR,
				(void *)&argp->file, (void *)&res);
		if(!lkupres)
			insres = cache_insert(rqstp->rq_vers, NFSPROC_GETATTR, NULL, (void *)&res);
	}

	result.attrstat = res;
	map_return_attributes(&result.attrstat.attrstat_u.attributes,
			nfsmount, rqstp);
	return (result);
}

union result_types
nfsd_nfsproc_root_2(void *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	attrstat res;
	union result_types result;
	nfs_mount *nfsmount = NULL;

	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	/* Forward request to server */
	FORWARD_REQUEST(nfsproc_root_2, "nfsproc_root_2");
	NOTOK;

	/* Map attributes and return to client */
	result.attrstat = res;

	map_return_attributes(&result.attrstat.attrstat_u.attributes,
			nfsmount, rqstp);
	return (result);
}

union result_types
nfsd_nfsproc_lookup_2(diropargs *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	diropres res;
	int lkupres=0, lkupres2=0, insres;
	attrstat res_attr;
	union result_types result;
	nfs_mount *nfsmount = NULL;

	Dprintf(D_CACHE, "*** LOOKUP: %s\n", argp->name);
	hexdump(argp->dir.data, NFS_FHSIZE, D_CACHE);

	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	if(attr_cache_enabled) { /* look up cache; retrieve results if hit */
		lkupres = cache_lookup(rqstp->rq_vers, rqstp->rq_proc,
			                   (void *)argp, (void *)&result.diropres);
		if(lkupres) {
			if(result.diropres.status != NFS_OK) { /* cached NOENT */
				Dprintf(D_CACHE, "file does not exist (status: %d)\n",
					    result.diropres.status);
#ifdef CALL_PROFILING
				log_cache_hits(rqstp->rq_proc);
#endif
				result.nfsstat = result.diropres.status;
				return result;
			}

			/* attributes stored in LOOKUP cache may be outdated;
			   use the attributes in GETATTR cache instead */
			lkupres2 = cache_lookup(rqstp->rq_vers, NFSPROC_GETATTR,
			                       (void *)&result.diropres.diropres_u.diropres.file,
			                       (void *)&res_attr);
			if(lkupres2) { /* keep consistent with GETATTR cache */
#ifdef CALL_PROFILING
				log_cache_hits(rqstp->rq_proc);
#endif
				result.diropres.diropres_u.diropres.attributes = res_attr.attrstat_u.attributes;
				goto out;
			}
		}
	}

	/* cache not enabled or cache miss; forward request to server */
	FORWARD_REQUEST(nfsproc_lookup_2, "nfsproc_lookup_2");
	if(res.status != NFS_OK) { /* cache NOENT directly */
		if(attr_cache_enabled && res.status == NFSERR_NOENT) {
			Dprintf(D_CACHE, "file does not exist\n");
			result.diropres = res;
			if (lkupres)
				lkupres = cache_update(rqstp->rq_vers, rqstp->rq_proc, (void *)argp,
						(void *)&result.diropres);
			else
				insres = cache_insert(rqstp->rq_vers, rqstp->rq_proc, NULL,
						(void *)&result.diropres);
		}
		result.nfsstat = res.status;
		return result;
	}

	result.diropres = res;

	if(attr_cache_enabled) {
		/* stored results in LOOKUP cache */
		if(lkupres)
			cache_update(rqstp->rq_vers, rqstp->rq_proc, (void *)argp,
				         (void *)&result.diropres);
		else
			cache_insert(rqstp->rq_vers, rqstp->rq_proc, (void *)argp,
				         (void *)&result.diropres);
		/* store returned attributes in GETATTR cache */
		if(!lkupres) /* GETATTR cache is not checked yet */
			lkupres2 = cache_lookup(rqstp->rq_vers, NFSPROC_GETATTR,
			                       (void *)&result.diropres.diropres_u.diropres.file,
			                       (void *)&res_attr);
		if(!lkupres2) {
			res_attr.status = NFS_OK;
			res_attr.attrstat_u.attributes = result.diropres.diropres_u.diropres.attributes;
			insres = cache_insert(rqstp->rq_vers, NFSPROC_GETATTR,
					(void *)&result.diropres.diropres_u.diropres.file, (void *)&res_attr);
		} else {
			if(wb_cache_enabled) /* return locally modified attributes */
				result.diropres.diropres_u.diropres.attributes = res_attr.attrstat_u.attributes;
			else {
				res_attr.attrstat_u.attributes = result.diropres.diropres_u.diropres.attributes;
				cache_update(rqstp->rq_vers, NFSPROC_GETATTR,
				             (void *)&result.diropres.diropres_u.diropres.file, (void *)&res_attr);
			}
		}
	}

out:
	hexdump(result.diropres.diropres_u.diropres.file.data, NFS_FHSIZE, D_CACHE);
	Dprintf(D_CACHE, "current file size %d, mtime %d\n",
			result.diropres.diropres_u.diropres.attributes.size,
			result.diropres.diropres_u.diropres.attributes.mtime.seconds);
	/* Map attributes and return to client */
	map_return_attributes(&result.diropres.diropres_u.diropres.attributes,
			nfsmount, rqstp);
	return (result);
}

union result_types
nfsd_nfsproc_readlink_2(nfs_fh *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	readlinkres res;
	union result_types result;
	nfs_mount *nfsmount = NULL;
	res.readlinkres_u.data = malloc(MAX_NFS_BLOCKSIZE);

	Dprintf(D_CACHE, "*** READLINK:\n");
	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	/* Forward request to server */
	FORWARD_REQUEST(nfsproc_readlink_2, "nfsproc_readlink_2");
	NOTOK;

	result.readlinkres = res;

	/* No need to map attributes */
	return (result);
}

union result_types
nfsd_nfsproc_read_2(readargs *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	readres *res, read_res;
	int lkupres = 0,lkupres2 = 0,insres, crossed_boundary = 0, forward = 0;
	int requestoffset;
	readargs args;
	attrstat res_attr;
	union result_types result;
	nfs_mount *nfsmount = NULL;
	char *cache_buffer = malloc(MAX_NFS_BLOCKSIZE);

	Dprintf(D_CACHE, "*** READ: offset %d count %d \n", argp->offset, argp->count);
	hexdump(argp->file.data, NFS_FHSIZE, D_CACHE);
	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	result.readres.readres_u.reply.data.data_val = cache_buffer;
	res = &result.readres;

	if(!data_cache_enabled) goto no_caching;
	else { /* look up cache; retrieve results if hit */
		/* get attributes for validation of cached data */
		lkupres2 = cache_lookup(rqstp->rq_vers, NFSPROC_GETATTR,
			                    (void *)&argp->file, (void *)&res_attr);

		args = *argp;

		if(lkupres2 && args.offset >= res_attr.attrstat_u.attributes.size) {
			Dprintf(D_CACHE, "offset is greater than file size\n");
			result.readres.status = NFS_OK;
			result.readres.readres_u.reply.data.data_len = 0;
			result.readres.readres_u.reply.attributes = res_attr.attrstat_u.attributes;
			goto out;
		}

		if(((args.offset % MAX_NFS_BLOCKSIZE) + args.count) >
			 MAX_NFS_BLOCKSIZE) {
			/* read crossed boundary: lookup the first block */
			Dprintf(D_CACHE, "read crossed boundary\n");
			crossed_boundary = 1;
			args.count = MAX_NFS_BLOCKSIZE - (args.offset % MAX_NFS_BLOCKSIZE);
			args.totalcount = args.count;
		}

		/* try to fetch locally */
		lkupres = dcache_lookup(rqstp->rq_vers, rqstp->rq_proc,
				(void *)&args, (void *)&result.readres);

		if(lkupres) {
			if(result.readres.status == 100) { /* block is locally modified; no need to check server */
				Dprintf(D_CACHE, "block is locally modified");
				result.readres.status = NFS_OK;
			} else { /* Must check modification time from server */
				if(lkupres2 != 1) {
					if(nfsproc_getattr_2(&argp->file, &res_attr, nfsdclient) != RPC_SUCCESS) {
						Dprintf(D_CACHE, "getattr from server error %d\n");
						result.nfsstat = NFSERR_ACCES;
						return result;
					}
					if(res_attr.status != NFS_OK) {
						result.nfsstat = res_attr.status;
						return result;
					}
					lkupres2 = 1;
				}
				if(result.readres.readres_u.reply.attributes.mtime.seconds <
					res_attr.attrstat_u.attributes.mtime.seconds) {
					Dprintf(D_CACHE, "cached data may be stale(%d:%d); flush\n",
							result.readres.readres_u.reply.attributes.mtime.seconds,
							res_attr.attrstat_u.attributes.mtime.seconds);
					dcache_inval(rqstp->rq_vers, rqstp->rq_proc, (void *)&args);
					lkupres = 0;
				}
			}
		}

		if(lkupres == -1) goto no_caching; /* error */
		else if(lkupres == 0) { /* not cached */
			/* fetch the entire first block */
			args.offset = (argp->offset) - (argp->offset%MAX_NFS_BLOCKSIZE);
			args.count = MAX_NFS_BLOCKSIZE;
			args.totalcount = MAX_NFS_BLOCKSIZE;

			forward = 1;
			if((nfsproc_read_2(&args, res, nfsdclient)) != RPC_SUCCESS) {
				clnt_perror(nfsdclient,"nfsproc_read_2");
				goto no_caching;
			}
			if(res->status!=NFS_OK) {
				Dprintf(L_ERROR, "read from server error %d\n", res->status);
				goto no_caching;
			}
			Dprintf(D_CACHE, "forward: offset %d, return: data len %d\n",
					args.offset, res->readres_u.reply.data.data_len);

			/* insert the first block into cache */
			if(res->readres_u.reply.data.data_len > 0) {
				args.offset = argp->offset;
				if(crossed_boundary)
					args.count = MAX_NFS_BLOCKSIZE - (args.offset % MAX_NFS_BLOCKSIZE);
				else args.count = argp->count;
				args.totalcount = args.count;
				insres = dcache_insert((void *)&args,rqstp->rq_vers,
						rqstp->rq_proc,(void *)res);
				if(insres != 0) {
					goto no_caching;
				}
			}
			result.readres = *res;
		}

		if(crossed_boundary) { /* fetch second block */
			read_res = result.readres;
			if((read_res.readres_u.reply.data.data_val = malloc(argp->count))
					== NULL) {
				Dprintf(D_CACHE,  "error malloc memory\n");
				goto no_caching;
			}
			if(read_res.readres_u.reply.data.data_len > 0)
				if(memcpy(read_res.readres_u.reply.data.data_val,
						result.readres.readres_u.reply.data.data_val,
						read_res.readres_u.reply.data.data_len) == NULL) {
					Dprintf(D_CACHE, "error memory copy\n");
					free(read_res.readres_u.reply.data.data_val);
					goto no_caching;
			}
			Dprintf(D_CACHE, "got first part: data_len %d\n",
					read_res.readres_u.reply.data.data_len);

			requestoffset = argp->offset + argp->count;
			args.offset = requestoffset - (requestoffset % MAX_NFS_BLOCKSIZE);
			args.count = requestoffset % MAX_NFS_BLOCKSIZE;
			args.totalcount = args.count;

			/* try to fetch locally */
			lkupres = dcache_lookup(rqstp->rq_vers, rqstp->rq_proc,
					(void *)&args, (void *)&result.readres);

			if(lkupres) {
				if(result.readres.status == 100)
					result.readres.status = NFS_OK;
				else { /* Must check modification time from server */
					if(lkupres2 != 1) {
						if(nfsproc_getattr_2(&argp->file, &res_attr, nfsdclient) != RPC_SUCCESS) {
							Dprintf(D_CACHE, "getattr from server error %d\n");
							result.nfsstat = NFSERR_ACCES;
							return result;
						}
						if(res_attr.status != NFS_OK) {
							result.nfsstat = res_attr.status;
							return result;
						}
						lkupres2 = 1;
					}
					if(result.readres.readres_u.reply.attributes.mtime.seconds <
							res_attr.attrstat_u.attributes.mtime.seconds) {
						Dprintf(D_CACHE, "the file has been modified on server, \
								and cached data may be stale\n");
								dcache_inval(rqstp->rq_vers, rqstp->rq_proc, (void *)&args);
						lkupres = 0;
					}
				}
			}
			if(lkupres == -1) {  /* error */
				free(read_res.readres_u.reply.data.data_val);
				goto no_caching;
			}
			else if(lkupres == 0) { /* not cached */
				/* fetch the entire second block */
				args.count = MAX_NFS_BLOCKSIZE;
				args.totalcount = MAX_NFS_BLOCKSIZE;

				forward = 1;
				if((nfsproc_read_2(&args, res, nfsdclient)) != RPC_SUCCESS) {
					clnt_perror(nfsdclient,"nfsproc_read_2");
					free(read_res.readres_u.reply.data.data_val);
					goto no_caching;
				}
				if (res->status!=NFS_OK) {
					Dprintf(L_ERROR, "read from server error %d\n", res->status);
					free(read_res.readres_u.reply.data.data_val);
					goto no_caching;
				}
				Dprintf(D_CACHE, "forward: offset %d, return: data len %d\n",
						args.offset, res->readres_u.reply.data.data_len);
				/* insert the second block into cache */
				if(res->readres_u.reply.data.data_len > 0) {
					args.count = requestoffset % MAX_NFS_BLOCKSIZE;
					args.totalcount = args.count;
					insres = dcache_insert((void *)&args,rqstp->rq_vers,
							rqstp->rq_proc,(void *)res);
					if(insres != 0) {
						free(read_res.readres_u.reply.data.data_val);
						goto no_caching;
					}
				}
				result.readres = *res;
			}

			if(result.readres.readres_u.reply.data.data_len > 0)
				if(memcpy(&read_res.readres_u.reply.data.data_val[read_res.readres_u.reply.data.data_len],
					      result.readres.readres_u.reply.data.data_val,
			result.readres.readres_u.reply.data.data_len) == NULL) {
				Dprintf(D_CACHE, "error memory copy\n");
				free(read_res.readres_u.reply.data.data_val);
				goto no_caching;
			}
			read_res.readres_u.reply.data.data_len +=
					result.readres.readres_u.reply.data.data_len;
			Dprintf(D_CACHE, "got second part: data_len %d\n",
					read_res.readres_u.reply.data.data_len);

			result.readres = read_res;
			result.readres.readres_u.reply.data.data_val = cache_buffer;
			if(memcpy(result.readres.readres_u.reply.data.data_val,
					read_res.readres_u.reply.data.data_val,
					read_res.readres_u.reply.data.data_len) == NULL) {
				Dprintf(D_CACHE, "error memory copy\n");
				free(read_res.readres_u.reply.data.data_val);
				goto no_caching;
			}
			free(read_res.readres_u.reply.data.data_val);
		} /* endif crossed boundary */

		if(!forward) {
#ifdef CALL_PROFILING
			log_cache_hits(rqstp->rq_proc); /*log cache hit */
#endif
		}

		goto out;
	} /* endif data_cache_enabled */

no_caching: /* dcache disabled or cache error */
	if((nfsproc_read_2(argp, res, nfsdclient)) != RPC_SUCCESS) {
		clnt_perror(nfsdclient,"nfsproc_read_2");
		result.nfsstat = NFSERR_ACCES;
		return result;
	}
	if (res->status != NFS_OK) {
		result.nfsstat = res->status;
		return result;
	}
/*
	if(res->readres_u.reply.data.data_len ==
				result.readres.readres_u.reply.data.data_len) {
		if(memcmp(res->readres_u.reply.data.data_val,
				result.readres.readres_u.reply.data.data_val,
		result.readres.readres_u.reply.data.data_len) != 0) {
			Dprintf(D_CACHE, "data contents do not match\n");
		}
	}
	else Dprintf(D_CACHE, "data lengths do not match\n");
	data_attr_cache_hit = 0; *//* for test only */

	result.readres = *res;
	Dprintf(D_CACHE, "forwarding ... returned data_len %d\n",
			result.readres.readres_u.reply.data.data_len);

out:
	if(attr_cache_enabled) {
		if(!lkupres2) { /* store returned attributes */
			res_attr.status = NFS_OK;
			res_attr.attrstat_u.attributes = result.readres.readres_u.reply.attributes;
			cache_insert(rqstp->rq_vers, NFSPROC_GETATTR, (void *)&argp->file, (void *)&res_attr);
		} else {
			if(wb_cache_enabled) /* return locally modified attributes */
				result.readres.readres_u.reply.attributes = res_attr.attrstat_u.attributes;
			else if(forward) {
				res_attr.attrstat_u.attributes = result.readres.readres_u.reply.attributes;
				cache_update(rqstp->rq_vers, NFSPROC_GETATTR, (void *)&argp->file, (void *)&res_attr);
			}
		}
	}

	/* Map attributes and return to client */
	map_return_attributes(&result.readres.readres_u.reply.attributes,
			nfsmount, rqstp);
	return (result);
}

union result_types
nfsd_nfsproc_writecache_2(void *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	attrstat res;
	union result_types result;
	nfs_mount *nfsmount = NULL;

	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	READONLY;

	/* Forward request to server */
	FORWARD_REQUEST(nfsproc_writecache_2, "nfsproc_writecache_2");
	NOTOK;

	result.nfsstat = NFS_OK;

	/* No need to map attributes */
	return (result);
}

union result_types
nfsd_nfsproc_write_2(writeargs *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	attrstat res, res_attr;
	int lkupres = 0, lkupres2 = 0, insres, new_size, old_size=0, requestoffset, requestcnt = 0;
	readargs rarg;
	readres res_read;
	writeargs warg;
	int forward = 0, crossed_boundary = 0;
	union result_types result;
	nfs_mount *nfsmount = NULL;
	char data_buffer[MAX_NFS_BLOCKSIZE], data_buffer2[MAX_NFS_BLOCKSIZE];

	Dprintf(D_CACHE, "*** WRITE: offset %d totalcount %d \n",
		argp->offset, argp->totalcount);
	hexdump(argp->file.data, NFS_FHSIZE, D_CACHE);
	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	READONLY;

	if(data_cache_enabled) {
		/* Get file attributes from attributes cache or server (if cache miss),
		   which are necessary to determine cached block's data length and file
		   size, and construct the attributes in result.
		*/
		lkupres2 = cache_lookup(rqstp->rq_vers, NFSPROC_GETATTR,
				(void *)&argp->file, (void *)&res_attr);
		if(!lkupres2) {
			if (((nfsproc_getattr_2(&argp->file, &res_attr, nfsdclient)) != RPC_SUCCESS) ||
					(res.status != NFS_OK)) {
				Dprintf(D_CACHE, "write allocate for acache error\n");
				goto no_caching;
			}
		}
		/* save previous file size for later use */
		old_size = res_attr.attrstat_u.attributes.size;
		Dprintf(D_CACHE, "old size: %d\n", old_size);
	}

	if(wb_cache_enabled) {
		/* update data cache:
		   attributes fetched from GETATTR cache are used to update attrs in matched
		   dcache block[s]; if a required block is not in cache, it will be fetched from
		   server, unless the write is appending to file, in which case, we can update
		   cache directly without asking it from server. However, we should be careful,
		   because the block may already exist in cache with some stale data due to
		   a setattr which has truncated the file before this write. So we still need look
		   up cache and update the block upon hit, or insert the block upon miss.
		*/
		warg = *argp;
		if(((warg.offset % MAX_NFS_BLOCKSIZE) + warg.data.data_len) >
					MAX_NFS_BLOCKSIZE) {
			/* write crossed boundary: update the first block */
			Dprintf(D_CACHE, "write crossed boundary\n");
			crossed_boundary = 1;
			requestcnt = MAX_NFS_BLOCKSIZE - (warg.offset % MAX_NFS_BLOCKSIZE);
			warg.totalcount = requestcnt;
			warg.data.data_len = requestcnt;
		}
		/* try update locally */
		result.attrstat.attrstat_u.attributes.size = old_size;
		lkupres = dcache_update(rqstp->rq_vers, rqstp->rq_proc,
				(void *)&warg, (void *)&result.attrstat);
		if(lkupres)
			Dprintf(D_CACHE, "update: offset %d, data_len %d, size %d\n",
					warg.offset, warg.totalcount,
		result.attrstat.attrstat_u.attributes.size);

		if(lkupres == -1) goto no_caching;
		else if(lkupres == 0) { /* not cached */
			/* allocate the entire first block */
			rarg.file = argp->file;
			rarg.offset = (argp->offset) - (argp->offset%MAX_NFS_BLOCKSIZE);
			rarg.count = MAX_NFS_BLOCKSIZE;
			rarg.totalcount = MAX_NFS_BLOCKSIZE;

			if(rarg.offset >= old_size || (!crossed_boundary &&
					argp->data.data_len == MAX_NFS_BLOCKSIZE)) {
				/* if the write is appending to the file, construct the block directly */
				Dprintf(D_CACHE, "new block (offset: %d)\n", rarg.offset);
				res_read.status = NFS_OK;
				res_read.readres_u.reply.attributes = res_attr.attrstat_u.attributes;
				res_read.readres_u.reply.data.data_len = 0;
			} else { /* else, fetch the block from server */
				forward = 1;
				res_read.readres_u.reply.data.data_val = data_buffer;
				if(((nfsproc_read_2(&rarg, &res_read, nfsdclient)) != RPC_SUCCESS) ||
						(res_read.status != NFS_OK)) {
					Dprintf(D_CACHE, "write allocate for dcache error\n");
					goto no_caching;
				}
				Dprintf(D_CACHE,"forward: offset %d,count %d, data_len %d\n",
						rarg.offset, rarg.count,
						res_read.readres_u.reply.data.data_len);
			}
			/* insert the first block into cache */
			insres = dcache_insert((void *)&rarg,rqstp->rq_vers,
					rqstp->rq_proc,(void *)&res_read);
			if(insres != 0) goto no_caching;
			/* update locally */
			lkupres = dcache_update(rqstp->rq_vers, rqstp->rq_proc,
					(void *)&warg, (void *)&result.attrstat);
			if(lkupres)
				Dprintf(D_CACHE, "update: offset %d, data_len %d, size %d\n",
						warg.offset, warg.totalcount,
						result.attrstat.attrstat_u.attributes.size);
			else {
				goto no_caching;
			}
		}

		if(crossed_boundary) { /* update second block */
			requestoffset = argp->offset + argp->data.data_len;
			warg.offset = requestoffset - (requestoffset % MAX_NFS_BLOCKSIZE);
			warg.beginoffset = warg.offset;
			warg.totalcount = requestoffset % MAX_NFS_BLOCKSIZE;
			warg.data.data_len = warg.totalcount;
			warg.data.data_val = data_buffer2;
			if(memcpy(warg.data.data_val, &(argp->data.data_val[requestcnt]),
					warg.totalcount) == NULL) {
				Dprintf(D_CACHE, "memcopy for data error\n");
				goto no_caching;
			}

			lkupres = dcache_update(rqstp->rq_vers, rqstp->rq_proc,
					(void *)&warg, (void *)&result.attrstat);
			if(lkupres)
				Dprintf(D_CACHE, "update: offset %d, data_len %d, size %d\n",
						warg.offset, warg.totalcount,
						result.attrstat.attrstat_u.attributes.size);

			if(lkupres == -1) /* error */
				goto no_caching;
			else if(lkupres == 0) { /* missed */
				/* allocate the entire second block*/
				rarg.file = argp->file;
				rarg.offset = warg.offset;
				rarg.count = MAX_NFS_BLOCKSIZE;
				rarg.totalcount = MAX_NFS_BLOCKSIZE;

				if(rarg.offset >= old_size) {
					/* if the write is appending to the file, contruct the block directly */
					Dprintf(D_CACHE, "new block (offset: %d)\n", rarg.offset);
							res_read.status = NFS_OK;
							res_read.readres_u.reply.attributes = res_attr.attrstat_u.attributes;
							res_read.readres_u.reply.data.data_len = 0;
				}
				else { /* else, fetch the block from server */
					forward = 1;
					res_read.readres_u.reply.data.data_val = data_cache_buffer;
					if(((nfsproc_read_2(&rarg, &res_read, nfsdclient)) != RPC_SUCCESS) ||
							(res_read.status != NFS_OK)) {
						Dprintf(D_CACHE, "write allocate for dcache error\n");
						goto no_caching;
					}
					Dprintf(D_CACHE,"forward: offset %d,count %d, data_len %d\n",
							rarg.offset, rarg.count,
							res_read.readres_u.reply.data.data_len);
				}
				/* insert the second block into cache */
				insres = dcache_insert((void *)&rarg,rqstp->rq_vers,
						rqstp->rq_proc,(void *)&res_read);
				if(insres != 0)
					goto no_caching;
				lkupres = dcache_update(rqstp->rq_vers, rqstp->rq_proc,
						(void *)&warg, (void *)&result.attrstat);
				if(lkupres)
					Dprintf(D_CACHE, "update: offset %d, data_len %d, size %d\n",
							warg.offset, warg.totalcount,
							result.attrstat.attrstat_u.attributes.size);
				else {
					goto no_caching;
				}
			}  /* end if lkureps == 0 */
		} /* end if crossed_bounday */

		if(!forward) {
#ifdef CALL_PROFILING
			log_cache_hits(rqstp->rq_proc); /*log cache hit */
#endif
		}

		new_size = result.attrstat.attrstat_u.attributes.size;
		/* store up-to-date attributes, returned from dcache lookup, into acache */
		res_attr.attrstat_u.attributes.size = new_size;
		if(lkupres2) {
			if(new_size > old_size) {
				lkupres2 = cache_update(rqstp->rq_vers, NFSPROC_GETATTR,
						(void *)&argp->file, (void *)&res_attr);
			}
		} else {
			insres = cache_insert(rqstp->rq_vers, NFSPROC_GETATTR,
					NULL, (void *)&res_attr);
			if(insres != 0) goto no_caching;
		}

		/* Fill the gap between the previous file's end and the new data's starting offset */
		if(new_size > old_size) {
			requestoffset = argp->offset - (argp->offset%MAX_NFS_BLOCKSIZE);
			if(requestoffset >= old_size) {
			/* the previous last block should be completed */
				if((old_size % MAX_NFS_BLOCKSIZE) != 0) {
					Dprintf(D_CACHE, "incomplete block at previous file end(%d)\n",old_size);
					warg.file = argp->file;
					warg.offset = old_size;
					warg.beginoffset = warg.offset;
					warg.totalcount = MAX_NFS_BLOCKSIZE - (warg.offset %
							MAX_NFS_BLOCKSIZE);
					warg.data.data_len = warg.totalcount;
					warg.data.data_val = data_buffer;
					memset(warg.data.data_val, 0, warg.totalcount);
					lkupres = dcache_update(rqstp->rq_vers, rqstp->rq_proc,
							(void *)&warg, (void *)&result.attrstat);
					if(lkupres == -1)
						dcache_inval(rqstp->rq_vers, rqstp->rq_proc, (void *)&warg);
				}

				/* the blocks between the previous file end and the new beginning should
				   be filled with zero bytes */
				warg.offset = old_size-1 - ((old_size-1) % MAX_NFS_BLOCKSIZE) +
						MAX_NFS_BLOCKSIZE;
				while(warg.offset < requestoffset) {
					warg.file = argp->file;
					warg.beginoffset = requestoffset;
					warg.totalcount = MAX_NFS_BLOCKSIZE;
					warg.data.data_len = MAX_NFS_BLOCKSIZE;
					warg.data.data_val = data_buffer;
					memset(warg.data.data_val, 0, warg.totalcount);
					lkupres = dcache_update(rqstp->rq_vers, rqstp->rq_proc,
							(void *)&warg, (void *)&result.attrstat);
					if(lkupres == -1)
						dcache_inval(rqstp->rq_vers, rqstp->rq_proc, (void *)&warg);
					if(lkupres == 0) {
						rarg.file = argp->file;
						rarg.offset = warg.offset;
						rarg.count = MAX_NFS_BLOCKSIZE;

						res_read.status = NFS_OK;
						res_read.readres_u.reply.attributes =
								result.attrstat.attrstat_u.attributes;
						res_read.readres_u.reply.data.data_len = MAX_NFS_BLOCKSIZE;
						res_read.readres_u.reply.data.data_val = data_buffer;
						memset(res_read.readres_u.reply.data.data_val, 0, MAX_NFS_BLOCKSIZE);
						insres = dcache_insert((void *)&rarg,rqstp->rq_vers,
								rqstp->rq_proc,(void *)&res_read);
						if(insres != 0) goto no_caching;
					}
					warg.offset += MAX_NFS_BLOCKSIZE;;
				}
			}
		} /* end if new_size > old_size) */
//    FORWARD_REQUEST(nfsproc_write_2, "nfsproc_write_2");
		goto out;
	} /* endif wb_cache_enabled */

no_caching: /* write-back disabled or cache error */
	/* Forward request to server */
	FORWARD_REQUEST(nfsproc_write_2, "nfsproc_write_2");
	NOTOK;

	/* Invalidate cache entries (both data and attribute) if necessary */
	if(data_cache_enabled) {
		dcache_inval(rqstp->rq_vers, rqstp->rq_proc, (void *)argp);
		if(old_size < argp->offset - argp->offset%MAX_NFS_BLOCKSIZE
				&& old_size%MAX_NFS_BLOCKSIZE != 0) {
			Dprintf(D_CACHE, "incomplete block at previous file end(%d)\n", old_size);
			argp->offset = old_size;
			argp->data.data_len = 1;
			dcache_inval(rqstp->rq_vers, rqstp->rq_proc, (void *)argp);
		}
	}

	if(attr_cache_enabled) { /* store returned attributes */
		lkupres2 = cache_update(rqstp->rq_vers, NFSPROC_GETATTR,
				(void *)&argp->file, (void *)&res);
		if(!lkupres2) {
			insres = cache_insert(rqstp->rq_vers, NFSPROC_GETATTR,
				                  NULL, (void *)&res);
		}
	}

	result.attrstat = res;
	Dprintf(D_CACHE, "forwarding ... done (size: %d)\n",
			result.attrstat.attrstat_u.attributes.size);

out:
	/* Map attributes and return to client */
	map_return_attributes(&result.attrstat.attrstat_u.attributes,
			nfsmount, rqstp);
	return (result);
}

union result_types
nfsd_nfsproc_create_2(createargs *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	diropres res;
	attrstat res_attr;
	int lkupres, insres;
	union result_types result;
	nfs_mount *nfsmount = NULL;

	Dprintf(D_CACHE, "*** CREATE: %s\n", argp->where.name);
	hexdump(argp->where.dir.data, NFS_FHSIZE, D_CACHE);
	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	READONLY;

	/* Forward request to server */
	FORWARD_REQUEST(nfsproc_create_2, "nfsproc_create_2");
	NOTOK;

	hexdump(res.diropres_u.diropres.file.data, NFS_FHSIZE, D_CACHE);

	if(attr_cache_enabled) { /* insert created file handle into LOOKUP cache */
		lkupres = cache_update(rqstp->rq_vers, NFSPROC_LOOKUP, (void *)&argp->where, (void *)&res);
		if(!lkupres)
			cache_insert(rqstp->rq_vers, NFSPROC_LOOKUP, NULL, (void *)&res);

		res_attr.status = res.status;
		res_attr.attrstat_u.attributes = res.diropres_u.diropres.attributes;
		insres = cache_insert(rqstp->rq_vers, NFSPROC_GETATTR,
				(void *)&res.diropres_u.diropres.file, (void *)&res_attr);
	}

	/* Map attributes and return to client */
	result.diropres = res;

	map_return_attributes(&result.diropres.diropres_u.diropres.attributes,
			nfsmount, rqstp);
	return (result);
}

union result_types
nfsd_nfsproc_remove_2(diropargs *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	nfsstat res;
	int lkupres = 0;
	diropres res_dir;
	attrstat res_attr;
	nfs_fh *file;
	fh_node *node;
	union result_types result;
	nfs_mount *nfsmount = NULL;

	Dprintf(D_CACHE, "**** REMOVE: %s\n", argp->name);
	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	READONLY;

	if(attr_cache_enabled) {
		/* fetch file handle from cache or server (upon miss) */
		lkupres = cache_lookup(rqstp->rq_vers, NFSPROC_LOOKUP,
			                   (void *)argp, (void *)&res_dir);
		if(lkupres == 0) { /* need the file handle to invalidate cache */
			if((nfsproc_lookup_2(argp, &res_dir, nfsdclient)) != RPC_SUCCESS) {
				Dprintf(D_CACHE, "lookup %s error\n", argp->name);
				result.nfsstat = NFSERR_ACCES;
				return result;
			}
		}
		file = &res_dir.diropres_u.diropres.file;
		hexdump(file->data, NFS_FHSIZE, D_CACHE);
	}

	/* Forward request to server */
	FORWARD_REQUEST(nfsproc_remove_2, "nfsproc_remove_2");
	if (res != NFS_OK) {
		result.nfsstat = res;
		return result;
	}

	if(attr_cache_enabled) {
		if(wb_cache_enabled) {
			lkupres = cache_lookup(rqstp->rq_vers, NFSPROC_GETATTR,
								   (void *)file, (void *)&res_attr);
			if(lkupres) {
				/* preserve locally modified file size */
				if(res_attr.attrstat_u.attributes.nlink > 1) {
					res_attr.attrstat_u.attributes.nlink--;
					cache_update(rqstp->rq_vers, NFSPROC_GETATTR,
								 (void *)file, (void *)&res_attr);
				}
			} else {
				if(nfsproc_getattr_2(file, &res_attr, nfsdclient) != RPC_SUCCESS) {
					Dprintf(D_CACHE, "getattr from server error %d\n");
					result.nfsstat = NFSERR_ACCES;
					return result;
				}
			}
			/* add it into stale_fh_hash */
			if(res_attr.attrstat_u.attributes.nlink == 1) {
				node = (fh_node *)malloc(sizeof(fh_node));
				node->file.handle = *file;
				node->status = FH_STALE;
				fh_hash_insert(stale_fh_hash, node);
			}
		} else
			invalidate_nfs_fh_v2(file);

		res_dir.status = NFSERR_NOENT; /* store negative cache */
		cache_update(rqstp->rq_vers, NFSPROC_LOOKUP, (void *)argp, (void *)&res_dir);
	}

	/* No need to map attributes */
	result.nfsstat = res;

	return (result);
}

union result_types
nfsd_nfsproc_rename_2(renameargs *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	nfsstat res;
	diropres res_dir;
	attrstat res_attr;
	int lkupres = 0;
	nfs_fh *file = NULL;
	fh_node *node;
	union result_types result;
	nfs_mount *nfsmount = NULL;

	Dprintf(D_CACHE, "*** RENAME: %s -> %s\n", argp->from.name, argp->to.name);
	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	READONLY;

	if(attr_cache_enabled) { /* fetch file handle of dest file */
		lkupres = cache_lookup(rqstp->rq_vers, NFSPROC_LOOKUP,
				 (void *)&argp->to, (void *)&res_dir);
		if(lkupres == 0) {
			if((nfsproc_lookup_2(&argp->to, &res_dir, nfsdclient)) != RPC_SUCCESS) {
				Dprintf(D_CACHE, "lookup %s error\n", argp->to.name);
				result.nfsstat = NFSERR_ACCES;
				return result;
			}
		}
		if(res_dir.status == NFS_OK) { /* not negative cache */
			file = &res_dir.diropres_u.diropres.file;
			hexdump(file->data, NFS_FHSIZE, D_CACHE);
		}
	}

	/* Forward request to server */
	FORWARD_REQUEST(nfsproc_rename_2, "nfsproc_rename_2");
	if (res != NFS_OK) {
		result.nfsstat = res;
		return result;
	}

	if(attr_cache_enabled) {
		if(file != NULL) { /* in case nfs does not remove the file explicitly */
			if (wb_cache_enabled) {
				lkupres = cache_lookup(rqstp->rq_vers, NFSPROC_GETATTR,
									   (void *)&file, (void *)&res_attr);
				if(lkupres) {
					/* preserve locally modified file size */
					if(res_attr.attrstat_u.attributes.nlink > 1) {
						res_attr.attrstat_u.attributes.nlink--;
						cache_update(rqstp->rq_vers, NFSPROC_GETATTR,
									 (void *)file, (void *)&res_attr);
					}
				} else {
					if(nfsproc_getattr_2(file, &res_attr, nfsdclient) != RPC_SUCCESS) {
						Dprintf(D_CACHE, "getattr from server error %d\n");
						result.nfsstat = NFSERR_ACCES;
						return result;
					}
				}
				/* add to stale_fh_hash */
				if(res_attr.attrstat_u.attributes.nlink == 1) {
					node = (fh_node *)malloc(sizeof(fh_node));
					node->file.handle = *file;
					node->status = FH_STALE;
					fh_hash_insert(stale_fh_hash, node);
				}
			} else
				invalidate_nfs_fh_v2(file);
		}

		invalidate_diroparg_v2(&argp->to);

		res_dir.status = NFSERR_NOENT;
		lkupres = cache_update(rqstp->rq_vers, NFSPROC_LOOKUP, (void *)&argp->from,
			                   (void *)&res_dir);
	}

	/* No need to map attributes */
	result.nfsstat = res;
	return (result);
}

union result_types
nfsd_nfsproc_link_2(linkargs *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	nfsstat res;
	int lkupres = 0;
	diropres res_dir;
	union result_types result;
	nfs_mount *nfsmount = NULL;

	Dprintf(D_CACHE, "*** LINK: %s\n", argp->to.name);
	hexdump(argp->from.data, NFS_FHSIZE, D_CACHE);
	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	READONLY;

	if(wb_cache_enabled) { /* write-through with write-allocate */
		lkupres = cache_lookup(rqstp->rq_vers, NFSPROC_GETATTR,
	                           (void *)argp->from.data, (void *)&result.attrstat);
	}

	/* Forward request to server */
	FORWARD_REQUEST(nfsproc_link_2, "nfsproc_link_2");
	if(res != NFS_OK) {
		result.nfsstat = res;
		return (result);
	}

	/* Invalidate cache entries if necessary */
	if(attr_cache_enabled) {
		res_dir.status = NFS_OK;
		res_dir.diropres_u.diropres.file = argp->from;
		lkupres = cache_update(rqstp->rq_vers, NFSPROC_LOOKUP, (void *)&argp->to,
			                   (void *)&res_dir);
		if(!lkupres)
			cache_insert(rqstp->rq_vers, NFSPROC_LOOKUP, NULL, (void *)&res_dir);

		if(wb_cache_enabled && lkupres) { /* keep locally modified file size */
			result.attrstat.attrstat_u.attributes.nlink++;
			cache_update(rqstp->rq_vers, NFSPROC_GETATTR, (void *)&argp->from, (void *)&result.attrstat);
		} else
			invalidate_nfs_fh_v2(&argp->from);

		invalidate_nfs_fh_v2(&argp->to.dir);
	}

	/* No need to map attributes */
	result.nfsstat = res;
	return (result);
}

union result_types
nfsd_nfsproc_symlink_2(symlinkargs *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	nfsstat res;
	union result_types result;
	nfs_mount *nfsmount = NULL;

	Dprintf(D_CACHE, "*** SYMLINK: from %s to %s\n", argp->from.name, argp->to);
	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	READONLY;

	/* Forward request to server */
	FORWARD_REQUEST(nfsproc_symlink_2, "nfsproc_symlink_2");
	if (res != NFS_OK) {
		result.nfsstat = res;
		return result;
	}

	/* Invalidate cache entries if necessary */
	if(attr_cache_enabled)
		cache_inval(rqstp->rq_vers, rqstp->rq_proc, (void *)argp);

	/* No need to map attributes */
	result.nfsstat = res;
	return (result);
}

union result_types
nfsd_nfsproc_mkdir_2(createargs *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	diropres res;
	int lkupres, insres;
	attrstat res_attr;
	union result_types result;
	nfs_mount *nfsmount = NULL;

	Dprintf(D_CACHE, "*** MKDIR: %s\n", argp->where.name);
	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	READONLY;

	/* Forward request to server */
	FORWARD_REQUEST(nfsproc_mkdir_2, "nfsproc_mkdir_2");
	NOTOK;

	hexdump(res.diropres_u.diropres.file.data, NFS_FHSIZE, D_CACHE);

	if(attr_cache_enabled) { /* insert created file handle into LOOKUP cache */
		lkupres = cache_update(rqstp->rq_vers, NFSPROC_LOOKUP, (void *)&argp->where, (void *)&res);
		if(!lkupres)
			cache_insert(rqstp->rq_vers, NFSPROC_LOOKUP, NULL, (void *)&res);
		/* also insert created dir's attributes into GETATTR cache */
		res_attr.status = res.status;
		res_attr.attrstat_u.attributes = res.diropres_u.diropres.attributes;
		insres = cache_insert(rqstp->rq_vers, NFSPROC_GETATTR,
			                  (void *)&res.diropres_u.diropres.file, (void *)&res_attr);
	}

	/* Map attributes and return to client */
	result.diropres = res;

	map_return_attributes(&result.diropres.diropres_u.diropres.attributes,
			nfsmount, rqstp);
	return (result);
}

union result_types
nfsd_nfsproc_rmdir_2(diropargs *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	nfsstat res;
	diropres res_dir;
	int lkupres;
	nfs_fh file;
	union result_types result;
	nfs_mount *nfsmount = NULL;

	Dprintf(D_CACHE, "*** RMDIR: %s\n", argp->name);
	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	READONLY;

	if(attr_cache_enabled) {
		/* fetch file handle from cache or server (upon miss) */
		lkupres = cache_lookup(rqstp->rq_vers, NFSPROC_LOOKUP,
				 (void *)argp, (void *)&res_dir);
		if(lkupres == 0) {
			if((nfsproc_lookup_2(argp, &res_dir, nfsdclient)) != RPC_SUCCESS) {
				Dprintf(D_CACHE, "lookup %s error\n", argp->name);
				result.nfsstat = NFSERR_ACCES;
				return result;
			}
		}
		file = res_dir.diropres_u.diropres.file;
	}

	/* Forward request to server */
	FORWARD_REQUEST(nfsproc_rmdir_2, "nfsproc_rmdir_2");
	if (res != NFS_OK) {
		result.nfsstat = res;
		return result;
	}

	if(attr_cache_enabled) {
		invalidate_nfs_fh_v2(&file);

		res_dir.status = NFSERR_NOENT;
		lkupres = cache_update(rqstp->rq_vers, NFSPROC_LOOKUP, (void *)argp,
						(void *)&res_dir);
	}

	/* No need to map attributes */
	result.nfsstat = res;
	return (result);
}

union result_types
nfsd_nfsproc_readdir_2(readdirargs *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	readdirres res;
	union result_types result;
	nfs_mount *nfsmount = NULL;
	char *buffer = (char *)malloc(argp->count);
	memset(buffer, 0, argp->count);
	res.readdirres_u.reply.entries = (entry *)buffer;

	Dprintf(D_CACHE, "*** READDIR (%d):\n", argp->count);
	hexdump(argp->dir.data, NFS_FHSIZE, D_CACHE);

	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	/* Forward request to server */
	FORWARD_REQUEST(nfsproc_readdir_2, "nfsproc_readdir_2");
	NOTOK;

	/* No need to map attributes */
	result.readdirres = res;
	result.nfsstat = res.status;
	return (result);
}

union result_types
nfsd_nfsproc_statfs_2(nfs_fh *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	statfsres res;
	union result_types result;
	nfs_mount *nfsmount = NULL;

	Dprintf(D_CACHE, "*** STATFS:\n");
	hexdump(argp->data, NFS_FHSIZE, D_CACHE);

	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	/* Forward request to server */
	FORWARD_REQUEST(nfsproc_statfs_2, "nfsproc_statfs_2");
	NOTOK;

	/* No need to map attributes */
	result.statfsres = res;
	result.nfsstat = res.status;
	return (result);
}

union result_types
nfsd_nfsproc_null_3(void *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	getattr3res res;
	union result_types result;
	nfs_mount *nfsmount = NULL;

	Dprintf(D_CACHE, "*** NULL:\n");
	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	/* Forward request to server */
	FORWARD_REQUEST3(nfsproc3_null_3, "nfsproc_null_3");
	NOTOK;

	result.nfsstat = NFS_OK;
	/* No need to map attributes in NULL call */
	return (result);
}

union result_types
nfsd_nfsproc_getattr_3(nfs_fh3 *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	getattr3res res;
	int lkupres = 0, insres;
	union result_types result;
	nfs_mount *nfsmount = NULL;

	Dprintf(D_CACHE, "*** GETATTR:\n");
	hexdump(argp->data.data_val, argp->data.data_len, D_CACHE);
	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	if(attr_cache_enabled) { /* look up cache; retrieve results if hit */
		lkupres = cache_lookup(rqstp->rq_vers, rqstp->rq_proc,
				 (void *)argp, (void *)&result.getattr3res);
		if(lkupres==1) {
#ifdef CALL_PROFILING /* log cache hit */
			log_cache_hits(rqstp->rq_proc);
#endif
		}
	}
	if(lkupres != 1) { /* cache disabled or cache miss; forward request to server */
		/* Forward request to server */
		FORWARD_REQUEST3(nfsproc3_getattr_3, "nfsproc_getattr_3");
		NOTOK;

		if(attr_cache_enabled) /* insert RPC response in cache */
			insres = cache_insert(rqstp->rq_vers, rqstp->rq_proc, NULL, (void *)&res);
		result.getattr3res = res;
	}

	Dprintf(D_CACHE, "current file size %llu, uid %d, gid %d, mtime %d\n",
			result.getattr3res.getattr3res_u.attributes.size,
			result.getattr3res.getattr3res_u.attributes.uid,
			result.getattr3res.getattr3res_u.attributes.gid,
			result.getattr3res.getattr3res_u.attributes.mtime);

	/* Map attributes and return to client */
	map_return_attributes3(&result.getattr3res.getattr3res_u.attributes,
			 nfsmount, rqstp);
	return (result);
}

union result_types
nfsd_nfsproc_setattr_3(setattr3args *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	wccstat3 res;
	getattr3res res_attr;
	access3res res_access;
	int lkupres = 0, insres;
	union result_types result;
	nfs_mount *nfsmount = NULL;
	fh_node *node;
	uint32 cuid, cgid;

	Dprintf(D_CACHE, "*** SETATTR:\n");
	hexdump(argp->object.data.data_val, argp->object.data.data_len, D_CACHE);

	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	READONLY;

	if(wb_cache_enabled) { /* write-through with write-allocate */
		if(argp->new_attributes.size.set && argp->new_attributes.size.set_uint64_u.val == 0) {
			Dprintf(D_CACHE, "nfsd_nfsproc_setattr_3: truncate file\n");
			node = (fh_node *)malloc(sizeof(fh_node));
			node->file.handle3.length = argp->object.data.data_len;
			memset(node->file.handle3.handle, 0, NFS3_FHSIZE);
			memcpy(node->file.handle3.handle, argp->object.data.data_val, argp->object.data.data_len);
			node->status = FH_TRUNC;
			fh_hash_insert(stale_fh_hash, node);
		} else
			lkupres = cache_lookup(rqstp->rq_vers, NFSPROC3_GETATTR,
				                   (void *)&argp->object, (void *)&result.getattr3res);
	}

	/* Forward request to server */
	if(argp->new_attributes.uid.set) {
		cuid = (uint32) luid(argp->new_attributes.uid.set_uint32_u.val, nfsmount, rqstp);
		Dprintf(D_CACHE, "map uid in args from %u to %u\n",
			    argp->new_attributes.uid.set_uint32_u.val, cuid);
		argp->new_attributes.uid.set_uint32_u.val = cuid;
	}
	if(argp->new_attributes.uid.set) {
		cgid = (uint32) lgid(argp->new_attributes.gid.set_uint32_u.val, nfsmount, rqstp);
		Dprintf(D_CACHE, "map gid in args from %u to %u\n",
			    argp->new_attributes.gid.set_uint32_u.val, cgid);
		argp->new_attributes.gid.set_uint32_u.val = cgid;
	}
	FORWARD_REQUEST3(nfsproc3_setattr_3, "nfsproc_setattr_3");

	/* Map attributes and return to client */
	if (res.status != NFS3_OK) {
		/* map uid/gid even if this call failed (resfail) */
		map_return_attributes3(&res.wccstat3_u.wcc.after.post_op_attr_u.attributes,
				 nfsmount, rqstp);
		return(result);
	}

	if(attr_cache_enabled) {
		res_attr.status = NFS3_OK;
		res_attr.getattr3res_u.attributes = res.wccstat3_u.wcc.after.post_op_attr_u.attributes;
		if(wb_cache_enabled && lkupres) {
			Dprintf(D_CACHE, "Old size %llu, new size %llu",
						res_attr.getattr3res_u.attributes.size,
						result.getattr3res.getattr3res_u.attributes.size);
			res_attr.getattr3res_u.attributes.size = result.getattr3res.getattr3res_u.attributes.size;
		}
		if(argp->new_attributes.mode.set) { /* clear cached access */
			Dprintf(D_CACHE, "change mode: %d\n", argp->new_attributes.mode.set_uint32_u.val);
			res_access.status = NFS3_OK;
			res_access.access3res_u.resok.access = -1;
			res_access.access3res_u.resok.obj_attributes.present = TRUE;
			res_access.access3res_u.resok.obj_attributes.post_op_attr_u.attributes =
				res_attr.getattr3res_u.attributes;
			lkupres = cache_update(rqstp->rq_vers, NFSPROC3_ACCESS,
			                      (void *)&argp->object, (void *)&res_access);
		} else
			lkupres = cache_update(rqstp->rq_vers, NFSPROC3_GETATTR,
			                      (void *)&argp->object, (void *)&res_attr);
		if(!lkupres)
			insres = cache_insert(rqstp->rq_vers, NFSPROC3_GETATTR, NULL, (void *)&res_attr);
	}

	/* update invalidation list */
	if(inval_enabled == 1) {
		inval_insert(argp->object, NULL, client_id);
	}

	result.wccstat3 = res;
	map_return_attributes3(&result.wccstat3.wccstat3_u.wcc.after.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
	return (result);
}

union result_types
nfsd_nfsproc_lookup_3(diropargs3 *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	lookup3res res;
	int lkupres=0;
	getattr3res res_attr_obj, res_attr_dir;
	union result_types result;
	nfs_mount *nfsmount = NULL;
	char *fh3_buffer = (char *)malloc(NFS3_FHSIZE);
	memset(fh3_buffer, 0, NFS3_FHSIZE);
	res.lookup3res_u.resok.object.data.data_val = fh3_buffer;

	Dprintf(D_CACHE, "*** LOOKUP: %s:\n", argp->name);
	hexdump(argp->dir.data.data_val, argp->dir.data.data_len, D_CACHE);
	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	if(attr_cache_enabled) { /* look up cache; retrieve results if hit */
		result.lookup3res.lookup3res_u.resok.object.data.data_val = fh3_buffer;
		lkupres = cache_lookup(rqstp->rq_vers, rqstp->rq_proc,
				 (void *)argp, (void *)&result.lookup3res);

		if(lkupres) {
			if (result.lookup3res.status != NFS3_OK) { /* cached NOENT */
				Dprintf(D_CACHE, "file does not exist (status: %d)\n", result.lookup3res.status);
				lkupres = cache_lookup(rqstp->rq_vers, NFSPROC3_GETATTR,
					                   (void *)&argp->dir,
					                   (void *)&res_attr_dir);
				if(lkupres) {
					result.lookup3res.lookup3res_u.resfail.post_op_attr_u.attributes =
						res_attr_dir.getattr3res_u.attributes;
					result.lookup3res.lookup3res_u.resfail.present = TRUE;
					map_return_attributes3(&result.lookup3res.lookup3res_u.resfail.post_op_attr_u.attributes,
						                    nfsmount, rqstp);
				} else
					result.lookup3res.lookup3res_u.resfail.present = FALSE;
#ifdef CALL_PROFILING
				log_cache_hits(rqstp->rq_proc);
#endif
				return result;
			}

			lkupres = cache_lookup(rqstp->rq_vers, NFSPROC3_GETATTR,
			                        (void *)&result.lookup3res.lookup3res_u.resok.object,
			                        (void *)&res_attr_obj);
			if(lkupres) { /* keep consistent with GETATTR cache */
				result.lookup3res.lookup3res_u.resok.obj_attributes.post_op_attr_u.attributes
					= res_attr_obj.getattr3res_u.attributes;
				result.lookup3res.lookup3res_u.resok.obj_attributes.present = TRUE;
			} else
				result.lookup3res.lookup3res_u.resok.obj_attributes.present = FALSE;

			lkupres = cache_lookup(rqstp->rq_vers, NFSPROC3_GETATTR,
			                        (void *)&argp->dir, (void *)&res_attr_dir);
			if(lkupres) {
				result.lookup3res.lookup3res_u.resok.dir_attributes.post_op_attr_u.attributes
					= res_attr_dir.getattr3res_u.attributes;
				result.lookup3res.lookup3res_u.resok.dir_attributes.present = TRUE;
			} else
				result.lookup3res.lookup3res_u.resok.dir_attributes.present = FALSE;

#ifdef CALL_PROFILING
			log_cache_hits(rqstp->rq_proc);
#endif
			goto out;
		}
	}

	/* cache not enabled or cache miss; forward request to server */
	FORWARD_REQUEST(nfsproc3_lookup_3, "nfsproc_lookup_3");
	result.lookup3res = res;

	if(res.status != NFS3_OK) { /* Cache negative results */
		if(attr_cache_enabled && res.status == NFS3ERR_NOENT) {
			Dprintf(D_CACHE, "file does not exist\n");
			cache_insert(rqstp->rq_vers, rqstp->rq_proc, (void *)argp, (void *)&result.lookup3res);

			res_attr_dir.status = NFS3_OK;
			res_attr_dir.getattr3res_u.attributes = result.lookup3res.lookup3res_u.resfail.post_op_attr_u.attributes;
			lkupres = cache_update(rqstp->rq_vers, NFSPROC3_GETATTR,
						(void *)&argp->dir, (void *)&res_attr_dir);
			if(!lkupres)
				cache_insert(rqstp->rq_vers, NFSPROC3_GETATTR,
							(void *)&argp->dir, (void *)&res_attr_dir);
		}
		/* map uid/gid even if this call failed (resfail) */
		map_return_attributes3(&result.lookup3res.lookup3res_u.resfail.post_op_attr_u.attributes,
		                       nfsmount, rqstp);
		return(result);
	}

	if(attr_cache_enabled) {
		cache_insert(rqstp->rq_vers, rqstp->rq_proc, (void *)argp, (void *)&result.lookup3res);

		lkupres = cache_lookup(rqstp->rq_vers, NFSPROC3_GETATTR,
		                       (void *)&result.lookup3res.lookup3res_u.resok.object,
		                       (void *)&res_attr_obj);
		if(!lkupres) {
			res_attr_obj.status = NFS3_OK;
			res_attr_obj.getattr3res_u.attributes =
				result.lookup3res.lookup3res_u.resok.obj_attributes.post_op_attr_u.attributes;
			cache_insert(rqstp->rq_vers, NFSPROC3_GETATTR,
						(void *)&result.lookup3res.lookup3res_u.resok.object, (void *)&res_attr_obj);
		} else {
			if(wb_cache_enabled) /* use locally modified attributes */
				result.lookup3res.lookup3res_u.resok.obj_attributes.post_op_attr_u.attributes
						= res_attr_obj.getattr3res_u.attributes;
			else {
				res_attr_obj.status = NFS3_OK;
				res_attr_obj.getattr3res_u.attributes =
						result.lookup3res.lookup3res_u.resok.obj_attributes.post_op_attr_u.attributes;
				lkupres = cache_update(rqstp->rq_vers, NFSPROC3_GETATTR,
							 (void *)&result.lookup3res.lookup3res_u.resok.object, (void *)&res_attr_obj);
			}
		}

		res_attr_dir.status = NFS3_OK;
		res_attr_dir.getattr3res_u.attributes =
			result.lookup3res.lookup3res_u.resok.dir_attributes.post_op_attr_u.attributes;
		lkupres = cache_update(rqstp->rq_vers, NFSPROC3_GETATTR, (void *)&argp->dir, (void *)&res_attr_dir);
		if(!lkupres)
			cache_insert(rqstp->rq_vers, NFSPROC3_GETATTR, (void *)&argp->dir, (void *)&res_attr_dir);
	}

out:
	hexdump(result.lookup3res.lookup3res_u.resok.object.data.data_val, result.lookup3res.lookup3res_u.resok.object.data.data_len, D_CACHE);
	if(result.lookup3res.lookup3res_u.resok.obj_attributes.present)
		Dprintf(D_CACHE, "current file size %llu, mtime %d, link %d\n",
	        result.lookup3res.lookup3res_u.resok.obj_attributes.post_op_attr_u.attributes.size,
	        result.lookup3res.lookup3res_u.resok.obj_attributes.post_op_attr_u.attributes.mtime.seconds,
	        result.lookup3res.lookup3res_u.resok.obj_attributes.post_op_attr_u.attributes.nlink);

	map_return_attributes3(&result.lookup3res.lookup3res_u.resok.obj_attributes.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
	map_return_attributes3(&result.lookup3res.lookup3res_u.resok.dir_attributes.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
	return (result);
}

union result_types
nfsd_nfsproc_access_3(access3args *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	access3res res;
	union result_types result;
	nfs_mount *nfsmount = NULL;
	int lkupres = 0;

	Dprintf(D_CACHE, "*** ACCESS: %i\n", argp->access);
	hexdump(argp->object.data.data_val, argp->object.data.data_len, D_CACHE);
	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	if(attr_cache_enabled) { /* look up cache; retrieve results if hit */
		lkupres = cache_lookup(rqstp->rq_vers, rqstp->rq_proc,
		                       (void *)&argp->object, (void *)&result.access3res);
		if(lkupres) {
			if(result.access3res.access3res_u.resok.access != -1) { /* valid access cached */
#ifdef CALL_PROFILING /* log cache hit */
				log_cache_hits(rqstp->rq_proc);
#endif
				goto out;
			}
		}
	}

	/* cache disabled or cache miss; forward request to server */
	FORWARD_REQUEST3(nfsproc3_access_3, "nfsproc_access_3");

	/* Map attributes and return to client */
	if (res.status != NFS3_OK) {
		/* map uid/gid even if this call failed (resfail) */
		map_return_attributes3(&res.access3res_u.resfail.post_op_attr_u.attributes,
		                       nfsmount, rqstp);
		result.access3res = res;
		return (result);
	}

	if(attr_cache_enabled) { /* insert RPC response in cache */
		if(lkupres) {
			if(wb_cache_enabled) /* save locally modified file size */
				res.access3res_u.resok.obj_attributes.post_op_attr_u.attributes.size
					= result.access3res.access3res_u.resok.obj_attributes.post_op_attr_u.attributes.size;
			lkupres = cache_update(rqstp->rq_vers, rqstp->rq_proc, (void *)&argp->object, (void *)&res);
		} else
			cache_insert(rqstp->rq_vers, rqstp->rq_proc, NULL, (void *)&res);
	}

	result.access3res = res;


out:
	Dprintf(D_CACHE, "status: %i\n", result.access3res.status);
	Dprintf(D_CACHE, "access: %i\n", result.access3res.access3res_u.resok.access);
	map_return_attributes3(&result.access3res.access3res_u.resok.obj_attributes.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
	return (result);
}

union result_types
nfsd_nfsproc_readlink_3(nfs_fh3 *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	readlink3res res;
	getattr3res res_attr;
	int lkupres;
	union result_types result;
	nfs_mount *nfsmount = NULL;
	char *buffer = malloc(NFS_MAXPATHLEN);
	res.readlink3res_u.resok.data = buffer;

	Dprintf(D_CACHE, "*** READLINK:\n");
	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	/* Forward request to server */
	FORWARD_REQUEST3(nfsproc3_readlink_3, "nfsproc_readlink_3");

	/* Map attributes and return to client */
	result.readlink3res = res;
	if (res.status != NFS3_OK) {
		/* map uid/gid even if this call failed (resfail) */
		map_return_attributes3(&result.readlink3res.readlink3res_u.resfail.\
				 post_op_attr_u.attributes,
				 nfsmount, rqstp);
		return (result);
	}

	if(attr_cache_enabled) {
		res_attr.status = NFS_OK;
		res_attr.getattr3res_u.attributes =
			result.readlink3res.readlink3res_u.resok.symlink_attributes.post_op_attr_u.attributes;
		lkupres = cache_update(rqstp->rq_vers, NFSPROC3_GETATTR,
			                   (void *)argp, (void *)&res_attr);
		if(!lkupres)
			lkupres = cache_insert(rqstp->rq_vers, NFSPROC3_GETATTR,
			                       (void *)argp, (void *)&res_attr);
	}

	map_return_attributes3(&result.readlink3res.readlink3res_u.resok.\
			 symlink_attributes.post_op_attr_u.attributes,
			 nfsmount, rqstp);
	return (result);
}

union result_types
nfsd_nfsproc_read_3(read3args *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	read3res *res, read_res;
	getattr3res res_attr;
	int lkupres=0, lkupres2=0,insres, crossed_boundary = 0, forward = 0;
	int requestoffset;
	read3args args;
	char *data_buffer = (char *)malloc(MAX_NFS3_BLOCKSIZE);
	char data_buffer2[MAX_NFS3_BLOCKSIZE];
	union result_types result;
	nfs_mount *nfsmount = NULL;

	Dprintf(D_CACHE, "*** READ: offset %llu count %d\n", argp->offset, argp->count);
	hexdump(argp->file.data.data_val, argp->file.data.data_len, D_CACHE);
	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	result.read3res.read3res_u.resok.data.data_val = data_buffer;
	res = &result.read3res;

	if(!data_cache_enabled) goto no_caching;
	else { /* look up cache; retrieve results if hit */
		/* get attributes for validation of cached data */
		lkupres2 = cache_lookup(rqstp->rq_vers, NFSPROC3_GETATTR,
			 (void *)&argp->file, (void *)&res_attr);

		args = *argp;
		if(lkupres2 && args.offset >= res_attr.getattr3res_u.attributes.size) {
			Dprintf(D_CACHE, "offset is greater than file size\n");
			result.read3res.status = NFS3_OK;
			result.read3res.read3res_u.resok.count = 0;
			result.read3res.read3res_u.resok.eof = TRUE;
			result.read3res.read3res_u.resok.data.data_len = 0;
			result.read3res.read3res_u.resok.file_attributes.post_op_attr_u.attributes = res_attr.getattr3res_u.attributes;
			result.read3res.read3res_u.resok.file_attributes.present = TRUE;
			goto out;
		}

		if(((args.offset % MAX_NFS3_BLOCKSIZE) + args.count) > MAX_NFS3_BLOCKSIZE) {
			/* read crossed boundary: lookup the first block */
			Dprintf(D_CACHE, "read crossed boundary\n");
			crossed_boundary = 1;
			args.count = MAX_NFS3_BLOCKSIZE - (args.offset % MAX_NFS3_BLOCKSIZE);
		}

		/* try to fetch locally */
		lkupres = dcache_lookup(rqstp->rq_vers, rqstp->rq_proc,
					(void *)&args, (void *)&result.read3res);
		if(lkupres) {
			Dprintf(D_CACHE, "returned from cache: len %d, eof %d\n",
			        result.read3res.read3res_u.resok.data.data_len,
			result.read3res.read3res_u.resok.eof);
			if(result.read3res.status == 100) /* locally modified */
				result.readres.status = NFS_OK;
			else { /* must check modification time from server */
				if(!lkupres2) {
					if (nfsproc3_getattr_3(&argp->file, &res_attr, nfsdclient) != RPC_SUCCESS) {
						Dprintf(L_ERROR, "getattr from server error\n");
						result.nfsstat = NFS3ERR_ACCES;
						return result;
					}
					if(res_attr.status != NFS_OK) {
						result.nfsstat = res_attr.status;
						return result;
					}
				}
				if(result.read3res.read3res_u.resok.file_attributes.post_op_attr_u.attributes.mtime.seconds
						< res_attr.getattr3res_u.attributes.mtime.seconds) {
					Dprintf(D_CACHE, "cached data may be stale(%d:%d); flush\n",
						result.read3res.read3res_u.resok.file_attributes.post_op_attr_u.attributes.mtime.seconds,
						res_attr.getattr3res_u.attributes.mtime.seconds);
					dcache_inval(rqstp->rq_vers, rqstp->rq_proc, (void *)&args);
					lkupres = 0;
				}
			}
		}

		if(lkupres == -1) {
			Dprintf(L_ERROR, "dcache lookup error\n");
			goto no_caching; /* error */
		} else if(lkupres == 0) { /* not cached */
			/* fetch the entire first block */
			args.offset = (argp->offset) - (argp->offset%MAX_NFS3_BLOCKSIZE);
			args.count = MAX_NFS3_BLOCKSIZE;

			forward = 1;
			if(((nfsproc3_read_3(&args, res, nfsdclient)) != RPC_SUCCESS) ||
								(res->status != NFS_OK)) {
				Dprintf(L_ERROR, "read from server error\n");
				goto no_caching;
			}
			Dprintf(D_CACHE, "forward:offset %d, cnt %d, return:len %d, eof %d\n",
			        (int)args.offset, args.count, res->read3res_u.resok.data.data_len,
			        res->read3res_u.resok.eof);

			/* insert the first block into cache */
			if(res->read3res_u.resok.data.data_len > 0) {
				args.offset = argp->offset;
				if(crossed_boundary)
					args.count = MAX_NFS3_BLOCKSIZE - (args.offset % MAX_NFS3_BLOCKSIZE);
				else
					args.count = argp->count;
				dcache_insert((void *)&args, rqstp->rq_vers, rqstp->rq_proc, (void *)res);
			}
			result.read3res = *res;
		}

		if(crossed_boundary) { /* fetch second block */
			read_res = result.read3res;
			read_res.read3res_u.resok.data.data_val = data_buffer2;
			if(read_res.read3res_u.resok.data.data_len > 0)
				if(memcpy(read_res.read3res_u.resok.data.data_val,
									result.read3res.read3res_u.resok.data.data_val,
					 read_res.read3res_u.resok.data.data_len) == NULL) {
					Dprintf(L_ERROR, "error memory copy\n");
					goto no_caching;
				}
			Dprintf(D_CACHE, "got first part: data_len %d\n",
					read_res.read3res_u.resok.data.data_len);

			requestoffset = argp->offset + argp->count;
			args.offset = requestoffset - (requestoffset % MAX_NFS3_BLOCKSIZE);
			args.count = requestoffset % MAX_NFS3_BLOCKSIZE;
			/* try to fetch locally */
			lkupres = dcache_lookup(rqstp->rq_vers, rqstp->rq_proc,
						(void *)&args, (void *)&result.read3res);
			if(lkupres) {
				if(result.read3res.status == 100)
					result.read3res.status = NFS_OK;
				else
					if(result.read3res.read3res_u.resok.file_attributes.post_op_attr_u.attributes.mtime.seconds <
						res_attr.getattr3res_u.attributes.mtime.seconds) {
					Dprintf(D_CACHE, "cached data may be stale(%d:%d); flush\n",
					        result.read3res.read3res_u.resok.file_attributes.post_op_attr_u.attributes.mtime.seconds,
					        res_attr.getattr3res_u.attributes.mtime.seconds);
					dcache_inval(rqstp->rq_vers, rqstp->rq_proc, (void *)&args);
					lkupres = 0;
				}
			}
			if(lkupres == -1)
				goto no_caching;
			else if(lkupres == 0) { /* not cached */
				/* fetch the entire second block */
				args.count = MAX_NFS3_BLOCKSIZE;

				forward = 1;
				if(nfsproc3_read_3(&args, res, nfsdclient) != RPC_SUCCESS ||
				   res->status != NFS_OK) {
					Dprintf(L_ERROR, "read from server error\n");
					result.nfsstat = NFS3ERR_ACCES;
					return result;
				}
				Dprintf(D_CACHE, "forward:offset %llu cnt %d, return:len %d, eof %d\n",
				        args.offset, args.count, res->read3res_u.resok.data.data_len,
				        res->read3res_u.resok.eof);
				/* insert the second block into cache */
				if(res->read3res_u.resok.data.data_len > 0) {
					args.count = requestoffset % MAX_NFS3_BLOCKSIZE;
					insres = dcache_insert((void *)&args, rqstp->rq_vers, rqstp->rq_proc, (void *)res);
					if(insres != 0) {
						Dprintf(L_ERROR, "dcache insert error\n");
						goto no_caching;
					}
				}
				result.read3res = *res;
			}

			if(result.read3res.read3res_u.resok.data.data_len > 0)
				if(memcpy(&read_res.read3res_u.resok.data.data_val[read_res.read3res_u.resok.data.data_len],
				   result.read3res.read3res_u.resok.data.data_val,
			       result.read3res.read3res_u.resok.data.data_len) == NULL) {
					Dprintf(L_ERROR, "error memory copy\n");
					goto no_caching;
				}
			read_res.read3res_u.resok.data.data_len +=
				result.read3res.read3res_u.resok.data.data_len;
			read_res.read3res_u.resok.count = read_res.read3res_u.resok.data.data_len;
			read_res.read3res_u.resok.eof = result.read3res.read3res_u.resok.eof;
			Dprintf(D_CACHE, "got second part: data_len %d\n",
			        read_res.read3res_u.resok.data.data_len);

			result.read3res = read_res;
			result.read3res.read3res_u.resok.data.data_val = data_buffer;
			if(memcpy(result.read3res.read3res_u.resok.data.data_val,
			          read_res.read3res_u.resok.data.data_val,
				read_res.read3res_u.resok.data.data_len) == NULL) {
				Dprintf(L_ERROR, "error memory copy\n");
				goto no_caching;
			}
			Dprintf(D_CACHE, "total: data_len %d, eof %d\n",
			        read_res.read3res_u.resok.data.data_len,
			        read_res.read3res_u.resok.eof);
		} /* end if crossed boundary */

		if(!forward) {
#ifdef CALL_PROFILING
			log_cache_hits(rqstp->rq_proc); /*log cache hit */
#endif
		}

//     read_res.read3res_u.resok.data.data_val = cache_buffer;
//     if((nfsproc3_read_3(argp, &read_res, nfsdclient)) != RPC_SUCCESS) {
//       clnt_perror(nfsdclient,"nfsproc_read_3");
//       exit(0);
//     }
//
//     if(read_res.read3res_u.resok.data.data_len ==
//                           result.read3res.read3res_u.resok.data.data_len) {
//       if(memcmp(read_res.read3res_u.resok.data.data_val,
//                   result.read3res.read3res_u.resok.data.data_val,
//                   result.read3res.read3res_u.resok.data.data_len) != 0) {
//         Dprintf(D_CACHE, "data contents do not match\n");
//         exit(0);
//       }
//     }
//     else {
//       Dprintf(D_CACHE, "data lengths do not match\n");
//       exit(0);
//     }
//     free(read_res.read3res_u.resok.data.data_val);

		goto out;
	} /* endif  data_cache_enabled */

no_caching:
	/* Forward request to server */
	if((nfsproc3_read_3(argp, res, nfsdclient)) != RPC_SUCCESS) {
		clnt_perror(nfsdclient,"nfsproc_read_3");
		result.nfsstat = NFS3ERR_ACCES;
		return result;
	}

	/* Map attributes and return to client */
	if (result.read3res.status != NFS3_OK) {
		/* map uid/gid even if this call failed (resfail) */
		map_return_attributes3(&result.read3res.read3res_u.resfail.\
				 post_op_attr_u.attributes,
				 nfsmount, rqstp);
		return (result);
	}

	Dprintf(D_CACHE, "forwarding ... returned data_len %d\n",
	        result.read3res.read3res_u.resok.count);

out:
	if(attr_cache_enabled) {
		if(!lkupres2) {
			res_attr.status = NFS3_OK;
			res_attr.getattr3res_u.attributes =
				result.read3res.read3res_u.resok.file_attributes.post_op_attr_u.attributes;
			cache_insert(rqstp->rq_vers, NFSPROC3_GETATTR, (void *)&argp->file, (void *)&res_attr);
		} else {
			if(wb_cache_enabled) { /* return fresh attributes from acache */
				result.read3res.read3res_u.resok.file_attributes.post_op_attr_u.attributes.size
					= res_attr.getattr3res_u.attributes.size;
			} else {
				if(lkupres2) {
					if(forward)
						cache_update(rqstp->rq_vers, NFSPROC_GETATTR, (void *)&argp->file, (void *)&res_attr);
				} else {
					cache_insert(rqstp->rq_vers, NFSPROC3_GETATTR, (void *)&argp->file, (void *)&res_attr);
				}
			}
		}
	}

	Dprintf(D_CACHE, "file size %llu\n",
	        result.read3res.read3res_u.resok.file_attributes.post_op_attr_u.attributes.size);

	map_return_attributes3(&result.read3res.read3res_u.resok.file_attributes.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
	return (result);
}

union result_types
nfsd_nfsproc_write_3(write3args *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	write3res res;
	getattr3res res_attr;
	read3args rarg;
	write3args warg;
	read3res res_read;
	int lkupres = 0, lkupres2 = 0, insres, new_size, old_size=0, requestoffset, requestcnt = 0;
	int forward = 0, crossed_boundary = 0;
	union result_types result;
	nfs_mount *nfsmount = NULL;
	char data_buffer[MAX_NFS3_BLOCKSIZE], data_buffer2[MAX_NFS3_BLOCKSIZE];

	Dprintf(D_CACHE, "*** WRITE: offset %llu, len %d\n", argp->offset, argp->data.data_len);
	hexdump(argp->file.data.data_val, argp->file.data.data_len, D_CACHE);
	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	READONLY;

	if(data_cache_enabled) {
		lkupres2 = cache_lookup(rqstp->rq_vers, NFSPROC3_GETATTR,
		                        (void *)&argp->file, (void *)&res_attr);
		if(!lkupres2) {
			if ((nfsproc3_getattr_3(&argp->file, &res_attr, nfsdclient)) != RPC_SUCCESS) {
				Dprintf(D_CACHE, "getattr from server error\n");
				result.nfsstat = NFS3ERR_ACCES;
				return result;
			}
			if(res_attr.status != NFS3_OK) {
				result.nfsstat = res_attr.status;
				return result;
			}
		}
		/* save previous file size for later use */
		old_size = res_attr.getattr3res_u.attributes.size;
		Dprintf(D_CACHE, "old size: %d\n", old_size);
	}

	if(wb_cache_enabled) {
		warg = *argp;
		if(((warg.offset % MAX_NFS3_BLOCKSIZE) + warg.data.data_len) > MAX_NFS3_BLOCKSIZE) {
			/* write crossed boundary: update the first block */
			Dprintf(D_CACHE, "write crossed boundary\n");
			crossed_boundary = 1;
			requestcnt = MAX_NFS3_BLOCKSIZE - (warg.offset % MAX_NFS3_BLOCKSIZE);
			warg.count = requestcnt;
			warg.data.data_len = requestcnt;
		}
		/* try update locally */
		result.write3res.write3res_u.resok.file_wcc.after.post_op_attr_u.attributes = res_attr.getattr3res_u.attributes;
		lkupres = dcache_update(rqstp->rq_vers, rqstp->rq_proc,
					(void *)&warg, (void *)&result.write3res);
		if(lkupres)
			Dprintf(D_CACHE, "update: offset %llu, data_len %d, size %llu\n",
			        warg.offset, warg.count,
			        result.write3res.write3res_u.resok.file_wcc.after.post_op_attr_u.attributes.size);

		if(lkupres == -1) goto no_caching;
		else if(lkupres == 0) { /* not cached */
			/* allocate the entire first block */
			rarg.file = argp->file;
			rarg.offset = (argp->offset) - (argp->offset%MAX_NFS3_BLOCKSIZE);
			rarg.count = MAX_NFS3_BLOCKSIZE;

			if(rarg.offset >= old_size || (!crossed_boundary && argp->data.data_len == MAX_NFS3_BLOCKSIZE)) {
				/* if the write is appending to the file, construct the block directly */
				Dprintf(D_CACHE, "append or replace a complete block (offset: %llu)\n",
				        rarg.offset);
				res_read.status = NFS3_OK;
				res_read.read3res_u.resok.count = 0;
				res_read.read3res_u.resok.data.data_len = 0;
				if(rarg.offset >= old_size)  res_read.read3res_u.resok.eof = TRUE;
				else res_read.read3res_u.resok.eof = FALSE;
				res_read.read3res_u.resok.file_attributes.post_op_attr_u.attributes = res_attr.getattr3res_u.attributes;
				res_read.read3res_u.resok.file_attributes.present = TRUE;
			} else if(lkupres == 0) { /* else, fetch the block from server */
				res_read.read3res_u.resok.data.data_val = data_buffer;
				forward=1;
				if(nfsproc3_read_3(&rarg, &res_read, nfsdclient) != RPC_SUCCESS ||
				   res_read.status != NFS_OK) {
					Dprintf(D_CACHE, "write allocate for dcache error\n");
					result.nfsstat = NFS3ERR_ACCES;
					return result;
				}
				Dprintf(D_CACHE,"forward: offset %llu, count %d, data_len %d\n",
				        rarg.offset, rarg.count,
				        res_read.read3res_u.resok.data.data_len);
			}
			/* insert the first block into cache */
			insres = dcache_insert((void *)&rarg, rqstp->rq_vers, rqstp->rq_proc, (void *)&res_read);
			if(insres != 0)	goto no_caching;

			/* update locally */
			lkupres = dcache_update(rqstp->rq_vers, rqstp->rq_proc,
			                        (void *)&warg, (void *)&result.write3res);
			if(lkupres)
				Dprintf(D_CACHE, "update: offset %llu, data_len %d, size %llu\n",
				        warg.offset, warg.count,
				        result.write3res.write3res_u.resok.file_wcc.after.post_op_attr_u.attributes.size);
			else
				goto no_caching;
		}

		if(crossed_boundary) { /* update second block */
			requestoffset = argp->offset + argp->data.data_len;
			warg.offset = requestoffset - requestoffset % MAX_NFS3_BLOCKSIZE;
			warg.count = requestoffset % MAX_NFS3_BLOCKSIZE;
			warg.data.data_len = warg.count;
			warg.data.data_val = data_buffer2;
			if(memcpy(warg.data.data_val, &(argp->data.data_val[requestcnt]), warg.count) == NULL) {
				Dprintf(D_CACHE, "memcopy for data error\n");
				goto no_caching;
			}

			lkupres = dcache_update(rqstp->rq_vers, rqstp->rq_proc,
					(void *)&warg, (void *)&result.write3res);
			if(lkupres)
				Dprintf(D_CACHE, "update: offset %llu, data_len %d, size %llu\n",
				        warg.offset, warg.count,
				        result.write3res.write3res_u.resok.file_wcc.after.post_op_attr_u.attributes.size);
			if(lkupres == -1) /* error */
				goto no_caching;
			else if(lkupres == 0) { /* missed */
				/* allocate the entire second block*/
				rarg.file = argp->file;
				rarg.offset = warg.offset;
				rarg.count = MAX_NFS3_BLOCKSIZE;

				if(rarg.offset >= old_size) {
					/* if the write is appending to the file, contruct the block directly */
					Dprintf(D_CACHE, "new block (offset: %d)\n", rarg.offset);
					res_read.status = NFS3_OK;
					res_read.read3res_u.resok.file_attributes.post_op_attr_u.attributes = res_attr.getattr3res_u.attributes;
					res_read.read3res_u.resok.file_attributes.present = TRUE;
					res_read.read3res_u.resok.count = 0;
					res_read.read3res_u.resok.data.data_len = 0;
					if(rarg.offset >= old_size)  res_read.read3res_u.resok.eof = TRUE;
					else res_read.read3res_u.resok.eof = FALSE;
				} else { /* else, fetch the block from server */
					res_read.read3res_u.resok.data.data_val = data_buffer;
					forward = 1;
					if(((nfsproc3_read_3(&rarg, &res_read, nfsdclient)) != RPC_SUCCESS) ||
							(res_read.status != NFS3_OK)) {
						Dprintf(D_CACHE, "write allocate for dcache error\n");
						result.nfsstat = NFS3ERR_ACCES;
						return result;
					}
					Dprintf(D_CACHE,"forward: offset %llu, count %d, data_len %d\n",
					        rarg.offset, rarg.count,
					        res_read.read3res_u.resok.data.data_len);
				}
				/* insert the second block into cache */
				insres = dcache_insert((void *)&rarg, rqstp->rq_vers,  rqstp->rq_proc, (void *)&res_read);
				if(insres != 0) goto no_caching;

				lkupres = dcache_update(rqstp->rq_vers, rqstp->rq_proc,
				                        (void *)&warg, (void *)&result.write3res);
				if(lkupres)
					Dprintf(D_CACHE, "update: offset %llu, data_len %d, size %llu\n",
					        warg.offset, warg.count,
					        result.write3res.write3res_u.resok.file_wcc.after.post_op_attr_u.attributes.size);
				else
					goto no_caching;
			}  /* end if lkureps == 0 */
		} /* end if crossed_bounday */

		if(!forward) {
#ifdef CALL_PROFILING
			log_cache_hits(rqstp->rq_proc); /*log cache hit */
#endif
		}

		/* store up-to-date attributes, returned from dcache lookup, into acache */
		new_size = result.write3res.write3res_u.resok.file_wcc.after.post_op_attr_u.attributes.size;
		res_attr.getattr3res_u.attributes.size = new_size;
		if (lkupres2) {
			if(new_size > old_size)
				/* look up cache; update cache if hit */
				cache_update(rqstp->rq_vers, NFSPROC3_GETATTR,
				             (void *)&argp->file, (void *)&res_attr);
		} else
			cache_insert(rqstp->rq_vers, NFSPROC3_GETATTR, NULL, (void *)&res_attr);

		/* Fill the gap between the previous file's end and the new data's starting offset */
		if(new_size > old_size) {
			requestoffset = argp->offset - (argp->offset%MAX_NFS3_BLOCKSIZE);
			if(requestoffset >= old_size) {
				/* the previous last block should be completed */
				if((old_size % MAX_NFS3_BLOCKSIZE) != 0) {
					Dprintf(D_CACHE, "incomplete block at previous file end (%llu)\n",
							old_size);
					warg.file = argp->file;
					warg.offset = old_size;
					warg.count = MAX_NFS3_BLOCKSIZE - (warg.offset % MAX_NFS3_BLOCKSIZE);
					warg.data.data_len = warg.count;
					warg.data.data_val = data_buffer;
					memset(warg.data.data_val, 0, warg.count);
					dcache_update(rqstp->rq_vers, rqstp->rq_proc, (void *)&warg, (void *)&result.write3res);
				}

				/* the blocks between the previous file end and the new beginning should
				   be filled with zero bytes. assuming latency is more important than space, we also store these
				   all-zero blocks in cache.
				 */
				warg.offset = old_size / MAX_NFS3_BLOCKSIZE * MAX_NFS3_BLOCKSIZE;
				if(old_size % MAX_NFS3_BLOCKSIZE > 0) warg.offset += MAX_NFS3_BLOCKSIZE;
				while(warg.offset < requestoffset) {
					warg.file = argp->file;
					warg.count = MAX_NFS3_BLOCKSIZE;
					warg.data.data_len = MAX_NFS3_BLOCKSIZE;
					warg.data.data_val = data_buffer;
					memset(warg.data.data_val, 0, warg.count);
					lkupres = dcache_update(rqstp->rq_vers, rqstp->rq_proc,
											(void *)&warg, (void *)&result.write3res);
					if(lkupres == -1)
						dcache_inval(rqstp->rq_vers, rqstp->rq_proc, (void *)&warg);
					else if(lkupres == 0) {
						rarg.file = argp->file;
						rarg.offset = warg.offset;
						rarg.count = MAX_NFS3_BLOCKSIZE;

						res_read.status = NFS3_OK;
						res_read.read3res_u.resok.count = MAX_NFS3_BLOCKSIZE;
						res_read.read3res_u.resok.data.data_len = MAX_NFS3_BLOCKSIZE;
						res_read.read3res_u.resok.file_attributes.post_op_attr_u.attributes =
							result.write3res.write3res_u.resok.file_wcc.after.post_op_attr_u.attributes;
						res_read.read3res_u.resok.file_attributes.present = TRUE;
						res_read.read3res_u.resok.eof = FALSE;
						res_read.read3res_u.resok.data.data_val = data_buffer;
						memset(res_read.read3res_u.resok.data.data_val, 0, MAX_NFS3_BLOCKSIZE);
						dcache_insert((void *)&rarg, rqstp->rq_vers, rqstp->rq_proc, (void *)&res_read);
					}
					warg.offset += MAX_NFS3_BLOCKSIZE;
				}
			}
		} /* end if new_size > old_size) */
		/* different count of data may be written in cache; return the requested count */
		result.write3res.write3res_u.resok.count = argp->count;
		goto out;
	} /* endif wb_cache_enabled */

no_caching:

	/* Forward request to server */
	FORWARD_REQUEST3(nfsproc3_write_3, "nfsproc_write_3");

	/* Map attributes and return to client */
	result.write3res = res;
	if (res.status != NFS3_OK) {
		/* map uid/gid even if this call failed (resfail) */
		map_return_attributes3(&result.write3res.write3res_u.resfail.after.\
				 post_op_attr_u.attributes,
				 nfsmount, rqstp);
		return (result);
	}

	if(data_cache_enabled) {
		dcache_inval(rqstp->rq_vers, rqstp->rq_proc, (void *)argp);
		if(old_size < argp->offset - argp->offset%MAX_NFS3_BLOCKSIZE
		   && old_size % MAX_NFS3_BLOCKSIZE != 0) {
			Dprintf(D_CACHE, "incomplete block at previous file end(%d)\n", old_size);
			argp->offset = old_size;
			argp->count = 1;
			dcache_inval(rqstp->rq_vers, rqstp->rq_proc, (void *)argp);
		}
	}

	if(attr_cache_enabled) {
		res_attr.status = NFS3_OK;
		res_attr.getattr3res_u.attributes = result.write3res.write3res_u.resok.file_wcc.after.post_op_attr_u.attributes;
		lkupres2 = cache_update(rqstp->rq_vers, NFSPROC3_GETATTR, (void *)&argp->file, (void *)&res_attr);
		if(lkupres2 != 1)
			cache_insert(rqstp->rq_vers, NFSPROC3_GETATTR, NULL, (void *)&res_attr);
	}

out:
	/* update invalidation list */
	if(inval_enabled == 1) {
		inval_insert(argp->file, NULL, client_id);
	}

	Dprintf(D_CACHE, "%d written, file size: %llu\n",
	        result.write3res.write3res_u.resok.count,
	        result.write3res.write3res_u.resok.file_wcc.after.post_op_attr_u.attributes.size);
	map_return_attributes3(&result.write3res.write3res_u.resok.file_wcc.after.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
	return (result);
}

union result_types
nfsd_nfsproc_create_3(create3args *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	diropres3 res;
	getattr3res res_attr;
	int lkupres, insres;
	lookup3res lookup_res;
	union result_types result;
	nfs_mount *nfsmount = NULL;
	char *fh3_buffer = (char *)malloc(NFS3_FHSIZE);
	memset(fh3_buffer, 0, NFS3_FHSIZE);
	res.diropres3_u.resok.obj.post_op_fh3_u.handle.data.data_val = fh3_buffer;

	Dprintf(D_CACHE, "*** CREATE: %s\n", argp->where.name);
	hexdump(argp->where.dir.data.data_val, argp->where.dir.data.data_len, D_CACHE);

	 /* Authenticate client */
	AUTHENTICATE_CLIENT;

	READONLY;

	/* Forward request to server */
	FORWARD_REQUEST3(nfsproc3_create_3, "nfsproc_create_3");

	/* Map attributes and return to client */
	result.diropres3 = res;
	if (res.status != NFS3_OK) {
		/* map uid/gid even if this call failed (resfail) */
		map_return_attributes3(&result.diropres3.diropres3_u.resfail.after.\
				 post_op_attr_u.attributes,
				 nfsmount, rqstp);
		return (result);
	}

	if(attr_cache_enabled) {
		lookup_res.status = result.diropres3.status;
		lookup_res.lookup3res_u.resok.object = result.diropres3.diropres3_u.resok.obj.post_op_fh3_u.handle;
		lkupres = cache_update(rqstp->rq_vers, NFSPROC3_LOOKUP, (void *)&argp->where, (void *)&lookup_res);
		if(!lkupres)
			cache_insert(rqstp->rq_vers, NFSPROC3_LOOKUP, NULL, (void *)&lookup_res);

		res_attr.status = NFS3_OK;
		res_attr.getattr3res_u.attributes = result.diropres3.diropres3_u.resok.obj_attributes.post_op_attr_u.attributes;
		cache_insert(rqstp->rq_vers, NFSPROC3_GETATTR, (void *)&result.diropres3.diropres3_u.resok.obj.post_op_fh3_u.handle, (void *)&res_attr);

		res_attr.getattr3res_u.attributes = result.diropres3.diropres3_u.resok.dir_wcc.after.post_op_attr_u.attributes;
		lkupres = cache_update(rqstp->rq_vers, NFSPROC3_GETATTR, (void *)&argp->where.dir, (void *)&res_attr);
		if(!lkupres)
			insres = cache_insert(rqstp->rq_vers, NFSPROC3_GETATTR, NULL, (void *)&res_attr);
	}

	if(inval_enabled == 1) {
		inval_insert(argp->where.dir, argp->where.name, client_id);
		inval_insert(argp->where.dir, NULL, client_id);
	}

	hexdump(result.diropres3.diropres3_u.resok.obj.post_op_fh3_u.handle.data.data_val,
			result.diropres3.diropres3_u.resok.obj.post_op_fh3_u.handle.data.data_len, D_CACHE);

	map_return_attributes3(&result.diropres3.diropres3_u.resok.dir_wcc.\
			 after.post_op_attr_u.attributes,
			 nfsmount, rqstp);
	map_return_attributes3(&result.diropres3.diropres3_u.resok.\
			 obj_attributes.post_op_attr_u.attributes,
			 nfsmount, rqstp);
	return (result);
}

union result_types
nfsd_nfsproc_mkdir_3(mkdir3args *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	diropres3 res;
	getattr3res res_attr;
	int lkupres;
	lookup3res lookup_res;
	union result_types result;
	nfs_mount *nfsmount = NULL;
	char *fh3_buffer = (char *)malloc(NFS3_FHSIZE);
	res.diropres3_u.resok.obj.post_op_fh3_u.handle.data.data_val = fh3_buffer;

	Dprintf(D_CACHE, "*** MKDIR:\n", argp->where.name);
	hexdump(argp->where.dir.data.data_val, argp->where.dir.data.data_len, D_CACHE);

	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	READONLY;

	/* Forward request to server */
	FORWARD_REQUEST3(nfsproc3_mkdir_3, "nfsproc_mkdir_3");

	/* Map attributes and return to client */
	result.diropres3 = res;
	if (res.status != NFS3_OK) {
		/* map uid/gid even if this call failed (resfail) */
		map_return_attributes3(&result.diropres3.diropres3_u.resfail.after.\
				 post_op_attr_u.attributes,
				 nfsmount, rqstp);
		return (result);
	}

	if(attr_cache_enabled) {
		lookup_res.status = result.diropres3.status;
		lookup_res.lookup3res_u.resok.object = result.diropres3.diropres3_u.resok.obj.post_op_fh3_u.handle;
		lkupres = cache_update(rqstp->rq_vers, NFSPROC3_LOOKUP, (void *)&argp->where, (void *)&lookup_res);
		if(!lkupres)
			cache_insert(rqstp->rq_vers, NFSPROC3_LOOKUP, NULL, (void *)&lookup_res);

		res_attr.status = NFS_OK;
		res_attr.getattr3res_u.attributes = result.diropres3.diropres3_u.resok.obj_attributes.post_op_attr_u.attributes;
		cache_insert(rqstp->rq_vers, NFSPROC3_GETATTR, (void *)&result.diropres3.diropres3_u.resok.obj.post_op_fh3_u.handle, (void *)&res_attr);

		res_attr.getattr3res_u.attributes = result.diropres3.diropres3_u.resok.dir_wcc.after.post_op_attr_u.attributes;
		lkupres = cache_update(rqstp->rq_vers, NFSPROC3_GETATTR, (void *)&argp->where.dir, (void *)&res_attr);
		if(!lkupres)
			cache_insert(rqstp->rq_vers, NFSPROC3_GETATTR, NULL, (void *)&res_attr);
	}

	if(inval_enabled == 1) {
		inval_insert(argp->where.dir, argp->where.name, client_id);
		inval_insert(argp->where.dir, NULL, client_id);
	}

	hexdump(result.diropres3.diropres3_u.resok.obj.post_op_fh3_u.handle.data.data_val, result.diropres3.diropres3_u.resok.obj.post_op_fh3_u.handle.data.data_len, D_CACHE);

	map_return_attributes3(&result.diropres3.diropres3_u.resok.dir_wcc.after.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
	map_return_attributes3(&result.diropres3.diropres3_u.resok.obj_attributes.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
	return (result);
}

union result_types
nfsd_nfsproc_symlink_3(symlink3args *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	diropres3 res;
	getattr3res res_attr;
	int lkupres;
	lookup3res lookup_res;
	union result_types result;
	nfs_mount *nfsmount = NULL;
	char *fh3_buffer = (char *)malloc(NFS3_FHSIZE);
	res.diropres3_u.resok.obj.post_op_fh3_u.handle.data.data_val = fh3_buffer;

	Dprintf(D_CACHE, "*** SYMLINK:\n");
	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	READONLY;

	/* Forward request to server */
	FORWARD_REQUEST3(nfsproc3_symlink_3, "nfsproc_symlink_3");

	/* Map attributes and return to client */
	result.diropres3 = res;
	if (res.status != NFS3_OK) {
		/* map uid/gid even if this call failed (resfail) */
		map_return_attributes3(&result.diropres3.diropres3_u.resfail.after.\
				 post_op_attr_u.attributes,
				 nfsmount, rqstp);
		return (result);
	}

	if(attr_cache_enabled) {
		res_attr.status = NFS_OK;
		res_attr.getattr3res_u.attributes = result.diropres3.diropres3_u.resok.dir_wcc.after.post_op_attr_u.attributes;
		lkupres = cache_update(rqstp->rq_vers, NFSPROC3_GETATTR, (void *)&argp->where.dir, (void *)&res_attr);
		if(!lkupres)
			cache_insert(rqstp->rq_vers, NFSPROC3_GETATTR, NULL, (void *)&res_attr);

		res_attr.getattr3res_u.attributes = result.diropres3.diropres3_u.resok.obj_attributes.post_op_attr_u.attributes;
		cache_insert(rqstp->rq_vers, NFSPROC3_GETATTR, (void *)&result.diropres3.diropres3_u.resok.obj.post_op_fh3_u.handle, (void *)&res_attr);

		lookup_res.status = result.diropres3.status;
		lookup_res.lookup3res_u.resok.object = result.diropres3.diropres3_u.resok.obj.post_op_fh3_u.handle;
		lkupres = cache_update(rqstp->rq_vers, NFSPROC3_LOOKUP, (void *)&argp->where, (void *)&lookup_res);
		if(!lkupres)
			cache_insert(rqstp->rq_vers, NFSPROC3_LOOKUP, NULL, (void *)&lookup_res);
	}

	if(inval_enabled == 1) {
		inval_insert(argp->where.dir, argp->where.name, client_id);
		inval_insert(argp->where.dir, NULL, client_id);
	}

	hexdump(result.diropres3.diropres3_u.resok.obj.post_op_fh3_u.handle.data.data_val, result.diropres3.diropres3_u.resok.obj.post_op_fh3_u.handle.data.data_len, D_CACHE);

	map_return_attributes3(&result.diropres3.diropres3_u.resok.dir_wcc.after.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
	map_return_attributes3(&result.diropres3.diropres3_u.resok.obj_attributes.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
	return (result);
}

union result_types
nfsd_nfsproc_mknod_3(mknod3args *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	diropres3 res;
	union result_types result;
	nfs_mount *nfsmount = NULL;
	char *fh3_buffer = (char *)malloc(NFS3_FHSIZE);
	res.diropres3_u.resok.obj.post_op_fh3_u.handle.data.data_val = fh3_buffer;

	Dprintf(D_CACHE, "*** MKNOD:\n");
	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	READONLY;

	/* Forward request to server */
	FORWARD_REQUEST3(nfsproc3_mknod_3, "nfsproc_mknod_3");

	/* Map attributes and return to client */
	result.diropres3 = res;
	if (res.status != NFS3_OK) {
		/* map uid/gid even if this call failed (resfail) */
		map_return_attributes3(&result.diropres3.diropres3_u.resfail.after.\
				 post_op_attr_u.attributes,
				 nfsmount, rqstp);
		return (result);
	}
	map_return_attributes3(&result.diropres3.diropres3_u.resok.dir_wcc.after.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
	map_return_attributes3(&result.diropres3.diropres3_u.resok.obj_attributes.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
	return (result);
}

union result_types
nfsd_nfsproc_remove_3(diropargs3 *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	wccstat3 res;
	getattr3res res_attr;
	int lkupres, lkupres2 = 0;
	lookup3res res_lookup, lookup_res;
	union result_types result;
	nfs_mount *nfsmount = NULL;
	char fh3_buffer[NFS3_FHSIZE];
	fh_node *node;
	nfs_fh3 *file = NULL;
	uint32 nlink = 0;

	Dprintf(D_CACHE, "*** REMOVE: %s\n", argp->name);
	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	READONLY;

	if(attr_cache_enabled) {
		res_lookup.lookup3res_u.resok.object.data.data_val = fh3_buffer;
		lkupres = cache_lookup(rqstp->rq_vers, NFSPROC3_LOOKUP,
				 (void *)argp, (void *)&res_lookup);
		if(!lkupres) {
			if((nfsproc3_lookup_3(argp, &res_lookup, nfsdclient)) != RPC_SUCCESS) {
				Dprintf(D_CACHE, "lookup %s error\n", argp->name);
				result.nfsstat = NFS3ERR_ACCES;
				return result;
			}
		}
		file = &res_lookup.lookup3res_u.resok.object;
		hexdump(file->data.data_val, file->data.data_len, D_CACHE);
		if(wb_cache_enabled) { /* need to know whether it is the last link */
			lkupres2 = cache_lookup(rqstp->rq_vers, NFSPROC3_GETATTR,
								   (void *)file, (void *)&res_attr);
			if(!lkupres2) { /* use attrs returned from server */
				if(!lkupres)
					res_attr.getattr3res_u.attributes =
						res_lookup.lookup3res_u.resok.obj_attributes.post_op_attr_u.attributes;
				else
					if(nfsproc3_getattr_3(file, &res_attr, nfsdclient) != RPC_SUCCESS) {
						Dprintf(L_ERROR, "getattr from server error\n");
						result.nfsstat = NFS3ERR_ACCES;
						return result;
					}
			}
			nlink = res_attr.getattr3res_u.attributes.nlink;
			Dprintf(D_CACHE, "this file has %u links\n", nlink);
		}
	}

	if(inval_enabled == 1) { /* proxy server */
		res_lookup.lookup3res_u.resok.object.data.data_val = fh3_buffer;
		if((nfsproc3_lookup_3(argp, &res_lookup, nfsdclient)) != RPC_SUCCESS) {
			Dprintf(D_CACHE, "lookup %s error\n", argp->name);
			inval_enabled = 0;
		}
	}

	/* Forward request to server */
	FORWARD_REQUEST3(nfsproc3_remove_3, "nfsproc_remove_3");

	/* Map attributes and return to client */
	if (res.status != NFS3_OK) {
		result.wccstat3 = res;
		/* map uid/gid even if this call failed (resfail) */
		map_return_attributes3(&res.wccstat3_u.wcc.after.\
				 post_op_attr_u.attributes,
				 nfsmount, rqstp);
		return (result);
	}

	if(attr_cache_enabled) {
		if(wb_cache_enabled) {
			if(nlink == 1) { /* add it into stale_fh_hash */
				node = (fh_node *)malloc(sizeof(fh_node));
				node->file.handle3.length = file->data.data_len;
				memset(node->file.handle3.handle, 0, NFS3_FHSIZE);
				memcpy(node->file.handle3.handle, file->data.data_val, file->data.data_len);
				node->status = FH_STALE;
				fh_hash_insert(stale_fh_hash, node);
				invalidate_nfs_fh_v3(file);
			} else {
				res_attr.getattr3res_u.attributes.nlink--;
				if(lkupres2) /* preserve locally modified file size */
					cache_update(rqstp->rq_vers, NFSPROC3_GETATTR,
					             (void *)file, (void *)&res_attr);
				else
					cache_insert(rqstp->rq_vers, NFSPROC3_GETATTR,
					             (void *)file, (void *)&res_attr);
			}
		} else
			invalidate_nfs_fh_v3(file);

		lookup_res.status = NFS3ERR_NOENT;
		lkupres = cache_update(rqstp->rq_vers, NFSPROC3_LOOKUP, (void *)argp, (void *)&lookup_res);

		res_attr.status = NFS_OK;
		res_attr.getattr3res_u.attributes = res.wccstat3_u.wcc.after.post_op_attr_u.attributes;
		lkupres = cache_update(rqstp->rq_vers, NFSPROC3_GETATTR, (void *)&argp->dir, (void *)&res_attr);
		if(!lkupres)
			cache_insert(rqstp->rq_vers, NFSPROC3_GETATTR, NULL, (void *)&res_attr);
	}

	/* update invalidation list */
	if(inval_enabled == 1) {
		if(res_lookup.status == NFS3_OK)
			inval_insert(res_lookup.lookup3res_u.resok.object, NULL, client_id);
		inval_insert(argp->dir, argp->name, client_id);
		inval_insert(argp->dir, NULL, client_id);
	}

	result.wccstat3 = res;
	map_return_attributes3(&result.wccstat3.wccstat3_u.wcc.after.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
	return (result);
}

union result_types
nfsd_nfsproc_rmdir_3(diropargs3 *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	wccstat3 res;
	getattr3res res_attr;
	int lkupres;
	union result_types result;
	nfs_mount *nfsmount = NULL;
	lookup3res res_lookup, lookup_res;
	char fh3_buffer[NFS3_FHSIZE];
	nfs_fh3 *file = NULL;

	Dprintf(D_CACHE, "*** RMDIR:\n");
	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	READONLY;

	if(attr_cache_enabled) {
		res_lookup.lookup3res_u.resok.object.data.data_val = fh3_buffer;
		lkupres = cache_lookup(rqstp->rq_vers, NFSPROC3_LOOKUP,
				 (void *)argp, (void *)&res_lookup);
		if(!lkupres) {
			if((nfsproc3_lookup_3(argp, &res_lookup, nfsdclient)) != RPC_SUCCESS) {
				Dprintf(D_CACHE, "lookup %s error\n", argp->name);
				result.nfsstat = NFS3ERR_ACCES;
				return result;
			}
		}
		file = &res_lookup.lookup3res_u.resok.object;
		hexdump(file->data.data_val, file->data.data_len, D_CACHE);
	}

	if(inval_enabled == 1) { /* proxy server */
		res_lookup.lookup3res_u.resok.object.data.data_val = fh3_buffer;
		if((nfsproc3_lookup_3(argp, &res_lookup, nfsdclient)) != RPC_SUCCESS) {
			Dprintf(D_CACHE, "lookup %s error\n", argp->name);
			inval_enabled = 0;
		}
	}

	/* Forward request to server */
	FORWARD_REQUEST3(nfsproc3_rmdir_3, "nfsproc_rmdir_3");

	/* Map attributes and return to client */
	result.wccstat3 = res;
	if (res.status != NFS3_OK) {
		/* map uid/gid even if this call failed (resfail) */
		map_return_attributes3(&result.wccstat3.wccstat3_u.wcc.after.\
				 post_op_attr_u.attributes,
				 nfsmount, rqstp);
		return (result);
	}

	if(attr_cache_enabled) {
		invalidate_nfs_fh_v3(file);

		lookup_res.status = NFS3ERR_NOENT;
		cache_update(rqstp->rq_vers, NFSPROC3_LOOKUP, (void *)argp, (void *)&lookup_res);

		res_attr.status = NFS_OK;
		res_attr.getattr3res_u.attributes = result.wccstat3.wccstat3_u.wcc.after.post_op_attr_u.attributes;
		lkupres = cache_update(rqstp->rq_vers, NFSPROC3_GETATTR, (void *)&argp->dir, (void *)&res_attr);
		if(!lkupres)
			cache_insert(rqstp->rq_vers, NFSPROC3_GETATTR, NULL, (void *)&res_attr);
	}

	/* update invalidation list */
	if(inval_enabled == 1) {
		if(res_lookup.status == NFS3_OK)
			inval_insert(res_lookup.lookup3res_u.resok.object, NULL, client_id);
		inval_insert(argp->dir, argp->name, client_id);
		inval_insert(argp->dir, NULL, client_id);
	}

	map_return_attributes3(&result.wccstat3.wccstat3_u.wcc.after.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
	return (result);
}

union result_types
nfsd_nfsproc_rename_3(rename3args *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	rename3res res;
	getattr3res res_attr;
	int lkupres, lkupres2 = 0;
	union result_types result;
	nfs_mount *nfsmount = NULL;
	lookup3res res_lookup;
	char fh3_buffer[NFS3_FHSIZE];
	fh_node *node;
	nfs_fh3 *file = NULL;
	uint32 nlink = 0;

	Dprintf(D_CACHE, "*** RENAME: %s -> %s\n", argp->from.name, argp->to.name);
	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	READONLY;

	if(attr_cache_enabled) {
		res_lookup.lookup3res_u.resok.object.data.data_val = fh3_buffer;
		lkupres = cache_lookup(rqstp->rq_vers, NFSPROC3_LOOKUP,
		                       (void *)&argp->to, (void *)&res_lookup);
		if(!lkupres) {
			if((nfsproc3_lookup_3(&argp->to, &res_lookup, nfsdclient)) != RPC_SUCCESS) {
				Dprintf(D_CACHE, "lookup %s error\n", argp->to.name);
				result.nfsstat = NFS3ERR_ACCES;
				return result;;
			}
		}
		if(res_lookup.status == NFS3_OK) { /* not negative cache */
			file = &res_lookup.lookup3res_u.resok.object;
			hexdump(file->data.data_val, file->data.data_len, D_CACHE);

			if(wb_cache_enabled) { /* need to know whether it is the last link */
				lkupres2 = cache_lookup(rqstp->rq_vers, NFSPROC3_GETATTR,
									   (void *)file, (void *)&res_attr);
				if(!lkupres2) { /* use attrs returned from server */
					if(!lkupres)
						res_attr.getattr3res_u.attributes =
							res_lookup.lookup3res_u.resok.obj_attributes.post_op_attr_u.attributes;
					else
						if(nfsproc3_getattr_3(file, &res_attr, nfsdclient) != RPC_SUCCESS) {
							Dprintf(L_ERROR, "getattr from server error\n");
							result.nfsstat = NFS3ERR_ACCES;
							return result;
						}
				}
				nlink = res_attr.getattr3res_u.attributes.nlink;
				Dprintf(D_CACHE, "this file has %u links\n", nlink);
			}
		}
	}

	if(inval_enabled == 1) { /* proxy server */
		res_lookup.lookup3res_u.resok.object.data.data_val = fh3_buffer;
		if((nfsproc3_lookup_3(&argp->to, &res_lookup, nfsdclient)) != RPC_SUCCESS) {
			Dprintf(D_CACHE, "lookup %s error\n", argp->to.name);
			inval_enabled = 0; // TODO
		}
	}

	/* Forward request to server */
	FORWARD_REQUEST3(nfsproc3_rename_3, "nfsproc_rename_3");

	/* Map attributes and return to client */
	result.rename3res = res;
	if (res.status != NFS3_OK) {
		/* map uid/gid even if this call failed (resfail) */
		map_return_attributes3(&result.rename3res.rename3res_u.res.\
				 fromdir_wcc.after.post_op_attr_u.attributes,
				 nfsmount, rqstp);
		map_return_attributes3(&result.rename3res.rename3res_u.res.\
				 todir_wcc.after.post_op_attr_u.attributes,
				 nfsmount, rqstp);
		return (result);
	}

	if(attr_cache_enabled) {
		if(file != NULL) {
			if(wb_cache_enabled) {
				if(nlink == 1) { /* add it into stale_fh_hash */
					node = (fh_node *)malloc(sizeof(fh_node));
					node->file.handle3.length = file->data.data_len;
					memset(node->file.handle3.handle, 0, NFS3_FHSIZE);
					memcpy(node->file.handle3.handle, file->data.data_val, file->data.data_len);
					node->status = FH_STALE;
					fh_hash_insert(stale_fh_hash, node);
					invalidate_nfs_fh_v3(file);
				} else {
					res_attr.getattr3res_u.attributes.nlink--;
					if(lkupres2) /* preserve locally modified file size */
						cache_update(rqstp->rq_vers, NFSPROC3_GETATTR,
									 (void *)file, (void *)&res_attr);
					else
						cache_insert(rqstp->rq_vers, NFSPROC3_GETATTR,
									 (void *)file, (void *)&res_attr);
				}
			} else
				invalidate_nfs_fh_v3(file);
		}

		invalidate_diroparg_v3(&argp->to);

		res_lookup.status = NFS3ERR_NOENT;
		cache_update(rqstp->rq_vers, NFSPROC3_LOOKUP, (void *)argp, (void *)&res_lookup);

		res_attr.status = NFS3_OK;
		res_attr.getattr3res_u.attributes =
			result.rename3res.rename3res_u.res.fromdir_wcc.after.post_op_attr_u.attributes;
		lkupres = cache_update(rqstp->rq_vers, NFSPROC3_GETATTR,
		                       (void *)&argp->from.dir, (void *)&res_attr);
		if(!lkupres)
			cache_insert(rqstp->rq_vers, NFSPROC3_GETATTR, NULL,
						(void *)&res_attr);

		if(argp->from.dir.data.data_len != argp->to.dir.data.data_len ||
		   memcmp(argp->from.dir.data.data_val, argp->to.dir.data.data_val,
		   argp->from.dir.data.data_len) != 0) {
			res_attr.status = NFS_OK;
			res_attr.getattr3res_u.attributes =
				result.rename3res.rename3res_u.res.todir_wcc.after.post_op_attr_u.attributes;
			lkupres = cache_update(rqstp->rq_vers, NFSPROC3_GETATTR,
						(void *)&argp->to.dir, (void *)&res_attr);
			if(!lkupres)
				cache_insert(rqstp->rq_vers, NFSPROC3_GETATTR, NULL, (void *)&res_attr);
		}
	}

	/* update invalidation list */
	if(inval_enabled == 1 && res_lookup.status == NFS3_OK) {
		inval_insert(res_lookup.lookup3res_u.resok.object, NULL, client_id);
		inval_insert(argp->from.dir, argp->from.name, client_id);
		inval_insert(argp->to.dir, argp->to.name, client_id);
		inval_insert(argp->to.dir, NULL, client_id);
		if(argp->to.dir.data.data_len != argp->from.dir.data.data_len ||
		   memcmp(argp->to.dir.data.data_val, argp->from.dir.data.data_val,
		   argp->to.dir.data.data_len) != 0)
			inval_insert(argp->from.dir, NULL, client_id);
	}

	map_return_attributes3(&result.rename3res.rename3res_u.res.\
			 fromdir_wcc.after.post_op_attr_u.attributes,
			 nfsmount, rqstp);
	map_return_attributes3(&result.rename3res.rename3res_u.res.\
			 todir_wcc.after.post_op_attr_u.attributes,
			 nfsmount, rqstp);
	return (result);
}

union result_types
nfsd_nfsproc_link_3(link3args *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	link3res res;
	getattr3res res_attr;
	int lkupres, lkupres2 = 0;
	lookup3res lookup_res;
	union result_types result;
	nfs_mount *nfsmount = NULL;

	Dprintf(D_CACHE, "*** LINK: %s\n", argp->link.name);
	hexdump(argp->file.data.data_val, argp->file.data.data_len, D_CACHE);
	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	READONLY;

	if(wb_cache_enabled)
		lkupres2 = cache_lookup(rqstp->rq_vers, NFSPROC3_GETATTR,
		                        (void *)&argp->file, (void *)&res_attr);

	/* Forward request to server */
	FORWARD_REQUEST3(nfsproc3_link_3, "nfsproc_link_3");

	/* Map attributes and return to client */
	result.link3res = res;
	if (res.status != NFS3_OK) {
		/* map uid/gid even if this call failed (resfail) */
		map_return_attributes3(&result.link3res.link3res_u.res.linkdir_wcc.after.\
				 post_op_attr_u.attributes,
				 nfsmount, rqstp);
		map_return_attributes3(&result.link3res.link3res_u.res.file_attributes.\
				 post_op_attr_u.attributes,
				 nfsmount, rqstp);
		return (result);
	}

	if(attr_cache_enabled) {
		lookup_res.status = NFS3_OK;
		lookup_res.lookup3res_u.resok.object = argp->file;
		lkupres = cache_update(rqstp->rq_vers, NFSPROC3_LOOKUP,
		                       (void *)&argp->link, (void *)&lookup_res);
		if(!lkupres)
			cache_insert(rqstp->rq_vers, NFSPROC3_LOOKUP, NULL, (void *)&lookup_res);

		res_attr.status = NFS3_OK;
		if(wb_cache_enabled && lkupres2) /* keep locally modified file size */
			res.link3res_u.res.file_attributes.post_op_attr_u.attributes.size =
				res_attr.getattr3res_u.attributes.size;
		res_attr.getattr3res_u.attributes =
			res.link3res_u.res.file_attributes.post_op_attr_u.attributes;
		lkupres = cache_update(rqstp->rq_vers, NFSPROC3_GETATTR,
		                       (void *)&argp->file, (void *)&res_attr);
		if(!lkupres)
			cache_insert(rqstp->rq_vers, NFSPROC3_GETATTR, NULL, (void *)&res_attr);

		res_attr.status = NFS_OK;
		res_attr.getattr3res_u.attributes =
			res.link3res_u.res.linkdir_wcc.after.post_op_attr_u.attributes;
		lkupres = cache_update(rqstp->rq_vers, NFSPROC3_GETATTR,
		                       (void *)&argp->link.dir, (void *)&res_attr);
		if(!lkupres)
			cache_insert(rqstp->rq_vers, NFSPROC3_GETATTR, NULL, (void *)&res_attr);
	}

	/* update invalidation list */
	if(inval_enabled == 1) {
		inval_insert(argp->file, NULL, client_id);
		inval_insert(argp->link.dir, argp->link.name, client_id);
		inval_insert(argp->link.dir, NULL, client_id);
	}

	map_return_attributes3(&result.link3res.link3res_u.res.linkdir_wcc.after.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
	map_return_attributes3(&result.link3res.link3res_u.res.file_attributes.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
	return (result);
}

union result_types
nfsd_nfsproc_readdir_3(readdir3args *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	readdir3res res;
	union result_types result;
	nfs_mount *nfsmount = NULL;
	char *buffer = (char *)malloc(argp->count);
	memset(buffer, 0, argp->count);
	res.readdir3res_u.resok.reply.entries = (entry3 *)buffer;

	Dprintf(D_CACHE, "*** READDIR:\n");
	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	/* Forward request to server */
	FORWARD_REQUEST3(nfsproc3_readdir_3, "nfsproc_readdir_3");

	/* Map attributes and return to client */
	result.readdir3res = res;
	if (res.status != NFS3_OK) {
		/* map uid/gid even if this call failed (resfail) */
		map_return_attributes3(&result.readdir3res.readdir3res_u.resfail.\
				 post_op_attr_u.attributes,
				 nfsmount, rqstp);
		return (result);
	}
	map_return_attributes3(&result.readdir3res.readdir3res_u.resok.\
			 dir_attributes.post_op_attr_u.attributes,
			 nfsmount, rqstp);
	return (result);
}

union result_types
nfsd_nfsproc_readdirplus_3(readdirplus3args *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	readdirplus3res res;
	struct entryplus3 *scan;
	union result_types result;
	diropargs3 dir_arg;
	lookup3res lookup_res;
	int lkupres;
	getattr3res res_attr;
	nfs_mount *nfsmount = NULL;
	char *buffer = (char *)malloc(argp->maxcount);
	memset(buffer, 0, argp->maxcount);
	res.readdirplus3res_u.resok.reply.entries = (entryplus3 *)buffer;

	Dprintf(D_CACHE, "*** READDIRPLUS (%d %d):\n", argp->dircount, argp->maxcount);
	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	/* Forward request to server */
	FORWARD_REQUEST3(nfsproc3_readdirplus_3, "nfsproc_readdirplus_3");

	/* Map attributes and return to client */
	result.readdirplus3res = res;
	if (res.status != NFS3_OK) {
		/* map uid/gid even if this call failed (resfail) */
		map_return_attributes3(&result.readdirplus3res.readdirplus3res_u.resfail.\
				 post_op_attr_u.attributes,
				 nfsmount, rqstp);
		return (result);
	}

	/* need to map all returned attributes */
	for(scan = res.readdirplus3res_u.resok.reply.entries; scan != NULL; scan = scan->nextentry) {
		Dprintf(D_CACHE, "%s\n", scan->name);
		if(attr_cache_enabled) {
			if(scan->name_handle.present && strcmp(scan->name, ".") != 0 &&
			   strcmp(scan->name, "..") != 0) {
				hexdump(scan->name_handle.post_op_fh3_u.handle.data.data_val,
				        scan->name_handle.post_op_fh3_u.handle.data.data_len, D_CACHE);

				dir_arg.dir = argp->dir;
				dir_arg.name = scan->name;
				lookup_res.status = NFS3_OK;
				lookup_res.lookup3res_u.resok.object = scan->name_handle.post_op_fh3_u.handle;
				lkupres = cache_update(rqstp->rq_vers, NFSPROC3_LOOKUP, (void *)&dir_arg, (void *)&lookup_res);
				if(!lkupres)
					cache_insert(rqstp->rq_vers, NFSPROC3_LOOKUP, NULL, (void *)&lookup_res);

				if(scan->name_attributes.present) {
					if(wb_cache_enabled) {
						lkupres = cache_lookup(rqstp->rq_vers, NFSPROC3_GETATTR,
						                       (void *)&scan->name_handle.post_op_fh3_u.handle,
						                       (void *)&res_attr);
						if(lkupres) /* return locally modified file size */
							scan->name_attributes.post_op_attr_u.attributes.size =
								res_attr.getattr3res_u.attributes.size;
						else {
							res_attr.status = NFS3_OK;
							res_attr.getattr3res_u.attributes = scan->name_attributes.post_op_attr_u.attributes;
							cache_insert(rqstp->rq_vers, NFSPROC3_GETATTR, NULL, (void *)&res_attr);
						}
					} else {
						res_attr.status = NFS3_OK;
						res_attr.getattr3res_u.attributes = scan->name_attributes.post_op_attr_u.attributes;
						lkupres = cache_update(rqstp->rq_vers, NFSPROC3_GETATTR, (void *)&scan->name_handle.post_op_fh3_u.handle, (void *)&res_attr);
						if(!lkupres)
							cache_insert(rqstp->rq_vers, NFSPROC3_GETATTR, NULL, (void *)&res_attr);

					}

					Dprintf(D_CACHE, "file size %llu, mtime %d, link %d\n",
					        res_attr.getattr3res_u.attributes.size,
					        res_attr.getattr3res_u.attributes.mtime.seconds,
					        res_attr.getattr3res_u.attributes.nlink);
				}
			}
		} // end of if(attr_cache_enabled)
		map_return_attributes3(&scan->name_attributes.post_op_attr_u.attributes,
				 nfsmount, rqstp);
	} // end for

	map_return_attributes3(&result.readdirplus3res.readdirplus3res_u.resok.\
			 dir_attributes.post_op_attr_u.attributes,
			 nfsmount, rqstp);

	return (result);
}

union result_types
nfsd_nfsproc_fsstat_3(nfs_fh3 *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	fsstat3res res;
	union result_types result;
	nfs_mount *nfsmount = NULL;

	Dprintf(D_CACHE, "*** FSSTAT:\n");
	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	/* Forward request to server */
	FORWARD_REQUEST3(nfsproc3_fsstat_3, "nfsproc_fstat_3");

	/* Map attributes and return to client */
	result.fsstat3res = res;
	if (res.status != NFS3_OK) {
		/* map uid/gid even if this call failed (resfail) */
		map_return_attributes3(&result.fsstat3res.fsstat3res_u.resfail.\
				 post_op_attr_u.attributes,
				 nfsmount, rqstp);
		return (result);
	}
	map_return_attributes3(&result.fsstat3res.fsstat3res_u.resok.obj_attributes.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
	return (result);
}

union result_types
nfsd_nfsproc_fsinfo_3(nfs_fh3 *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	fsinfo3res res;
	union result_types result;
	nfs_mount *nfsmount = NULL;

	Dprintf(D_CACHE, "*** FSINFO:\n");
	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	/* Forward request to server */
	FORWARD_REQUEST3(nfsproc3_fsinfo_3, "nfsproc_fsinfo_3");

	/* Map attributes and return to client */
	result.fsinfo3res = res;
	if (res.status != NFS3_OK) {
		/* map uid/gid even if this call failed (resfail) */
		map_return_attributes3(&result.fsinfo3res.fsinfo3res_u.resfail.\
				 post_op_attr_u.attributes,
				 nfsmount, rqstp);
		return (result);
	}
	map_return_attributes3(&result.fsinfo3res.fsinfo3res_u.resok.obj_attributes.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
	Dprintf(D_CALL, "rtmax: %u\n", result.fsinfo3res.fsinfo3res_u.resok.rtmax);

	return (result);
}

union result_types
nfsd_nfsproc_pathconf_3(nfs_fh3 *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	pathconf3res res;
	union result_types result;
	nfs_mount *nfsmount = NULL;

	Dprintf(D_CACHE, "*** PATHCONF:\n");
	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	/* Forward request to server */
	FORWARD_REQUEST3(nfsproc3_pathconf_3, "nfsproc_pathconf_3");

	/* Map attributes and return to client */
	result.pathconf3res = res;
	if (res.status != NFS3_OK) {
		/* map uid/gid even if this call failed (resfail) */
		map_return_attributes3(&result.pathconf3res.pathconf3res_u.resfail.\
				 post_op_attr_u.attributes,
				 nfsmount, rqstp);
		return (result);
	}
	map_return_attributes3(&result.pathconf3res.pathconf3res_u.resok.\
			 obj_attributes.post_op_attr_u.attributes,
			 nfsmount, rqstp);
	return (result);
}

union result_types
nfsd_nfsproc_commit_3(commit3args *argp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	nfsstat astat;
	commit3res res;
	union result_types result;
	nfs_mount *nfsmount = NULL;

	Dprintf(D_CACHE, "*** COMMIT:\n");
	/* Authenticate client */
	AUTHENTICATE_CLIENT;

	READONLY;

	/* Forward request to server */
	FORWARD_REQUEST3(nfsproc3_commit_3, "nfsproc_commit_3");

	/* Map attributes and return to client */
	result.commit3res = res;
	if (res.status != NFS3_OK) {
		/* map uid/gid even if this call failed (resfail) */
		map_return_attributes3(&result.commit3res.commit3res_u.resfail.\
				 after.post_op_attr_u.attributes,
				 nfsmount, rqstp);
		return (result);
	}
	map_return_attributes3(&result.commit3res.commit3res_u.resok.file_wcc.after.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
	return (result);
}

union result_types
nfsd_nfsproc_getinv_3(u_int *tstamp, struct svc_req *rqstp, CLIENT *nfsdclient)
{
	u_int timestamp;
	int i, j = 0, k;
	nfs_client *nfsclient = NULL;
	union result_types result;
	char *cache_buffer = malloc(MAX_NFS3_BLOCKSIZE);
	result.read3res.read3res_u.resok.data.data_val = cache_buffer;

	Dprintf(D_CACHE, "*** GETINV\n");
	/* Authenticate client */
	if ((nfsclient = auth_clnt(rqstp)) == NULL) {
		result.read3res.status = NFS3ERR_ACCES;
		return result;
	}
	if((client_id = auth_sk(rqstp)) < 1) {
		result.read3res.status = NFS3ERR_ACCES;
		return result;
	}

	if(inval_enabled != 1) {
		Dprintf(D_CACHE, "server invalidation is not enabled\n");
		result.read3res.status = 1;
		return result; /* server invalidation is not enabled */
	}

	timestamp = *tstamp;
	Dprintf(D_CACHE, "client: %d, timestamp: %d\n", client_id, timestamp);

	if(timestamp < 0) {
		Dprintf(D_CACHE, "invalid timestamp\n");
		result.read3res.status = 2;
		return result; /* invalidate timestamp */
	}

	if(timestamp == 0) {
		result.read3res.read3res_u.resok.file_attributes.present = TRUE;
		result.read3res.read3res_u.resok.file_attributes.post_op_attr_u.attributes.mode = 0;
		result.read3res.read3res_u.resok.data.data_len = 0;
		result.read3res.read3res_u.resok.file_attributes.post_op_attr_u.attributes.size = inval_index+inval_base+1;
		result.read3res.status = 0;
		client_tstamps[client_id-1] = inval_index+inval_base+1;
		return result; //client initialization
	}

	if(timestamp-1 > inval_index+inval_base) {
		Dprintf(D_CACHE, "client reinitialization is necessary\n");
		result.read3res.status = 3;
		return result; /* client need reinitialize */
	}

	result.read3res.read3res_u.resok.file_attributes.present = TRUE;
	result.read3res.read3res_u.resok.file_attributes.post_op_attr_u.attributes.mode = 0;

	for(i=timestamp-1; i<inval_index+inval_base; i++) {
		if(inval_list[i-inval_base].client_id != client_id) {
			Dprintf(D_CACHE, "who did this? %d\n", inval_list[i-inval_base].client_id);
			if(inval_list[i-inval_base].flag == 1) {
				k = inval_list[i-inval_base].info.attr.file.data.data_len+sizeof(u_int);
							if(j+k+1 > MAX_NFS3_BLOCKSIZE && i < inval_index+inval_base-1) {
								Dprintf(D_CACHE, "another getinv request is necessary\n");
					result.read3res.read3res_u.resok.file_attributes.post_op_attr_u.attributes.mode = 1;
								i++;
					break;
				}
				result.read3res.read3res_u.resok.data.data_val[j++] = 'g';
				Dprintf(D_CACHE, "%d %c\n", j, result.read3res.read3res_u.resok.data.data_val[j-1]);
				memcpy(&result.read3res.read3res_u.resok.data.data_val[j],
							&inval_list[i-inval_base].info.attr.file.data.data_len,
											sizeof(u_int));
				j += sizeof(u_int);
				Dprintf(D_CACHE, "%d %d\n", j, result.read3res.read3res_u.resok.data.data_val[j-sizeof(u_int)]);
				if(memcpy(&result.read3res.read3res_u.resok.data.data_val[j],
					inval_list[i-inval_base].info.attr.file.data.data_val,
									inval_list[i-inval_base].info.attr.file.data.data_len) == NULL) {
					Dprintf(D_CACHE, "memory copy error\n");
					result.read3res.status = 4;
					return result;
				}
				j += inval_list[i-inval_base].info.attr.file.data.data_len;
				hexdump(&result.read3res.read3res_u.resok.data.data_val[j-inval_list[i-inval_base].info.attr.file.data.data_len], inval_list[i-inval_base].info.attr.file.data.data_len, D_CACHE);
				Dprintf(D_CACHE, "inval size: %d\n", j);
			}
			else {
				k = inval_list[i-inval_base].info.diro.dir.data.data_len+sizeof(u_int)+strlen(inval_list[i-inval_base].info.diro.name)+1;
				if(j+k+1 > MAX_NFS3_BLOCKSIZE && i<inval_index+inval_base-1) {
								Dprintf(D_CACHE, "another getinv request is necessary\n");
					result.read3res.read3res_u.resok.file_attributes.post_op_attr_u.attributes.mode = 1;
								i++;
					break;
				}
				result.read3res.read3res_u.resok.data.data_val[j++] = 'd';
				Dprintf(D_CACHE, "%d %c\n", j, result.read3res.read3res_u.resok.data.data_val[j-1]);
				memcpy(&result.read3res.read3res_u.resok.data.data_val[j],
										&inval_list[i-inval_base].info.diro.dir.data.data_len,
										sizeof(u_int));
				j += sizeof(u_int);
				Dprintf(D_CACHE, "%d %d\n", j, result.read3res.read3res_u.resok.data.data_val[j-sizeof(u_int)]);
				if(memcpy(&result.read3res.read3res_u.resok.data.data_val[j],
								inval_list[i-inval_base].info.diro.dir.data.data_val,
												inval_list[i-inval_base].info.diro.dir.data.data_len) == NULL) {
								Dprintf(D_CACHE, "memeory copy error\n");
								result.read3res.status = 4;
								return result;
				}
				j += inval_list[i-inval_base].info.diro.dir.data.data_len;
				hexdump(&result.read3res.read3res_u.resok.data.data_val[j-inval_list[i-inval_base].info.attr.file.data.data_len], inval_list[i-inval_base].info.attr.file.data.data_len, D_CACHE);
				strcpy(&result.read3res.read3res_u.resok.data.data_val[j],
									 inval_list[i-inval_base].info.diro.name);
				Dprintf(D_CACHE, "%s\n", &result.read3res.read3res_u.resok.data.data_val[j]);
				j += strlen(inval_list[i-inval_base].info.diro.name) + 1;
				Dprintf(D_CACHE, "inval size: %d\n", j);
			}
		}
	}
	result.read3res.read3res_u.resok.data.data_len = j;
	result.read3res.read3res_u.resok.file_attributes.post_op_attr_u.attributes.size = i+1;
	client_tstamps[client_id-1] = i+1;
	Dprintf(D_CACHE, "client_id: %d, timestamp: %d->%d, inval_index+inval_base: %d+%d, data_len: %d",
		client_id, timestamp, (int)result.read3res.read3res_u.resok.file_attributes.post_op_attr_u.attributes.size,
		inval_index, inval_base, result.read3res.read3res_u.resok.data.data_len);
	result.read3res.status = 0;
	return result;
}


/****************************************************************************
 * MAIN function
 *  read command line arguments; initialize signal handlers;
 *  setup NFS client and server
 ****************************************************************************/
int main(int argc, char **argv)
{
	char *auth_file = NULL;
	char *cachebasepath = NULL;
	char *cookiefile = NULL;
	char *cachesessionid = NULL;
	int attrcachesize = DEFAULT_ATTR_CACHE_SIZE;
	int attrcacheassoc = DEFAULT_ATTR_CACHE_ASSOC;
	int attrcachebanks = DEFAULT_ATTR_CACHE_BANKS;
	int datacachesize = DEFAULT_DATA_CACHE_SIZE;
	int datacacheassoc = DEFAULT_DATA_CACHE_ASSOC;
	int datacachebanks = DEFAULT_DATA_CACHE_BANKS;
	int	foreground = 0;
	int c;

	program_name = argv[0];
	localhostname[0] = '\x0';
	chdir("/");

	/* Parse the command line options and arguments. */
	opterr = 0;
	while ((c = getopt_long(argc, argv, shortopts, longopts, NULL)) != EOF)
		switch (c) {
		case 'h':
			usage(stdout, 0);
			break;
		case 'd':
			enable_logging(optarg);
			break;
		case 'F':
			foreground = 1;
			break;
		case 'f':
			auth_file = optarg;
			break;
		case 'i':
			cachesessionid = optarg;
			break;
		case 'C':
#ifndef COOKIE_AUTH
			fprintf(stderr, "Cookie authentication has not been"
				" enabled in this executable.\n");
			exit(-1);
#endif /* COOKIE_AUTH */
			cookiefile = optarg;
			break;
		case 'V':
			nfsvers = atoi(optarg);
			if(nfsvers<2 || nfsvers>3) {
				fprintf(stderr, "NFS version %i not supported\n",nfsvers);
				exit(-1);
			}
			break;
		case 'A':
			attr_cache_enabled = 1;
			break;
		case 'D':
			data_cache_enabled = 1;
			break;
		case 'W':
			wb_cache_enabled = 1;
			break;
		case 'c':
			cachebasepath = optarg;
			break;
		case 'z':
			attrcachesize = atoi(optarg);
			break;
		case 'a':
			attrcacheassoc = atoi(optarg);
			break;
		case 'b':
			attrcachebanks = atoi(optarg);
			break;
		case 'e':
			datacachesize = atoi(optarg);
			break;
		case 'o':
			datacacheassoc = atoi(optarg);
			break;
		case 'k':
			datacachebanks = atoi(optarg);
			break;
		case 'H':
			strncpy(&localhostname[0], optarg, HOSTNAMESIZE);
			break;
		case 'P':
			nfsportin = atoi(optarg);
			if (nfsportin < 0) {
				fprintf(stderr, "nfsd: bad port number: %s\n",
					optarg);
				usage(stderr, -1);
			}
			break;
		case 'O':
			nfsportout = atoi(optarg);
			if (nfsportout <= 0) {
				fprintf(stderr, "nfsd: bad port number: %s\n",
					optarg);
				usage(stderr, -1);
			}
			break;
		case 'S':
			rootsock = atoi(optarg);
			if (rootsock <= 0) {
				fprintf(stderr, "nfsd: bad socket number: %s\n",
					optarg);
				usage(stderr, -1);
			}
			break;
		case 't':
			if(strcmp(optarg,"tcp")==0) {
				transport = NFS_OVER_TCP;
			} else {
				if(strcmp(optarg,"udp")==0) {
					transport = NFS_OVER_UDP;
				} else {
					fprintf(stderr, "nfsd: bad transport: %s\n",
						optarg);
					usage(stderr, -1);
				}
			}
			break;
		case 'T':
			sync_interval = atoi(optarg);
			if (sync_interval <= 0) {
				fprintf(stderr, "nfsd: bad sync interval value: %s\n",
					optarg);
				usage(stderr, -1);
			}
			break;
		case 'I':
			if(strcmp(optarg, "server") == 0) inval_enabled = SVCROLE;
			else if(strcmp(optarg, "client") == 0) inval_enabled = CLNTROLE;
			else if(strcmp(optarg, "alive") == 0) inval_enabled = 3;
			else {
				fprintf(stderr, "nfsd: bad invalidation request: %s\n", optarg);
				usage(stderr, -1);
			}
			break;
		case 'x':
			inv_min_interval = atoi(optarg);
			if (inv_min_interval <= 0) {
				fprintf(stderr, "nfsd: bad minimum invalidation interval value: %s\n",
					optarg);
				usage(stderr, -1);
			}
			break;
		case 'y':
			inv_max_interval = atoi(optarg);
			if (inv_max_interval <= 0) {
				fprintf(stderr, "nfsd: bad maximum invalidation interval value: %s\n",
					optarg);
				usage(stderr, -1);
			}
			break;
		case 's':
			if((send_sk_path = optarg) == NULL) {
				fprintf(stderr,"Must specify path to the session key\n");
				usage(stderr, -1);
			}
			send_sk_enabled = 1;
			break;
		case 'm':
			if((match_sk_path = optarg) == NULL) {
				fprintf(stderr,"Must specify path to the session key\n");
				usage(stderr, -1);
			}
			match_sk_enabled = 1;
			break;
		case 'p':
			promiscuous = 1;
			break;
		case 'v':
			printf("%s\n", version);
			exit(0);
		case 0:
			break;
		case '?':
		default:
			usage(stderr, -1);
		} /* endof switch & while*/

	/* No more arguments allowed. */
	if (optind != argc)
		usage(stderr, -1);

	if(auth_file == NULL) {
		fprintf(stderr,"Must specify exports file\n");
		exit(-1);
	}

	/* needed for auth_clnt.c */
	authvers = nfsvers;

	/* Initialize logging. */
	log_open("nfsd", foreground);

	/* Initialize RPC stuff */
	rpc_init("nfsd", NFS_PROGRAM, nfsvers, nfs_dispatch,
		     nfsportin, NFS_MAXDATA);

	/* Obtain local host name for default forwarding address if
		 not given as command-line argument (-H) */
	if(localhostname[0] == '\x0') {
		if(gethostname(localhostname,HOSTNAMESIZE)!=0) {
			fprintf(stderr,"gethostname(): couldn't obtain local host name\n");
			exit(-1);
		}
	}

	/* initialize v2 client */
	if(nfsvers == 2) {
		if((nfsdclient2=open_nfsdclient(localhostname,nfsportout,2,rootsock,
						transport))==NULL) {
			fprintf(stderr,"Couldn't create v2 client to %s:%i; bailing out\n",
				localhostname,nfsportout);
			exit(-1);
		}
		nfsdclient = nfsdclient2;
	}
	/* initialize v3 client */
	if(nfsvers == 3) {
		if((nfsdclient3=open_nfsdclient(localhostname,nfsportout,3,rootsock,
			            transport))==NULL) {
			fprintf(stderr,"Couldn't create v3 client to %s:%i; bailing out\n",
				localhostname,nfsportout);
			exit(-1);
		}
		nfsdclient = nfsdclient3;
	}

	if((data_cache_enabled || attr_cache_enabled) && cachesessionid==NULL) {
		fprintf(stderr,"Must specify cache session ID string\n");
		exit(-1);
	}
	if(wb_cache_enabled && (!data_cache_enabled || !attr_cache_enabled)) {
		fprintf(stderr,"Must enable data cache and attributes cache\n");
		wb_cache_enabled = 0;
	}

	/* Initialize attribute cache */
	if(attr_cache_enabled) {
		if((init_attr_cache(attrcachesize, attrcacheassoc, attrcachebanks,
			                cachebasepath,cachesessionid))==-1) {
			Dprintf(L_ERROR, "nfsd: cannot initialize attribute cache\n");
			exit(-1);
		} else {
			umask(2); /* set mask for sharing cache banks among group */
		}
	}

	/* Initialize data cache */
	if(data_cache_enabled) {
		if((init_data_cache(datacachesize, datacacheassoc, datacachebanks,
			                cachebasepath,cachesessionid))==-1) {
			fprintf(stderr, "nfsd: cannot initialize data cache\n");
			exit(-1);
		} else {
			umask(2); /* set mask for sharing cache banks among group */
		}
	}

	/* If not in foreground, fork off a child. */
	if (!foreground) {
		if ((c = fork()) > 0)
			/* parent exits */
			exit(0);
		if (c < 0) {
			Dprintf(L_FATAL, "nfsd: cannot fork: %s\n",
				strerror(errno));
		}
	}

	/* Initialize the AUTH module. */
	auth_init(auth_file);

	if(send_sk_enabled || match_sk_enabled)
		if(sk_init(send_sk_path, match_sk_path) != 0) {
			Dprintf(L_ERROR, "Session key initialization error\n");
			exit(-1);
		}

	if(inval_enabled > 0) {
		if(inval_enabled == 2) {
			if(!attr_cache_enabled || !data_cache_enabled) {
				Dprintf(L_ERROR, "Must enable attribute cache and data cache\n");
				exit(-1);
			}
			if(!send_sk_enabled) {
				Dprintf(L_ERROR, "Must use session key authentication\n");
				exit(-1);
			}
		} else if(inval_enabled == 1) {
			if(!match_sk_enabled) {
				Dprintf(L_ERROR, "Must use session key authentication\n");
				exit(-1);
			}
			if(inval_init() != 0) {
				Dprintf(L_ERROR, "Invalidation initialization error\n");
				exit(-1);
			}
		}
	}

	if (!foreground) {
		/* No more logging to stderr */
		background_logging();

		/* Now we remove ourselves from the foreground. */
		close(0);
		close(1);
		close(2);
#ifdef HAVE_SETSID
		setsid();
#else
		{
			int fd;
			if ((fd = open("/dev/tty", O_RDWR)) >= 0) {
				ioctl(fd, TIOCNOTTY, (char *) NULL);
				close(fd);
			}
		}
#endif
	}

	/* Install signal handlers:
	   USR1 toggles logging on/off
	   USR2 dumps statistics to /tmp/gvfs.nfsd_profile.PID
	   TERM also dumps statistics, but then exits
	   QUIT flushes caches and dumps statistics
	   HUP re-reads input files (exports, map)
	 */
	install_signal_handler(SIGUSR1, toggle_logging);
	install_signal_handler(SIGQUIT, flush_caches);
#ifdef CALL_PROFILING
	install_signal_handler(SIGUSR2, sigusr2);
	gettimeofday(&inittime, NULL);
#endif
	install_signal_handler(SIGHUP, reinitialize);
	install_signal_handler(SIGTERM, sigterm);
	/* atexit registers "terminate" to be called at normal
	   program termination */
	atexit(terminate);

	if(inval_enabled == 2) {
		if(getinv() == 0) {
			install_signal_handler(SIGALRM, setalarm);
		}
	}

	/* Run the NFS server. */
	fprintf(stdout,"Proxy started successfully.\n");
	fflush(stdout);

	svc_run();

	Dprintf(L_ERROR, "Oh no Mr. Bill... nfs_server() returned!\n");
	exit(-1);
}


/****************************************************************************
 * Shows program usage
 ****************************************************************************/
static void
usage(FILE *fp, int n)
{
	fprintf(fp,
		"\nUsage: %s -f exports-file\n"
		"       [-d kind] [-H servername] [-S socketd] [-t tcp|udp]\n"
		"       [-C cookiefile] [-P inputport] [-O outputport]\n"
		"       [-V nfsvers] [-F] [-h] [-p] [-v] [-A] [-D] [-W]\n"
		"       [-T syncinterval] [-i cachesessionid] [-c cachebasepath]\n"
		"       [-z attrcachesize] [-a attrcacheassoc] [-b attrcachebanks] \n"
		"       [-e datacachesize] [-o datacacheassoc] [-k datacachebanks] \n"
		"       [-s sendskpath] [-m matchskpath] \n"
		, program_name);
	fprintf(fp,
		"\nOptions: \n"
		"       -d: enables debugging mode (auth,all,call,devtab,cache,general,rmtab,stale,ugid)\n"
		"       -F: puts process to foreground\n"
		"       -h: prints help\n"
		"       -H: specifies server name (local hostname is used if not specified)\n"
		"       -S: specifies socket\n"
		"       -t: specifies transport protocol (tcp <default> or udp)\n"
		"       -V: specifies NFS version (2 <default> or 3)\n"
		"       -C: specifies cookie file\n"
		"       -A: enables attributes cache (only works with NFS version 2)\n"
		"       -D: enables data cache (only works with NFS version 2)\n"
		"       -W: enables write-back caching\n"
		"       -T: specifies periodical sync interval (seconds) of caches with server\n"
		"       -i: specifies cache session id (required if cache is enabled)\n"
		"       -c: specifies cache base path (required if cache is enabled)\n"
		"       -z: specifies attributes cache size (default is 4096 blocks)\n"
		"       -a: specifies attributes cache associativity (default is 8)\n"
		"       -b: specifies attributes cache bank number (default is 64)\n"
		"       -e: specifies data cache size (default is 65536 blocks)\n"
		"       -o: specifies data cache associativity (default is 16)\n"
		"       -k: specifies data cache bank number (default is 128)\n"
		"       -p: enables promiscuous mode\n"
		"       -v: prints version number\n"
		"       -f: specifies path of exports file (required)\n"
		"       -P: specifies input port number (a random port is chosen and printed out if not specified)\n"
		"       -O: specifies output port number (a random port is chosen and printed out if not specified)\n"
		"       -s: specifies path of the session key to be sent\n"
		"       -m: specifies path of the session key to match with\n");
	fprintf(fp,
		"\nSignals:\n"
		"       -USR1: toggles logging on/off\n"
		"       -USR2: dumps statistics to /tmp/gvfs.nfsd profile.PID\n"
		"       -HUP:  re-reads input files (exports and map)\n"
		"       -ALRM: syncs caches with server\n"
		"       -QUIT: flushes caches and dumps statistics to /tmp/gvfs.nfsd-profile.PID\n"
		"       -TERM: exits and dumps statistics to /tmp/gvfs.nfsd-profile.PID\n");

	exit(n);
}


/****************************************************************************
 * Signal handler for SIGTERM
 ****************************************************************************/
static RETSIGTYPE
sigterm(int sig)
{
	/* Dump statistics if enabled */
	if(wb_cache_enabled) flush_dirty_blocks();
#ifdef CALL_PROFILING
	if(inval_enabled && inval_list != NULL) free(inval_list);
	dump_stats(SIGTERM);
#endif /* CALL_PROFILING */
	exit(0);
}

/****************************************************************************
 * Signal handler for SIGUSR2
 ****************************************************************************/
static RETSIGTYPE
sigusr2(int sig)
{
	/* Dump statistics if enabled */
#ifdef CALL_PROFILING
	dump_stats(SIGUSR2);
#endif /* CALL_PROFILING */
}

/****************************************************************************
 * Exit subroutine called upon normal program termination
 ****************************************************************************/
static void
terminate(void)
{
	/* finish server */
	rpc_exit(NFS_PROGRAM, nfsvers);
	/* finish client(s) */
	 close_nfsdclient(nfsdclient2);
	 close_nfsdclient(nfsdclient3);
}

/****************************************************************************
 * Signal handler for SIGQUIT
 ****************************************************************************/
RETSIGTYPE flush_caches(sig)
{
	static volatile int	inprogress = 0;

	/* in the middle of processing a request; do not reinitialize yet */
	if (_rpcsvcdirty) {
		need_flush = 1;
		return;
	}
	if (inprogress++)	/* SIGUSR2 handling in progress already? */
		return;
	/* not in the middle of processing a request, not handling another
		 SIGUSR - go ahead and flush caches */
	if(wb_cache_enabled) flush_dirty_blocks();
	if(attr_cache_enabled)
		flush_attr_cache();	/* attribute cache */
	if(data_cache_enabled)
		flush_data_cache();   /* data cache */
#ifdef CALL_PROFILING
	dump_stats(0);         /* dump profiling stats */
#endif
	inprogress = 0;
	need_flush = 0;
}

/****************************************************************************
 * Signal handler for SIGHUP
 ****************************************************************************/
RETSIGTYPE reinitialize(sig)
{
	static volatile int	inprogress = 0;

	/* in the middle of processing a request; do not reinitialize yet */
	if (_rpcsvcdirty) {
		need_reinit = 1;
		return;
	}
	if (inprogress++)	/* SIGHUP handling in progress already? */
		return;
	/* not in the middle of processing a request, not handling another
		 SIGHUP - go ahead and reinitialize input files */
	auth_init(NULL);	/* auth_init remembers the exports file name */
	if(match_sk_enabled)
		if(sk_init(send_sk_path, match_sk_path) != 0) {
			Dprintf(L_ERROR, "Session key initialization error\n");
			match_sk_enabled = 0;
	}
	inprogress = 0;
	need_reinit = 0;
}

/****************************************************************************
 * Signal handler for SIGALRM
 ****************************************************************************/
RETSIGTYPE setalarm(sig)
{
	static volatile int inprogress = 0;

	/* in the middle of processing a request; do not reinitialize yet */
	if (_rpcsvcdirty) {
		need_alarm = 1;
		return;
	}
	if (inprogress++)	/* SIGALRM handling in progress already? */
		return;

	getinv();

	inprogress = 0;
	need_alarm = 0;
}
