/*
 * rpcmisc	Miscellaneous functions for RPC startup and shutdown.
 *		This code is partially snarfed from rpcgen -s tcp -s udp,
 *		partly written by Mark Shand, Donald Becker, and Rick
 *		Sladkey. It was tweaked slightly by Olaf Kirch to be
 *		useable by both nfsd and mountd.
 *
 *              Modified by Renato Figueiredo to work with both TLI
 *              and sockets. The main functions provided in this file are:
 *              - rpc_init(): create RPC service (server side of proxy)
 *              - rpc_exit(): terminate RPC service
 *
 *
 *		This software may be used for any purpose provided
 *		the above copyright notice is retained.  It is supplied
 *		as is, with no warranty expressed or implied.
 *
 *		Auth daemon code by Olaf Kirch.
 */

#include "system.h"
#include <stdio.h>
#include <stdlib.h>
#include <rpc/pmap_clnt.h>
#include <string.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <memory.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include "rpcmisc.h"
#include "logging.h"
#include <time.h>

#ifdef SOLARIS
#include <strings.h>
#define PORTMAP
#include <rpc/rpc.h>
#include <rpc/svc_soc.h>
#endif /* SOLARIS */

/* Another undefined function in RPC */
extern SVCXPRT *svcfd_create(int sock, u_int ssize, u_int rsize);

static int	makesock(int port, int proto, int socksz);

#define _RPCSVC_CLOSEDOWN	120
time_t		closedown = 0;
int		_rpcpmstart = 0;
int		_rpcfdtype = 0;
int		_rpcsvcdirty = 0;

/****************************************************************************
 * Initialize UDP and TCP RPC services - the server-side of the procy
 * The RPC program number and versions are given by prog, vers
 * The RPC dispatch function is given by dispatch
 * The port to listen to is given by defport, and the buffer size by bufsiz
 *
 * If successful, the RPC service is established and RPC calls for prog/vers
 * will all be routed to the dispatch function. If not successful, will
 * exit with fatal error.
 *
 * Called by both mountd.c and nfsd.c when setting up the server-side of
 * the nfsd/mountd proxies.
 ****************************************************************************/
void
rpc_init(const char *name, int prog, int vers, void (*dispatch)(),
			int defport, int bufsiz)
{
  struct sockaddr_in saddr;
  SVCXPRT	*transp;
  int	sock;
  int	asize;
  struct sigaction pipeact;

  /* This is still not very clear to me... */
  asize = sizeof(saddr);
  sock = 0;
  if (getsockname(0, (struct sockaddr *) &saddr, (socklen_t *)&asize) == 0) {
    int	ssize = sizeof (int);

    if (saddr.sin_family != AF_INET)
      goto not_inetd;
    if (getsockopt(0, SOL_SOCKET, SO_TYPE, &_rpcfdtype, (socklen_t *)&ssize) < 0)
      goto not_inetd;
    background_logging();	/* no more logging to stderr */
    closedown = time(NULL) + _RPCSVC_CLOSEDOWN;
  } else {
  not_inetd:
    _rpcfdtype = 0;
//    pmap_unset(prog, vers);
    sock = RPC_ANYSOCK;
  }


  /* initialize UDP service */
  if ((_rpcfdtype == 0) || (_rpcfdtype == SOCK_DGRAM)) {
    if (_rpcfdtype == 0 && defport != 0) {
      sock = makesock(defport, IPPROTO_UDP, bufsiz);
    } else {
      sock = RPC_ANYSOCK;
    }
    /* create buffer for UDP */
    transp = svcudp_bufcreate(sock, bufsiz, bufsiz);
    if (transp == NULL)
      Dprintf(L_FATAL, "cannot create udp service.");
    if (defport == 0) {
      fprintf(stdout, "UDP=%i\n",transp->xp_port);
    } else {
      if (transp->xp_port != defport)
        Dprintf(L_FATAL, "cannot connect to specified udp port %i.",defport);
    }
    /* Register function in dispatch.c for all versions of the service */
    /* Note: svc_register's last argument of "0" replaces IPPROTO_UDP;
             with argument 0, proxy does not register itself w/ portmap */
    if (!svc_register(transp, prog, vers, dispatch, 0)) {
      Dprintf(D_CALL,
	      "unable to register (%s, %d, udp).",
	      name, vers);
    }
  }

  /* initialize TCP service */
  if ((_rpcfdtype == 0) || (_rpcfdtype == SOCK_STREAM)) {
    if (_rpcfdtype == 0 && defport != 0) {
      sock = makesock(defport, IPPROTO_TCP, bufsiz);
    } else {
      sock = RPC_ANYSOCK;
    }
    /* create buffer for TCP */
    transp = svctcp_create(sock, bufsiz, bufsiz);
    if (transp == NULL)
      Dprintf(L_FATAL, "cannot create tcp service.");
    if (defport == 0) {
      fprintf(stdout, "TCP=%i\n",transp->xp_port);
    } else {
      if (transp->xp_port != defport)
        Dprintf(L_FATAL, "cannot connect to specified tcp port %i.",defport);
    }

    /* Register function in dispatch.c for all versions of the service */
    /* Note: svc_register's last argument of "0" replaces IPPROTO_TCP;
             with argument 0, proxy does not register itself w/ portmap */
    if (!svc_register(transp, prog, vers, dispatch, 0)) {
      Dprintf(D_CALL,
	      "unable to register (%s, %d, tcp).",
	      name, vers);
    }
  }

  /* We ignore SIGPIPE. SIGPIPE is being sent to a daemon when trying
   * to do a sendmsg() on a TCP socket whose peer has disconnected.
   */

  memset(&pipeact, 0, sizeof(pipeact));
  pipeact.sa_handler = SIG_IGN;
  sigaction(SIGPIPE, &pipeact, NULL);
}

/****************************************************************************
 * Un-register program w/ portmapper
 ****************************************************************************/
void
rpc_exit(int prog, int vers)
{
  pmap_unset(prog, vers);
}

/****************************************************************************
 * Create a socket; used by rpc_init()
 ****************************************************************************/
static int
makesock(int port, int proto, int socksz)
{
#ifdef PUNCH_RPCTLI
  struct t_bind req;
#endif
  struct sockaddr_in sin;
  const char	*prot_name;
  int		s;
  int		sock_type;
  int           val;

  switch (proto) {
  case IPPROTO_UDP:
    sock_type = SOCK_DGRAM;
    prot_name = "UDP";
    break;
  case IPPROTO_TCP:
    sock_type = SOCK_STREAM;
    prot_name = "TCP";
    break;
  default:
    Dprintf(L_FATAL, "Invalid protocol in makesock\n");
    return -1; /* NOTREACHED */
  }

  /* handles TLI socket interface; probably needs cleaner implementation.
     Anyone? */
#ifdef PUNCH_RPCTLI
  /* open socket "s" */
  if((s = t_open((sock_type == SOCK_DGRAM ? "/dev/udp" : "/dev/tcp"),
		 O_RDWR, (struct t_info *)0)) < 0) {
    Dprintf(L_FATAL, "can't open /dev/[udp/tcp]\n");
  }

  /* Create name data structure to bind to socket. */
  bzero((char *) &sin, sizeof(sin));
  sin.sin_addr.s_addr = htonl(INADDR_ANY);
  sin.sin_port = htons(port);
  sin.sin_family = AF_INET;
  req.addr.maxlen = sizeof(sin);
  req.addr.len = sizeof(sin);
  req.addr.buf = (char *) &sin;
  req.qlen = 5;

  /* bind name to socket */
  if (t_bind(s, &req, (struct t_bind *)0) < 0) {
    Dprintf(L_FATAL, "can't bind local address\n");
  }
#endif /* PUNCH_RPCTLI */

  /* handles regular sockets */
#ifdef PUNCH_RPCSOCK
  /* create unnamed socket */
  s = socket(AF_INET, sock_type, proto);
  if (s < 0)
    Dprintf(L_FATAL, "Could not make a %s socket: %s\n",
	    prot_name, strerror(errno));

  memset((char *) &sin, 0, sizeof(sin));
  sin.sin_family = AF_INET;
  sin.sin_addr.s_addr = INADDR_ANY;
  sin.sin_port = htons(port);


  val = 1;
  if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &val, sizeof(val)) < 0)
    Dprintf(L_ERROR, "setsockopt failed: %s\n", strerror(errno));

#ifdef SO_SNDBUF
  if (socksz != 0) {
    int sblen, rblen;

    /* 1024 for rpc & transport overheads */
    sblen = rblen = 8 * (socksz + 1024);
    if (setsockopt(s, SOL_SOCKET, SO_SNDBUF, &sblen, sizeof sblen) < 0 ||
	setsockopt(s, SOL_SOCKET, SO_RCVBUF, &rblen, sizeof rblen) < 0)
      Dprintf(L_ERROR, "setsockopt failed: %s\n", strerror(errno));
  }
#endif				/* SO_SNDBUF */

  if (bind(s, (struct sockaddr *) &sin, sizeof(sin)) == -1)
    Dprintf(L_FATAL, "Could not bind %s socket to %s:%d: %s\n",
	    prot_name, inet_ntoa(sin.sin_addr),
	    ntohs(sin.sin_port),
	    strerror(errno));
#endif /* PUNCH_RPCSOCK */

  return (s);
}

